/*****************************************************************************************
 * $Id: orva_basel_aggr.sas,v 1.1.4.21.2.20.8.1 2012/03/06 00:53:23 jariou Exp $
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved. 
 *
 * NAME    : orva_basel_aggr.sas  
 * PURPOSE : Calculates both aggregate and basel aggregate VaR.     
 *           The BASEL method sorts by loss amount and then sum across reps.
 *           The AGGR method sorts by rep and then sums across reps.      
 * SUPPORT : stchri         
 *
 * PARAMETERS :
 *
 *   varSimData     = input/output VaR Simulation dataset for interior cells.
 *                    This dataset may be updated by this macro! 
 *   varStatData    = the VaR Statistics dataset for interior cells.      
 *   ccData         = the cross classification dataset for project.
 *   confIntData    = the confidence interval dataset. 
 *   aggrCorr       = the user defined correlation matrix for
 *                    aggregate cells.
 *   outSimBasel    = output basel aggregate simulation dataset.
 *   outStatBasel   = output basel aggregate statistics dataset.
 *   outSimAggr     = output aggregate simulation dataset.
 *   outStatAggr    = output aggregate statistics dataset.
 *   outStatus      = output status dataset.
 *   copulaFamily   = NORMAL, NORMALMIX, or T.
 *   copulaParams   = the parameters going along the copula specification.
 *                    See documentation for orva_aggr_corr().
 *   aggrType       = STD or AGGR or BOTH 
 *****************************************************************************************/
%macro orva_basel_aggr( 
                        varSimData   = ,
                        varStatData  = , 
                        ccData       = ,
                        confIntData  = ,
                        aggrCorr     = ,
                        outSimBasel  = ,
                        outStatBasel = ,
                        outSimAggr   = ,
                        outStatAggr  = ,
                        outStatus    = status,
                        copulaFamily = NORMAL,
                        copulaParams = ,
                        aggrType     = BOTH
                        );
                       
    /*- create the status data set up front -*/
    %let reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
    %orva_status_ds( statusds = &outstatus, status = 1, reason = &reason );
    %let status = 0;
    %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));
    %let abort  = 0;
 
    /*- Do we have statistics data to aggregate? -*/
    %let nobs = 0;
    %orvu_getObsCount(&varStatData, nobs);
    %if &nobs eq 0 %then %do;
        %let status = 1;
        %let reason = %sysfunc(sasmsg(&orvads, bag_no_varstat_data_error, noquote)); 
        %put &orvar_note %sysfunc(sasmsg(&orvads, bag_no_varstat_data_error, noquote)); 
        %goto DONE;
    %end;
    
    /*- Do we have simulated data to aggregate? -*/
    %let nobs = 0;
    %orvu_getObsCount(&varSimData, nobs);
    %if &nobs eq 0 %then %do;
       %let reason = %sysfunc(sasmsg(&orvads, bag_no_varsim_data_error, noquote));
       %orva_status_ds( statusds = &outstatus, status = 1, reason = &reason ); 
       %put &orvar_note %sysfunc(sasmsg(&orvads, bag_no_varsim_data_error, noquote)); 
       %goto DONE;
    %end; 
    
    /*- Get the number of replications -*/
    %let nreps = 0;
    data _NULL_;
      set &varStatData (obs = 1);
      call symputx('NREPS', nreps);
    run;

    /*- initialize timers -*/
    %baselaggr_setupTime;
    
    /*****************************************************************/
    /***************** Run orva_aggr_corr() first ********************/
    /*****************************************************************/
    %if &aggrCorr NE %str() %then %do;
       %orva_aggr_corr(
                       simData      = &varSimData, 
                       nreps        = &nreps, 
                       corrData     = &aggrCorr,
                       copulaFamily = &copulaFamily,
                       copulaParams = &copulaParams,
                       outStatus    = &outStatus
                       );
      proc sql noprint; 
         SELECT 
                  status, 
                  reason 
         INTO 
                  :status, 
                  :reason 
         FROM 
                  &outStatus; 
       quit;
       %if &status ne 0 %then %do;
           %put &orvar_error %sysfunc(sasmsg(&orvads, bag_aggr_call_fail_error, noquote, &reason));
           %put &orvar_warning %sysfunc(sasmsg(&orvads, bag_orig_var_simdat_agg_warning, noquote)); 
           %let status = 0;
           %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote));
       %end;
       
       %baselaggr_getElapsedTime;
       %put %sysfunc(sasmsg(&orvads, bag_aggr_corr_time, noquote, &_TimeElapsed_));
   %end;
    
   /*** Get a list of cross-class variables into BYVARS ***/
   proc contents data = &ccdata out = c noprint; 
   run;

   data _NULL_; set c;
      retain byvars;
      length byvars $ 512;

      if upcase(name) ^= upcase("CCUID") then do;
         byvars = trim(byvars) || ' ' || trim(name);
         call symput( 'byvars', byvars );
      end;
   run;

   /*** Create dataset with total cells only ***/
   data _totcc_; 
      set &ccData end = LAST;
      where
        %let nbyvars = 1;
        %let bvar = %scan(&byvars, &nbyvars, ' ');

        %do %while (&bvar ne ); 
              &bvar = 0 or
              %let nbyvars = %eval(&nbyvars + 1);
              %let bvar = %scan(&byvars, &nbyvars, ' ');
        %end;
           0;
      retain count 0;
      count + 1;

      if LAST then call symputx( "totobs", count);
   run;
   
   /**** create macros for each total cell ****/
   proc SQL noprint;
      SELECT 
               CCUID 
      INTO 
               :Tot_CCUID_1 - :Tot_CCUID_&totobs
      FROM 
               _totcc_;
   quit;


   /*****************************************************************/
   /* 
    *  Order simulation dataset from smallest to largest
    *  so we can aggregate like losses together across cells
    */
   /*****************************************************************/
   proc sort 
            data = &varSimData 
            out  = _tmpSimData_;
       by CCUID loss;
   run;  
   %baselaggr_getElapsedTime;
   %put %sysfunc(sasmsg(&orvads, bag_sort_sim_time, noquote, &_TimeElapsed_));
  
   /*****************************************************************/
   /*************** loop over aggregate cells ***********************/
   /*****************************************************************/
 
   %do iiii = 1 %to &totobs;
   
      %let CELLS_IN =;
      %orva_cells_in(ccuid=&&Tot_CCUID_&iiii, ccData=&ccData);
           
      /****************************************************************/
      /**************** aggregate SIMULATION for BOTH *****************/
      /****************************************************************/
      %let haveData= false;
      data _tempsim_basel _tempsim_aggr; 
        set _tmpSimData_ end=LAST;
        /* get interior cells for this aggregate cell */
        where CCUID in(
             %let ncells=1;
             %let cell=%scan(&CELLS_IN, &ncells, ' ');
             %do %while (&cell ne %str()); 
                  %if &ncells ne 1 %then %do; , %end; &cell 
                   
               %let ncells=%eval(&ncells+1);
               %let cell=%scan(&CELLS_IN, &ncells, ' ');
             %end;
        );
       
        /*- handle BASEL summation, sums sorted losses and recreates _rep_ -*/
        %if &aggrType = BOTH OR &aggrType = STD %then %do;    
           retain _basel_rep_ 1;
           array mat_basel[&nreps] _temporary_;         
           mat_basel[_basel_rep_] + LOSS;
           if (_basel_rep_ eq &NREPS) then _basel_rep_ = 1;
           else _basel_rep_ + 1;
        %end;
        
        /*- handle AGGR summation, sums across _rep_ -*/
        %if &aggrType = BOTH OR &aggrType = AGGR %then %do;    
           array mat_aggr[&nreps] _temporary_;
           mat_aggr[_rep_] + LOSS;
        %end;
        
        /*- write to BOTH output datasets -*/
        if LAST then do;
             CCUID = &&Tot_CCUID_&iiii;
             do i=1 to &NREPS;
             
                _rep_=i;
             
                %if &aggrType = BOTH OR &aggrType = STD %then %do;    
                   loss = mat_basel[_rep_];
                   output _tempsim_basel;
                %end;
     
                %if &aggrType = BOTH OR &aggrType = AGGR %then %do;    
                   loss = mat_aggr[_rep_];
                   output _tempsim_aggr;
                %end;
             end;
             call symput('haveData', 'true');
        end;
     
        keep CCUID Loss _rep_ ;

      run;
      
      /*- append to output simulation datasets -*/
      %if &haveData eq true %then %do;
           %if &aggrType = BOTH OR &aggrType = STD %then %do;    
              %orvu_append(base=&outSimBasel,new=_tempsim_basel,canUseAppend=true);
           %end;
           %if &aggrType = BOTH OR &aggrType = AGGR %then %do;   
              %orvu_append(base=&outSimAggr,new=_tempsim_aggr,canUseAppend=true);           
           %end;
      %end;
      %else %goto NEXT_CELL;

      /****************************************************************/
      /**************** aggregate BASEL STATISTICS ********************/
      /****************************************************************/
      %let haveData= false;
      %if &aggrType = BOTH OR &aggrType = STD %then %do;
         /* perform simple summation of statistics */
         data _tempstat_basel; 
           set &varStatData end=LAST;
           where CCUID in(
                %let ncells=1;
                %let cell=%scan(&CELLS_IN, &ncells, ' ');
                %do %while (&cell ne %str()); 
                     %if &ncells ne 1 %then %do; , %end; &cell 
                  %let ncells=%eval(&ncells+1);
                  %let cell=%scan(&CELLS_IN, &ncells, ' ');
                %end;
           );
                 
           retain basel_MEAN 
             %do i=1 %to 3; basel_P_&i basel_UCL&i basel_LCL&i basel_VAREVT_&i basel_CONDEXPECTEDLOSS_&i %end;
             ;        
           basel_MEAN + MEAN;
           %do i=1 %to 3;
              basel_P_&i + P_&i;
              basel_UCL&i + UCL&i;
              basel_LCL&i + LCL&i;
              basel_VAREVT_&i + VAREVT_&i;
              basel_CONDEXPECTEDLOSS_&i + CONDEXPECTEDLOSS_&i;
              basel_param_risk_lower_&i + param_risk_lower_&i;
              basel_param_risk_upper_&i + param_risk_upper_&i;
           %end;   
           
           if LAST then do;
              CCUID = &&Tot_CCUID_&iiii;
              MEAN = basel_MEAN;
              %do i=1 %to 3;
                P_&i = basel_P_&i;
                UCL&i = basel_UCL&i;
                LCL&i = basel_LCL&i;
                VAREVT_&i = basel_VAREVT_&i;
                CONDEXPECTEDLOSS_&i =  basel_CONDEXPECTEDLOSS_&i;
                param_risk_lower_&i = basel_param_risk_lower_&i;
                param_risk_upper_&i = basel_param_risk_upper_&i;
              %end;
              output;
              call symput('haveData', 'true');
           end;
           
           keep CCUID MEAN 
             %do i=1 %to 3; P_&i UCL&i LCL&i VAREVT_&i CONDEXPECTEDLOSS_&i %end;
           ;
           
         run;
      
         /*- append to outStatBasel dataset -*/
         %if &haveData eq true %then %do;
           %orvu_append(base=&outStatBasel,new=_tempstat_basel,canUseAppend=true);
         %end;
         
      %end;

      /********************************************************/
      /**************** aggregate AGGR STATISTICS *************/
      /********************************************************/
      %if &aggrType = BOTH OR &aggrType = AGGR %then %do;     
      
         %orva_get_var_stats(
            simData     = _tempsim_aggr, 
            outStat     = _tempstat_aggr, 
            lossVar     = loss,
            confIntData = &confIntData
            );   

         %orva_compute_expected_shortfall( 
                    simdata = _tempsim_aggr, 
                    outstat = _tempstat_aggr, 
                    nreps   = &nreps 
                    );
         data _tempstat_aggr; set _tempstat_aggr;
            CCUID = &&Tot_CCUID_&iiii;
         run;

         /*----------------------------------------------------------*/
         /*- Compute the aggregate parameter risk numbers -----------*/
         /*- This is an approximation: 
             Reduce the basel confidence limits by the same percentage
             as the reduction in aggregate var             
             --------------------------------------------------------*/
         data _NULL_; set _tempstat_aggr;
             call symputX("P_1", put(p_1,BEST20.));
             call symputX("P_2", put(p_2,BEST20.));
             call symputX("P_3", put(p_3,BEST20.));
         run;
         data _tempstat_parameter_risk; 
           set &varStatData end=LAST;
           where CCUID in(
                %let ncells=1;
                %let cell=%scan(&CELLS_IN, &ncells, ' ');
                %do %while (&cell ne %str()); 
                     %if &ncells ne 1 %then %do; , %end; &cell 
                  %let ncells=%eval(&ncells+1);
                  %let cell=%scan(&CELLS_IN, &ncells, ' ');
                %end;
           );
           %do i=1 %to 3;
              basel_P_&i + P_&i;
              basel_param_risk_lower_&i + param_risk_lower_&i;
              basel_param_risk_upper_&i + param_risk_upper_&i;
              keep param_risk_lower_&i param_risk_upper_&i;
           %end;   
           if LAST then do;
              %do i=1 %to 3;
                 ratio =  &&P_&i / basel_P_&i ;
                 param_risk_lower_&i = basel_param_risk_lower_&i * ratio;
                 param_risk_upper_&i = basel_param_risk_upper_&i * ratio;
              %end;
              output;
           end;
         run;
         data _tempstat_aggr;
            set _tempstat_parameter_risk;
            set _tempstat_aggr;
         run;

         /*----------------------------------------------------------*/
         /*- End parameter risk inclusion --------------------------*/
         /*----------------------------------------------------------*/

         /*- append to outStatAggr dataset -*/
         %orvu_append(base=&outStatAggr, new=_tempstat_aggr, canUseAppend=true);
       
     %end;

     %baselaggr_getElapsedTime;
     %put %sysfunc(sasmsg(&orvads, bag_aggr_cell_time, noquote, &&Tot_CCUID_&iiii, &_TimeElapsed_));
     
%NEXT_CELL:
   %end;
       
   
%DONE:
   data _NULL_;
      now = datetime();
      elapsed = now-symget('_TimeStart0_');
      call symputx('_TimeElapsed_',elapsed);
   run;
   %put %sysfunc(sasmsg(&orvads, bag_total_time, noquote, &_TimeElapsed_));
 
  
   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, bag_abnormal_error, noquote));
      %let status=1;
   %end;   
   %orva_status_ds( statusds=&outstatus, status=&status, reason=&reason );

   proc datasets library=work memtype=data nolist;
   delete _tempstat_: _tempsim_:  _totcc_  c;
quit;

%mend orva_basel_aggr;

%macro baselaggr_getElapsedTime;
   data _NULL_;
      now = datetime();
      elapsed = now-symget('_TimeStart_');
      call symputx('_TimeElapsed_',elapsed);
      call symputx('_TimeStart_',now);
   run;
%mend baselaggr_getElapsedTime;


%macro baselaggr_setupTime;
   data _NULL_;
      now = datetime();
      call symputx('_TimeStart_',now);
      call symputx('_TimeStart0_',now);
   run;
%mend baselaggr_setupTime;



/******************************************************************
 Sample call:

%orvsetup(version=3.3);
libname varprj "C:\_VaR331_Aug_Data\ver_1\components\projects\Proj12020";
option mprint notes symbolgen;
%orva_basel_aggr(                                     
                        varSimData   = varprj.varSim_Naked,     
                        varStatData  = varprj.varStat_Naked,    
                        ccData       = varprj.cc,              
                        confIntData  = VARPRJ.simIndices,      
                        outSimBasel  = baselSim_Naked,   
                        outStatBasel = baselStat_Naked,          
                        outSimAggr   = aggSim_Naked,    
                        outStatAggr  = aggStat_Naked,    
                        outStatus    = status                 
                        );                                    
                                                     
**********************************************************************/
