/*****************************************************************************************
 * $Id: orva_combine_cov.sas,v 1.1.4.18.2.16.2.2.8.1 2012/05/03 13:53:44 sasdje Exp $
 *  Copyright (C), 2005
 *  SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    ORVA_COMBINE_COV                                             
 *  PURPOSE: Take the kronecker product of two covariance matrices
 *  SUPPORT: jariou
 *
 *  PARAMETERS:
 *
 *  cov1  = Name of first covariance matrix - corresponds to ccuid_0
 *          If is blank a Identity matrix will be used
 *           
 *  cov2  = Name of the second covariance matrix - corresponds to ccuid_1
 *          If is blank a Identity matrix will be used
 *  
 *  cov3  = Name of the third covariance matrix - corresponds to ccuid_3
 *          If is blank a Identity matrix will be used
 *
 *  ccLibref =  the name of the libref with the cc data information
 *
 *  GenTotals = 0 or 1 to indicate whether covariances for total cells are needed.
 *      
 *  Aggr  = true, if this is the Aggregate Cov matrix
 *          false, if it is the Frequency Cov matrix
 *
 *  Reduced = true, to use Reduced Kronecker Product when combinging matrices
 *            false, to use Kronecker Product when combinging matrices
 *
 *  outCov = Output covariance matrix
 *  
 *  ostat = name of the out status dataset.  Will contain the following:
 *           status = 0 on success, 1 otherwise.
 *           reason = printable message.
 *      
 *****************************************************************************************/
 %macro ORVA_COMBINE_COV( COV1          = ,
                          COV2          = ,
                          COV3          = ,
                          ccLibref      = varprj., 
                          GenTotals     = 1,
                          Reduced       = true,
                          Aggr          = false,
                          outCov        = outcov,
                          ostat         =
									);


    /*- create the status dataset up front -*/
    %let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
    %orva_status_ds( statusds=&ostat, status=1, reason=&reason );

    %let status=0;
    %let abort=0;
    %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));

    options sortdup=logical;

   /*- Determine the number of dimensions ------------*/
   %let NUMCCVS=0;
   proc contents noprint data=&ccLibref.cc out=_cc; run;
   %let dsid=%sysfunc(open(_cc,I));
   %if &dsid ^= 0 %then %do;
      %let NUMCCVS=%sysfunc(attrn(&dsid,nlobs));
      %let rc=%sysfunc(close(&dsid));
   %end;
   %if &NUMCCVS eq 0 %then %return;
   %let NUMCCVS= %eval(&numccvs-1);

      /* for simple identity matrices skip most of the work ---------------------*/
   %if &cov1=%str() and &cov2=%str() and &cov3=%str() %then %do;
      %_orva_create_identity_matrix( &ccLibref, &outcov, &numccvs);
      %goto DONE;
   %end;

   /*- check input dataset if missing make a matrix  -*/
   options sortdup=logical;
   %_orva_Check_Cov_Mat(dimension=1);
   %let _var_names1=;
   %let _matrix_size1=;


	%let _var_names1   = %str();
	%let _matrix_size1 = %str();
	%_orva_List_Dims_And_Names(&ccLibref,0,&cov1,cov1_tmp,_var_names1,_matrix_size1);

   %_orva_Check_Cov_Mat(dimension=2);


	%let _var_names2   = %str();
	%let _matrix_size2 = %str();

   %_orva_List_Dims_And_Names(&ccLibref, 1, &cov2, cov2_tmp, _var_names2, _matrix_size2);

   %if &NUMCCVS = 3 %then %do;
      %_orva_Check_Cov_Mat(dimension=3);
      %let _var_names3   = %str();
      %let _matrix_size3 = %str();
      %_orva_List_Dims_And_Names(&ccLibref,2,&cov3,cov3_tmp,_var_names3,_matrix_size3);
   %end;

   %if &NUMCCVS = 3 %then %do;
	   %_orva_build_all( covs   = cov2_tmp cov1_tmp cov3_tmp,
                        sizes  = &_matrix_size2 &_matrix_size1 &_matrix_size3,
                        vars   = &_var_names2 &_var_names1 &_var_names3,
                        outCov = &outCov
                        );
   %end;
   %else %do;
      %_orva_build_all( covs   = cov2_tmp cov1_tmp,
                        sizes  = &_matrix_size2 &_matrix_size1,
                        vars   = &_var_names2 &_var_names1,
                        outCov = &outCov
                        );
   %end;
    
   /* Return a correlation matrix rather than a covariance matrix */
   %let NVARS=0;
   %let dsid=%sysfunc(open(&outCov,I));
   %if &dsid ^= 0 %then %do;
      %let NVARS=%sysfunc(attrn(&dsid,nlobs));
      %let rc=%sysfunc(close(&dsid));
   %end;
   
   %orva_cov2corr(&outCov,&outCov,&NVARS);
   
%DONE:
    %if &abort %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, cobv_abnormal_error, noquote)) ;
        %let status=1;
    %end;    
    %orva_status_ds( statusds=&ostat, status=&status, reason=&reason );

    /***- delete unwanted data sets -***/
    proc datasets library=work memtype=data nolist nodetails;
      delete _cc cov1_tmp cov2_tmp _cov1_ _cov2_
            %if &NUMCCVS = 3 %then %do;
               cov3_tmp
               _cov3_
            %end;
              ;
    quit;
%mend ORVA_COMBINE_COV;

%macro _orva_Check_Cov_Mat(dimension=);
    %if &&&COV&dimension=%str()  %then %do;

      %let dim=%sysevalf(&dimension-1);

		proc sort data = &ccLibref.cc(keep=ccvUID_&dim) out = _CC_sorted_(where=( ccvUID_&dim>0)) nodupkey;
			by ccvUID_&dim;
		run;

		data _CC_sorted_plus_;
			set _CC_sorted_;
			self = ccvUID_&dim;
		run;

      proc transpose data=_CC_sorted_plus_ out=_ccmap_tmp_ prefix=uid;
         id ccvUID_&dim;
      run;
        
       data _cov&dimension._; 
         length _NAME_ $20;
         set _ccmap_tmp_(drop=_name_);
         array y[*] uid:;

         do i =1 to dim(y);
            _NAME_ = trim(vname(y[i]));

            do j = 1 to dim(y);
               if ( j = i ) then y[j] = 1;
               else y[j] = 0;
            end;

            output;
          end;

           drop i j;
       run;
        %let cov&dimension=_cov&dimension._;
    %end;

%_orva_Check_Cov_Mat_DONE:

    proc datasets library=work memtype=data nolist nodetails;
      delete _ccmap_tmp_ _cc_sorted_ _cc_sorted_plus_ ;
    quit;
%mend _orva_Check_Cov_Mat;


%macro _orva_List_Dims_And_Names(libref,_dimCount,inMatrix,outMatrix,outVars,outSize);
   /* Sort cc dataset by CCVUID_??? to prepare for summary */
   /* Output has columns: CCUID, CCVUID_0, CCVUID_1,...    */
   proc sort data=&libref.cc out = _cc_ ;
      by ccvuid_&_dimCount;
   run;
   
   /* Keep a single value, the minimum, of CCUID per CCVUID_??? */
   /* Output has columns: CCUID and CCVUID_???                  */
   proc summary data= _cc_;
      var ccuid;
      by ccvuid_&_dimCount;
      where ccvuid_&_dimCount NE 0;
      output
         out= cc_&_dimCount(drop =_freq_ _type_)
         min = CCUID;
   run;

   proc sort data=cc_&_dimCount;
      by ccvuid_&_dimCount;
   run;
      
	proc SQL NOPRINT;
		SELECT _NAME_
		INTO : cov_vars separated by " "
		FROM &inMatrix;
 
		SELECT count(_NAME_)
		INTO : cov_SIZE 
		FROM &inMatrix;
	QUIT;




   %let &outVars = ;
   %let &outSize = ;

   %_orva_sum_one_dim(&inMatrix);

   proc transpose data= matrix_tmp_summ out=matrix_tmp_2;
      id _NAME_;
   run;

   %_orva_sum_one_dim(matrix_tmp_2);

   data &outMatrix;
      set matrix_tmp_summ;
   run;
   
   %let &outVars = &cov_vars;
   %let &outSize = &cov_size;

%_orva_List_Dims_And_Names_DONE:

    proc datasets library=work memtype=data nolist nodetails;
      delete _cc_ cc_&_dimCount cc_tmp matrix_tmp_2 matrix_tmp_summ ;
    quit;
%mend _orva_List_Dims_And_Names;
   
%macro _orva_sum_one_dim(matrix);

	proc sql noprint;
		CREATE TABLE matrix_tmp
		AS SELECT  * FROM &matrix A, cc_&_dimCount B
		WHERE input(substr(A._NAME_,4),BEST20.)  = B.CCVUID_&_dimCount;
	quit;

   /* Add a _NAME column */
   data matrix_tmp_summ;
      format _NAME_ $12. ;
      retain vars "";
      set matrix_tmp(drop=_name_) end=last;
      _NAME_="COL" || trim(left(_N_));
      drop ccuid ;

      if last then do;
         call symputx( "cov_size", _N_);
      end;
   run;

	proc SQL NOPRINT;
		SELECT _NAME_
		INTO : cov_vars separated by " "
		FROM matrix_tmp_summ;
 	QUIT;
 
%_orva_sum_one_dim_DONE:

    proc datasets library=work memtype=data nolist nodetails;
      delete matrix_tmp;
    quit;
%mend _orva_sum_one_dim;




/****************************************************************/
/* This macro builds the whole mess of covariance matrices      */
/* and assembles them together to form the one huge covariance  */
/* matrix including the sum cells.                              */
/*                                                              */
/* The COVS parameter is a space delimited list of names of     */
/* covariance matrix datasets.                                  */
/*                                                              */
/* The SIZES parameter is a space delimited list indicating the */
/* size of each corresponding covariance matric in COVS.        */
/*                                                              */
/* The VARS parameter is a space delimited list of variable     */
/* names for each corresponding covariance matric in COVS. The  */
/* lists are concatenatd into a single list.                    */
/****************************************************************/
%macro _orva_build_all( 
                        covs  =,
                        sizes =,
                        vars  =,
                        outCov=);
   %let cumSize = 0;
   %let nCov = 1;
   %let Vars2=;
   %let cov  = %scan(&covs, &nCov, ' ');
   %let size = %scan(&sizes, &nCov, ' ');
   
   /** Begin (1) ************************************************/
   /* Create a list of macro variables:                         */
   /*                                  _TheVars_1 to _TheVars_n */
   /*                                  _TheCov_1  to _TheCov_n  */
   /*                                  _TheSize_1 to _TheSize_n */
   /*************************************************************/
   %do %while ("&cov" ne "" );
      %let Vars1=;

      %do i=1 %to &size;
         %let Var1=%scan(&vars,%sysevalf(&cumSize+&i),' ');
         %let Vars1=&Vars1 &var1;
      %end;

      %let TheVars_&ncov=&Vars1;
      %let TheCov_&ncov=&cov;
      %let TheSize_&ncov=&size;

      %let cumSize=%eval(&cumSize + &size);
      %let size=%scan(&sizes, &nCov+1, ' ');
      %let nCov=%eval(&nCov+1);
      %let LastCov=&cov;
      %let cov=%scan(&covs, &nCov, ' ');
      %let Vars2=&vars1;
   %end;
   %let nCov=%eval(&ncov-1);
   /** End (1) **************************************************/
   



   /** Begin (2) ************************************************/
   /* Create a 1x1 matrix to augment each covariance matrix    **/
   /*************************************************************/
   data _one_;
      col1=1;
   run;
   /** End (2) **************************************************/


   /** Begin (3) ************************************************/
   /* Calculate the Kronecker products and join the resulting   */
   /* matrix with the previous results                          */
   /*************************************************************/
   %let _LastCov   = _one_;
   %let _LastNames = col1;
   %let _LastSize  = 1;
   %let _NewNames  = ;
   %let _NewSize   = ;

   %do _ncovCount=1 %to &nCov;
      %_orva_kronecker_product(&_LastCov,
                              &_LastNames, 
                              &&&TheCov_&_ncovCount,
                              &&&TheVars_&_ncovCount,
                              &Reduced,
                              _TmpCov1
                              );

      %let _NewSize=%eval(&_LastSize * &&TheSize_&_nCovCount);
      %_orva_list_Names(   col,
                           1,
                           &_NewSize ,
                           1,
                           _NewNames
                        );

      %_orva_block_matrix( _TmpCov1,
                           &_NewNames,
                           &_LastCov,
                           &_LastNames,
                           _TmpCov2
                           );

      %let _LastSize=%eval(&_NewSize + &_LastSize);
      %_orva_list_Names(   col,
                           1,
                           &_LastSize ,
                           1,
                           _LastNames
                        );
      
      %let _LastCov   = _tmpCov2;
   %end;

   /** End (3) **************************************************/

   %let _CompleteList_=;
   %_orva_list_all(&sizes, _CompleteList_);
   
   /* move the total columns to the right places according to _CompleteList_*/
   %let commaDelimtedCols=%_orva_createList(&_CompleteList_, %str( ), %str(,), COL);
   %let spaceDelimtedCols=%_orva_createList(&_CompleteList_, %str( ), %str( ), COL);
   
   /* move the total columns and total-total columns to the right place by 
    * selecting them according to the _completeList_ once, then a transpose 
    * opertaion with the same order, and them move the total columns one more 
    * time
    */
   proc sql noprint;
       create table WORK.reorder1 as select &commaDelimtedCols 
       from _tmpCov2;
   quit;
   
   proc transpose data= reorder1 out=_tmp_cov2;
       var &spaceDelimtedCols;
   run;
   
   proc sql noprint;
       create table WORK.reorder2 as select &commaDelimtedCols 
       from _tmp_Cov2;
   quit;
   
   /*no need to do another transpose assuming the matrix is symmetric */
   data &outCov;
      format _NAME_ $15.;
      format cell  &ORVAR_CELLFMT;

      set reorder2;
      array tmp(&_LastSize) _temporary_ (&_CompleteList_);
      cell=tmp[_N_];
      _NAME_=compress("Cell_"||put(cell,&ORVAR_CELLFMT)); 
      drop cell;

   run;
      
   proc transpose data= &outCov out=_tmp_cov;
      id _NAME_;     
   run;   

   data &outCov;
      format _NAME_ $15.;
      format cell  &ORVAR_CELLFMT;

      set _tmp_Cov;
      array tmp(&_LastSize) _temporary_ (&_CompleteList_);
      cell=tmp[_N_];
      _NAME_=compress("Cell_"||put(cell,&ORVAR_CELLFMT)); 
      drop cell;

   run;

%_orva_build_all_DONE:

   proc datasets library=work memtype=data nolist nodetails;
      delete _ONE_ _tmpCov1 _tmpCov2 _tmp_cov;
   quit;
%mend _orva_build_all;


/************************************************************/
/* This macro takes a name prefix and sequence parameters   */
/* FROM, TO, and BY to create a list of names all starting  */
/* with the prefix and having the number in the sequence    */
/* added to it.                                             */
/*                                                          */
/* The resulting list is put in a macro variable, the name  */
/* of which is passed as the LISTVAR parameter              */
/************************************************************/
%macro _orva_List_Names(
                        prefix, 
                        from, 
                        to, 
                        by, 
                        listVar
                        );
   %local _listVar_;

   %do ListName_i = &from %to &to %by &by;
      %let _listVar_ = &_listVar_ &prefix.&ListName_i;   
   %end;
   
   %let &listVar = &_listVar_;
%mend _orva_List_Names;


/****************************************************************/
/* This macro creates the Kronecker product of two matrices.    */
/*                                                              */
/* The parameters COV1 and COV2 are the names of datasets that  */
/* contain the two matrices data.                               */
/*                                                              */
/* The parameters VARS1 and VARS2 are list of variables in the  */
/* respective datasets that are to be included in the matrix    */
/*                                                              */
/* The result is put in a dataset the name of which is passed   */
/* in parameter OUTCOV                                          */
/****************************************************************/
%macro _orva_kronecker_product(
                                 cov1,
                                 vars1,
                                 cov2,
                                 vars2,
                                 Reduced, 
                                 outCov
                                 );

   proc iml;
      use &cov1;
      read all var{&vars1} into cov1;
      close &cov1;

              /* Default covariance to 1 */
      do i = 1 to ncol(cov1);
         if cov1[i,i] = 0 then cov1[i,i] = 1;
      end;

      use &cov2;
      read all var{&vars2} into cov2;
      close &cov2;

      do i = 1 to ncol(cov2);
         if cov2[i,i] = 0 then cov2[i,i] = 1;
      end;

      cov = cov2 @ cov1;

      %if &Reduced=true %then %do;
         cov_0 = (cov2-diag(cov2)) @ (cov1-diag(cov1));
         cov = cov - cov_0;
      %end;

      /* output the final covariance matrix */
      col="col";
      row="_name_";
      create &outCov from cov;
      append from cov;
      close &outCov;      
   quit;
%mend _orva_kronecker_product;

/****************************************************************/
/* This macro creates a diagonal block matrix with the two      */
/* parameters MATRIX1 and MATRIX2 being the names of datasets   */
/* containing the matrices data.                                */
/*                                                              */
/* The parameters VARS1 and VARS2 are list of variables in the  */
/* respective datasets that are to be included in the matrix    */
/*                                                              */
/* The result is put in a dataset the name of which is passed   */
/* in parameter OUTMATRIX                                       */
/****************************************************************/
%macro _orva_block_matrix(
                           matrix1,
                           vars1,
                           matrix2,
                           vars2,
                           outMatrix
                           );
   proc iml;
      use &matrix1;
      read all var{&vars1} into matrix1;
      close &matrix1;

      use &matrix2;
      read all var{&vars2} into matrix2;
      close &matrix2;

      matrix= (matrix1||j(nrow(matrix1),ncol(matrix2),0))//(j(nrow(matrix2),ncol(matrix1),0)||matrix2);

      /* output the final covariance matrix */
      col="col";
      row="_name_";
      create &outMatrix from matrix;
      append from matrix;
      close &outMatrix;      
   quit;
%mend _orva_block_matrix;



/****************************************************************/
/* This macro takes a list of integers in its "list" argument   */
/* and adds a value "x" to each member of the list              */
/*                                                              */  
/* The resulting list is placed in a mcro variable the name of  */
/* which is passed in the "ListPlus" argument                   */
/****************************************************************/
%macro _orva_Add_to_List(x, list, ListPlus);
   %let &ListPlus=;
   %let nList = 1;
   %let item = %scan(&list, &nList, ' ');
   
   %do %while ("&item" ne "" );  
      %let &ListPlus= &&&ListPlus %eval(&item+&x);
      %let nList=%eval(&nList+1);
      %let item=%scan(&list, &nList, ' ');
   %end;
%mend _orva_Add_to_List;

/****************************************************************/
/* This macro takes an ordered list of dimensions "dims" for a  */
/* covariance matrix and a list "code" of 0's or 1's indicating */
/* whether the corresponding dimension is a sum or not.         */
/*                                                              */ 
/* It creates a list of cell number that is placed in the macro */
/* variable whose name is passed in the "ListFromCode" argument */
/****************************************************************/
%macro _orva_list_from_code(dims, code, ListFromCode);
   %local nDims;
   %let nDims = 1;
   %let dim = %scan(&dims, &nDims, ' ');
   %let iCode = %scan(&code, &nDims, ' ');
   
   %do %while ("&dim" ne "" );  
      %let dim&nDims=&dim;
      %let code&nDims=&iCode;
      %let nDims=%eval(&nDims+1);
      %let dim=%scan(&dims, &nDims, ' ');
      %let iCode=%scan(&code, &nDims, ' ');
   %end;
   %let dimSize=%eval(&nDims-1);

   %let &ListFromCode=;
   %let theCode=;
   %let thatListPlus=;

   %do i=1 %to &dimSize;
      %if &i=1 %then %do;
         %if &Code1=0 %then %do;
            %do j=1 %to &dim1 ;
               %let &ListFromCode= &&&ListFromCode &j;
            %end;
         %end;
         %else %do;
               %let &ListFromCode= %eval(&dim1+1);
         %end;
      %end;
      %else %do;
         %if &&Code&i=0 %then %do;
            %let extendedList=&&&ListFromCode;
            %let factor=1;

            %do j=1 %to %eval(&i-1);
               %let factor=%eval(&factor*(&&dim&j+1));
            %end;

            %do j=1 %to %eval(&&dim&i-1) ;
               %_orva_Add_to_List(%eval(&j*&factor),&&&ListFromCode, thatListPlus)
               %let extendedList= &extendedList &thatListPlus ;
            %end;
            %let &ListFromCode= &extendedList;
         %end;
         %else %do;
            %let extendedList=&&&ListFromCode;
            %let factor=1;

            %do j=1 %to %eval(&i-1);
               %let factor=%eval(&factor*(&&dim&j+1));
            %end;

            %let factor= %eval(&factor * &&dim&i);
            %_orva_Add_to_List(&factor,&&&ListFromCode, thatListPlus)
            %let &ListFromCode= &thatListPlus;
         %end;
      %end;
   %end;
%mend _orva_list_from_code;


/****************************************************************/
/* This macro takes an ordered list of dimensions "dims" for a  */
/* covariance matrix and it creates a list of cell number that  */
/* that is placed in the macro variable whose name is passed in */
/* the "CompleteList" argument                                  */
/****************************************************************/
%macro _orva_list_all(dims, CompleteList);
   %let nDims = 1;
   %let _dim_ = %scan(&dims, &ndims, ' ');
   
   %do %while ("&_dim_" ne "" );  
      %let nDims=%eval(&nDims+1);
      %let _dim_=%scan(&dims, &nDims, ' ');
   %end;
   %let nDims=%eval(&nDims-1);

   %let _CompleteList_=;
   %let _code_=;

   %do _i_=0    %to %eval(2**&nDims-1);
      data _NULL_;
         length x $25;
         y=&_i_;

         do i=1 to &nDims;
            x=trim(left(x)) || " " || trim( left(mod(y,2)));
            y=int(y/2);
         end;

         call symputX("_code_", x);
      run;

      %let theList=;
      %_orva_list_from_code(&dims, &_code_, theList);
      %let _CompleteList_ = &_CompleteList_ &theList;
   %end;

   %let &CompleteList=&_CompleteList_;
%mend _orva_list_all;

%macro _orva_reverse_list(list,ReverseList );
   %let nItems = 1;
   %let item = %scan(&list, &nItems, ' ');
   %let _reverseList = &item;

   %do %while ("&item" ne "" );  
      %let nItems=%eval(&nItems+1);
      %let item=%scan(&list, &nItems, ' ');
      %let _reverseList = &item &_reverseList;
   %end;

   %let &reverseList=&_reverseList;
%mend;

/* this macro takes the inString (with dilimiter: indelimiter), prefixes
 * each item by prefix and separated by outdelimiter and return the new
 * string.
 */
%macro _orva_createList(inString, indelimiter, outdilimiter, prefix);
    %local numcell cell colList;
	%let numCell = 1;
	%let cell=%qscan(&inString, &numCell, &indelimiter);
	%let colList=;
	%if &cell ne %str() %then %do;
	    %let colList=&prefix.%left(%trim(&cell));
	%end;
	%do %while(&cell ne %str());
	    %let numCell=%eval(&numCell+1);
	    %let cell=%qscan(&inString, &numCell, &indelimiter);
	    %if &cell ne %str() %then %do;
	        %let colList=&colList.&outdilimiter.&prefix.%left(%trim(&cell));
	    %end;
	%end;
	&colList
%mend _orva_createList;

%macro _orva_create_identity_matrix( ccLibref, outcov, numccvs);


proc sort data=&ccLibref.cc out = _CC_sorted_ nodupkey;
			by ccuid;
run;

data _cc_sorted_; set _cc_sorted_;
    format cell $15.;
    cell = compress("Cell_"||put(ccuid,&ORVAR_CELLFMT));

run;

proc transpose data=_CC_sorted_(drop=ccvUID_0 ccvUID_1 %if &numccvs > 2 %then ccvUID_2;) out=_ccmap_tmp_ ;
id cell;
run;

data &outcov; 
         length _NAME_ $20;
         set _ccmap_tmp_(drop=_name_) ;
         array y[*] cell:;

         do i =1 to dim(y);
            _NAME_ = trim(vname(y[i]));

            do j = 1 to dim(y);
               if ( j = i ) then y[j] = 1;
               else y[j] = 0;
            end;

            output;
          end;

           drop i j;
       run;


%mend _orva_create_identity_matrix;
