/******************************************************************************
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    orva_data_plot.sas                                             
 * PURPOSE: Generate data to plot histogram for data before any analysis            
 * SUPPORT: jariou
 *
 * PARAMETERS:
 *
 *   data       = The name of the loss data set to be analyzed.
 *   ccUid      = Key for this cell.
 *   ccData     = Current CC dataset.
 *   maxPoints  = The maximum number of points to use in plotting component
 *   plotData   = The name of the dataset containing the points to plot.
 *                Will contain the following variables (for plotting):
 *                AMOUNT LogAMOUNT
 *   plotStatus = Specifies the name of the dataset with extra information
 *
 ******************************************************************************/
%macro orva_data_plot(
                      data       = ,      
                      ccUid      = ,
                      ccData     = ,
                      maxPoints  = 200,
                      plotData   = ,
                      plotStatus = plotStatus   
                      );   

   /*- initialize macro variables -*/
   %let status = 0;
   %let reason = %sysfunc(sasmsg(&orvads, successful_note, noquote)) ;
   %let abort  = 0;

   %let Multiplier = 0;
   %let Count      = 0;  
   %let BinMin     = .;
   %let BinMax     = .;
   %let MaxLoss    = .;
  
   /*- create subsetting where for loss dataset -*/
   %let CELL_WHERE =  ;
   %let NUMCCVS    = 0;
   %orva_cell_where(
                     ccuid  = &ccuid,
                     ccData = &ccData
                     );
   %if &NUMCCVS eq 0 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, datp_cant_subset_cell_withds_error, noquote, &ccUid, &data));
      %let reason = %sysfunc(sasmsg(&orvads, datp_cant_subset_cell_withds_error, noquote, &ccUid, &data));
      %let abort  = 1;
      %goto DONE;
   %end;
   /*- initialize status dataset -*/
   %else %do;
      %let reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
      %orva_status_ds( 
                        statusds = &plotStatus, 
                        status = 1, 
                        reason = &reason 
                        ); 
      data &plotStatus; 
         set &plotStatus; 
         count = 0; 
      run;
   %end;

   /*- Sort the data -*/
   proc sort 
         data = &data 
         out  = _plotData_(keep=AMOUNT THRESHOLD)
         ;
      by AMOUNT;
      where &CELL_WHERE;
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, datp_cant_sort_loss_dat_error, noquote));
      %let reason = %sysfunc(sasmsg(&orvads, datp_cant_sort_loss_dat_error, noquote)) ;
      %let abort  = 1;
      %goto DONE;
   %end;

   %let dsid    = %sysfunc(open(_PlotData_, I));
   %let nLosses = %sysfunc(attrn(&dsid, nlobs));
   %let rc = %sysfunc(close(&dsid));

   %if &nLosses = 0 %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, datp_no_obs_in_data_error, noquote)) ;
      %let reason = %sysfunc(sasmsg(&orvads, datp_no_obs_in_data_error, noquote));
      %let abort = 1;
      %goto DONE;
   %end;

   data _PlotData_;
      set _PlotData_ nobs = obsCount;
      /* Sometimes, the THRESHOLD variable is always missing
      in these cases, BinMin would stay 1e16 so instead
      we make it be the minimum loss. */  
      retain minthreshold 1e16 minamount 1e16;
      minthreshold = min(minthreshold,THRESHOLD);
      minthreshold = min(minthreshold,AMOUNT);
      minLogThreshold = log( 1 + minthreshold);
      minamount = min(minamount,AMOUNT);
      minLogamount = log( 1 + minamount);

      /* Take log of (1+x) instead of x to avoid negative numbers */
      LogAMOUNT = 100 * log(1 + AMOUNT);
      LogLogAMOUNT = log(1 + LogAMOUNT);

      if _N_ = obsCount then do;
         call symputX("LossCount",  put(_N_, BEST20.));

         
         call symputX("BinMin",     put(minamount, BEST20.));
         call symputX("LogBinMin",  put(minLogamount, BEST20.));

         call symputX("MaxLoss",    put(AMOUNT, BEST20.));
         call symputX("LogMaxLoss", put(LogAMOUNT, BEST20.));

         call symputX("BinMax",     put(AMOUNT, BEST20.));
         call symputX("LogBinMax",  put(LogAMOUNT, BEST20.));
      end;
   run;


   proc summary 
               data   = _PlotData_     
               vardef = n 
               noprint
               ;
      var LogAMOUNT;
      output 
            out      = _stat_
            min      = min
            max      = max
            std      = std
            mean     = mean
            kurtosis = kurtosis
            ;
   run;

   proc summary 
               data   = _PlotData_     
               vardef = n 
               noprint
               ;
      var LogLogAMOUNT;
      output 
            out      = _log_stat_
            min      = min
            max      = max
            std      = std
            mean     = mean
            kurtosis = kurtosis
            ;
   run;

   data _null_;
      set _stat_;
      BinCount = log(%sysevalf(&nLosses)) / log(2) + 1;
      BinCount = BInCount + log(1 + (kurtosis + 3) * sqrt(%sysevalf(&nLosses / 6)))/ log(2);
      BinCount = max(1, log(&nLosses)/log(10) * BinCount);
      BinCount = min(200, BinCount);
      BinMax   = &BinMin + 20 * (&MaxLoss - &BinMin) / BinCount;
      BinWidth = (&MaxLoss - &BinMin) / BinCount;
      call symputX( "BinWidth", put(BinWidth, BEST20.) );
      call symputX( "BinMax", put(BinMax, BEST20.) );
      call symputX( "BinCount", put(BinCount, BEST20. ));
   run;

   data _NULL_;
      set _log_stat_;
      LogBinCount = log(%sysevalf(&nLosses)) / log(2) + 1;
      LogBinCount = LogBinCount + log(1 + (kurtosis + 3) * sqrt(%sysevalf(&nLosses / 6)))/log(2);
      LogBinCount = max(1, log(&nLosses) / log(10) * LogBinCount);
      LogBinCount = min(200, LogBinCount);
      LogBinMax   = &LogBinMin + 20 * (&LogMaxLoss - &LogBinMin) / LogBinCount;
      LogBinWidth = (&LogMaxLoss - &LogBinMin) / LogBinCount;
      call symputX( "LogBinWidth", put(LogBinWidth, BEST20.) );
      call symputX( "LogBinMax", put(LogBinMax, BEST20.));
      call symputX( "LogBinCount", put(LogBinCount, BEST20.));
   run;



   %orv_binify( 
                  inData   = _PlotData_,
                  variable = AMOUNT, 
                  binWidth = &binWidth,
                  binCount = &binCount,
                  maxPoint = &maxLoss,
                  outData  = __temp__
                  );

   %orv_binify( 
                  inData   = _PlotData_,
                  variable = LOGAMOUNT, 
                  binWidth = &logBinWidth,
                  binCount = &logBinCount,
                  maxPoint = &logMaxLoss,
                  outData  = Log__temp__
                  );

   %orv_countify( 
                     points_data = _PlotData_, 
                     bin_data    = __temp__, 
                     outdata     = &PlotData, 
                     variable    = AMOUNT 
                     );
   %orv_countify( 
                     points_data = _PlotData_, 
                     bin_data    = Log__temp__, 
                     outdata     = Log&PlotData, 
                     variable    = LOGAMOUNT 
                     );

   data &PlotData;
      set &PlotData;
      Count = Count / (UB - LB) / &lossCount;
   run;

   data Log&PlotData;
      set Log&PlotData;
      UB = UB/100;
      LB = LB/100;
      MidPoint = MidPoint / 100;
      RawCount = COunt;
      Count    = Count / (UB - LB) / &lossCount;
   run;

   data _NULL_;
      retain scale 0;
      set &PlotData end = last;

      scale = scale + (UB - LB) * Count;

      if _N_ = 1 then do;
         call symputX("BinMin", put(LB, BEST20.));
      end;

      if last then do;
         call symputX("Scale", put(scale, BEST20.));
         call symputX ("Count", _N_);
      end;
   run;

   data _NULL_;
      retain LogScale 0;
      set Log&PlotData end = last;

      LogScale = LogScale + (UB - LB) * Count;

      if _N_ = 1 then do;
         call symputX( "LogBinMin", put(LB, BEST20.));
      end;

      if last then do;
         call symputX( "LogScale", put(Logscale, BEST20.));
         call symputX ( "LogCount", _N_ );
      end;
   run;

%DONE:
   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, datp_abnormal_error, noquote));
      %let status = 1;
   %end;

   %orva_status_ds( 
                     statusds = &plotStatus, 
                     status   = &status, 
                     reason   = &reason 
                     ); 
   data &plotStatus; 
      set &plotStatus;
      Count    = &Count;
      LogCount = &LogCount;

      Multiplier = &Multiplier;

      BinMin    = &BinMin;
      LogBinMin = &LogBinMin;

      BinMax  = &BinMax;
      LossMax = &MaxLoss;
      Scale    = 1.0;
      LogScale = 1.0;
   run;

   /*- delete unwanted data sets -*/
   %orvu_delete_data ( data =
                              _PlotData_
                              _stat_ 
                              _log_stat_
                              Log__temp__
                              __temp__
                              );
%mend orva_Data_Plot;

%macro orv_binify(
                  inData   = ,
                  variable = , 
                  binCount = , 
                  binWidth = ,
                  maxPoint = , 
                  outData  =
                  );
   data &outData;
      %if &binCount > 1 %then %do;
         retain MinBin 0 LB 0 UB 0;
         LogWidth = int(log10(&binWidth)) - 1;

         LB = int(MinBin / 10**LogWidth) * 10**LogWidth;
         UB = int((LB + &binWidth) / 10**LogWidth) * 10**LogWidth;
         Midpoint = (UB + LB)/2;
         output;

         do while( UB LT &maxPoint);
            LB = UB;
            UB = int((LB + &binWidth) / 10**LogWidth) * 10**LogWidth;
            Midpoint = (UB + LB) / 2;
            output;
         end;
      %end;
      %else %do;
         MinBin = 0;
         UB = %sysevalf(&maxpoint + 0.5);
         LB = %sysevalf(&maxpoint - 0.5);
         Midpoint = &maxpoint;
      %end;
   run;
%mend orv_binify;

%macro orv_countify(
                     points_data = , 
                     bin_data    = , 
                     outdata     = , 
                     variable    = 
                     );
   proc sql noprint;
      SELECT 
               left(put(count(*), Best20.)) 
      INTO 
               :PointsCount
      FROM 
               &points_data
               ;
   quit;

   data &bin_data;
      set &bin_data end=last;
      count = 0;

      if last then do;
         call symput("NBINS",_N_);
      end;
   run;


   proc fcmp;
      array points[1] / NOSYMBOLS;
      array output[1] / NOSYMBOLS; /*   lb, ub, midpoint, count */

      rc = read_array("&bin_data", output, "LB", "UB", "COUNT",  "MIDPOINT");
      rc = read_array("&points_data", points, "&variable");
 
      j = 1;
      outputj2 = output[j, 2];
      do i = 1 to &PointsCount;
         
         if outputj2 GE Points[i] then do;
             output[j, 3] + 1;
         end;
         else do;  /* find the next bucket */
             j+1;
            do while (output[j, 2] LT Points[i] and j LT &NBINS ); 
                j +1;
            end;
            output[j, 3] + 1;
            outputj2 = output[j, 2];   
         end;
      end;

      rc = write_array("&outdata", output, "LB", "UB", "COUNT", "MIDPOINT");
   run;
   

   data &outdata;
      set &outdata;
      if count EQ 0 then delete;
      keep LB UB Count Midpoint;
   run;
%mend orv_countify;
