/*****************************************************************************************
 *
 * Copyright (C), 2005
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    ORVA_ffit                                      
 *       
 * PURPOSE: Wrapper for orva_freq_selection and orva_freq_fit_kri macros that gets all
 *          necessary fitting options out of the fitoptsData or initData
 *          for frequency fitting and then appends new fits to outparm dataset.
 *
 * SUPPORT: stchri
 *
 * PARAMETERS:
 *     countdata        - count data from orva_freq_counts
 *     fitoptsData      - fitting options (NA if 'initData' is passed)
 *     initData         - fparm dataset to use as starting point
 *                        for fitting during incremental load
 *     iteration        - iteration to refit (only used with initData)
 *     instrataprops    - dataset containing stratprops used in orva_freq_make_counts (I think)
 *     useCellOverrides - use cell overrides (vs project options)
 *     cTool            - are we running Consortium Tool?
 *     interval         - project interval
 *     pobservInfo         - sparm dataset containing severity fit for probability of observability
 *     sevIteration     - iteration id for sev fit this goes with
 *                        (NOT related to PObservInfo)
 *     sevIterSelected  - is sevIteration selected?
 *     datatype         - datatype I(nternal), (C)onsortium, (E)xternal
 *     ccUid            - ccUid for cell to fit
 *     ccData           - Cross Classification dataset
 *     CCV_NAMES        - namse of current cross class variables
 *     parallel         - parallel processing?
 *     outparm          - output fparm dataset to write results to
 *     outBluidEst      - output Dummy Estimates dataset (one dimension of dummies per newprop
 *     outCov           - output Covariance Matrix dataset
 *     outModels        - output Frequency Models dataset. One obs per stratum.  The sum of these
 *                        is the total mean for the cell.
 *     scaleData        - scaling dataset (currently only for internal data)
 *     STANDARDIZE      - standardize the KRIs to mean 0, std 1?
 *
 * NOTES:  None
 *****************************************************************************************/
%macro ORVA_ffit(   countdata        = ,
                    lossdata         = ,
                    fitoptsData      = ,  
                    initData         = ,
                    iteration        = , 
                    instrataprops    = ,
                    useCellOverrides = true,  
                    cTool            = true,
                    interval         = YEAR,   
                    pobservInfo      = ,
                    sevIteration     = ,
                    sevIterSelected  = ,
                    datatype         = , 
                    ccUid            = , 
                    ccData           = ,   
                    CCV_NAMES        = , 
                    parallel         = 0,
                    outParm          = ,
                    outBluidEst      = ,
                    outCov           = ,
                    outModels        = ,
                    scaleData        = ,
                    STANDARDIZE      = true,
                    debug            = false
                );                     

   %if &debug EQ true %then %do;
      %let _ffit_cum_timer_=0;
      %let _ffit_timer_=0;
      %orvu_timer(op=RESET, tvar=_ffit_cum_timer_);
      %orvu_timer(op=RESET, tvar=_ffit_timer_);
   %end;

   /***** clear out out temporary output datasets *****/
   data _ffit_parm;  set _NULL_; run;
   data _ffit_cov;   set _NULL_; run;
   data _ffit_blest; set _NULL_; run;

   /***** initialize macro variables *****/
   %let abort                      = 0;
   %let reason                     = ;
   %let fdists                     = ;
   %let sdist                      = ;
   %let ndist                      = 0;
   %let CELL_WHERE                 = 1;
   %let NUMCCVS                    = 0;
   %let pobsreason                 = ;
   %let pobsabort                  = ;
   %let _sel_status                = 0;
   %let _sel_reason                = ;
   %let STRATAPROPS                = ;
   %let STANDARDIZE                = false;
   %let NAMESOFKRIS                = ;
   %let ccuidForOptions            = .;
   %let fdists                     = POIS NEGBIN;
   %let Freq_FitMeasures           = LOGL 1 SBC 0 AIC 0;
   %let Freq_FavFitMeasure         = LOGL;

   %if &dataType EQ I %then %do;
      %let Freq_Pobserv_Source = Internal;
   %end;
   %else %if &dataType EQ C %then %do;
      %let Freq_Pobserv_Source = Consortium;
   %end;
   %else %if &dataType EQ P %then %do;
      %let Freq_Pobserv_Source = Public;
   %end;
   %else %do;
      %let Freq_Pobserv_Source =;
   %end;

   %let Freq_Use_Data_Threshold      = ;
   %let Freq_Missing_Default         = ;
   %let Freq_Zero_Count_Default      = ;
   %let Freq_Use_Stratum_Mean        = ;
   %let Freq_Cut_Off=;
   %let Freq_Keep_Larger_Thresholds  = ;
   %let Freq_Use_Kris                = false;
   %let Selected_Kris                = ;
   %let Freq_Kris_Suggested          = ;
   %let Freq_Kris_Forced             = ;
   %let Freq_Selection_Method_Kris   = Stepwise;
   %let Freq_Use_Indicators          = true;
   %let Freq_Indicator_Properties    = internal_bl_uid;
   %let Freq_Model_Basis_Props       = internal_bl_uid;
   %let Freq_Use_Data_Threshold      = true;
   %let Freq_Missing_Default         = 0;
   %let Freq_Cut_Off                 = 0;
   %let Freq_Keep_Larger_Thresholds  = false;
   %let Freq_UseStratumMean_Default  = false;
   %let Freq_UseStratumMean_UserSpec = false;
   %let Freq_Zero_Count_Default      = 0;
   %let fitOptionsRetVal             = 0;


   /***** GET FROM INITDATA (if incremental fit)OR FITOPTS DATASET *****/
/* %_handle_fit_options( fitOptionsRetValM = fitOptionsRetVal);*/
   %_handle_fit_options(fitOptionsRetValM             = fitOptionsRetVal,
                        initDataM                     = &initData,
                        fitOptsDataM                  = &fitOptsData,
                        inStrataPropsM                = &inStrataProps,
                        STRATAPROPSM                  = STRATAPROPS,
                        abortM                        = abort,
                        reasonM                       = reason,
                        ccuidM                        = &ccuid,
                        ccuidForOptionsM              = ccuidForOptions,
                        useCellOverridesM             = &useCellOverrides, 
                        iterationM                    = &iteration,
                        dataTypeM                     = &dataType,
                        Freq_Use_KrisM                = Freq_Use_Kris,
                        Freq_Kris_SuggestedM          = Freq_Kris_Suggested,
                        Freq_Kris_ForcedM             = Freq_Kris_Forced,
                        Freq_Selection_Method_KrisM   = Freq_Selection_Method_Kris,
                        Freq_Use_IndicatorsM          = Freq_Use_Indicators,
                        Freq_Indicator_PropertiesM    = Freq_Indicator_Properties,
                        Freq_Model_Basis_PropsM       = Freq_Model_Basis_Props,
                        Freq_Pobserv_SourceM          = Freq_Pobserv_Source,
                        Freq_Use_Data_ThresholdM      = Freq_Use_Data_Threshold,
                        Freq_Missing_DefaultM         = Freq_Missing_Default,
                        Freq_Cut_OffM                 = Freq_Cut_Off,
                        Freq_Keep_Larger_ThresholdsM  = Freq_Keep_Larger_Thresholds,
                        Freq_UseStratumMean_DefaultM  = Freq_UseStratumMean_Default,
                        Freq_UseStratumMean_UserSpecM = Freq_UseStratumMean_UserSpec,
                        Freq_Zero_Count_DefaultM      = Freq_Zero_Count_Default,
                        fdistsM                       = fdists
                        );
   %if &debug EQ true %then %do;
      %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Getting options);
   %end;
   %if &fitOptionsRetVal=1 %then %do;
      %goto FFIT_DONE;
         %end;



   /***** Count Data Set Preparation. *****/
   %_count_data_set_prep( _ffit_countsM = _ffit_counts,
                           _ffit_lossesM = _ffit_losses
   );



    /***** Fill in Thresholds according to requested options  *****/
    %orva_freq_thresholds(  countdata            = _ffit_counts, 
                            instrataprops        = &STRATAPROPS, 
                            outcountdata         = _ffit_counts, 
                            outStatus            = _ft_status,
                            debug                = &debug
   ); 
     data _NULL_;
       set _ft_status;
       call symput('_ft_reason', reason);
       call symputx('_ft_abort', status);
     run;
     %if &_ft_abort %then %do;
       %let reason= &_ft_reason;
      %let abort= 0; /* End gracefully; display error messages to GUI */
       %goto FFIT_DONE;
     %end;

     %if &debug EQ true %then %do;
        %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Getting thresholds);
     %end;



   /***** Fill in p_obs using severity dist fitted to this cell *****/
    %let _fgp_returnVal = 1; /*  expect failure */
    %orva_freq_getpobs( countdata     = _ffit_counts, 
                        lossdata      = _ffit_losses, 
                        sevinfo       = &pobservInfo,
                        datatype      = &dataType, 
                        ccuid         = &ccUid, 
                        instrataprops = &STRATAPROPS,
                        outcountdata  = _ffit_counts, 
                        returnVal     = _fgp_returnVal,
                        pobsoutstatus = _ffit_pobsoutstatus,
                        debug         = &debug
   );
   %if &_fgp_returnVal GT 0 %then %do;
      data _NULL_;
         set _ffit_pobsoutstatus;
         call symput('reason', reason);
         call symput('pobsreason', reason);
         call symput('pobsabort', status);
      run;
   %end;
   %if &_fgp_returnVal EQ 3 %then %do;
      %let abort= &pobsabort;
      %goto FFIT_DONE;
   %end;
   %else %do;
      data _NULL_;
         set _ffit_pobsoutstatus;
         call symputX('avgCOUNTwtdpobs', put(avgCOUNTwtdpobs,BEST20.));
         call symputX('Count_Yellowflagged_Strata', put(Count_Yellowflagged_Strata,BEST20.));
      run;
   %end;

    %if &debug EQ true %then %do;
      %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Getting POBS);
    %end;
  


   /***** Combine counts/exposures/p_obs to get a coarser level of granularity (fewer strata) *****/
%if &dataType = I %then %do;
   %if %upcase(&cTool) EQ FALSE %then %do;
      %if %upcase(&Freq_Model_Basis_Props) EQ INTERNAL_BL_UID %then %do;
         %let Freq_Use_Indicators = true;
         %let Freq_Indicator_Properties = &Freq_Model_Basis_Props;  
         /* temporary */
/*       %if Freq_Use_Indicators = false %then %do;*/
         /*%let Freq_Indicator_Properties = ; */
/*       %end;*/
/*       %else %do;*/
/*          %let Freq_Indicator_Properties = &Freq_Model_Basis_Props; */
/*       %end;*/
      %end;
      %if %upcase(&Freq_Model_Basis_Props) EQ LEGAL_ENTITY_UID %then %do;
         %let Freq_Model_Basis_Props = INTERNAL_BL_UID; /* "make dummies on" will equal freq model basis props, for now */
/*       %let Freq_Use_Indicators = true;*/
/*       %let Freq_Indicator_Properties = &Freq_Model_Basis_Props;  */
         /* temporary */
         %let Freq_Use_Indicators = false;
         %let Freq_Indicator_Properties = ;  /* temporary */
      %end;
      %else %if %upcase(&Freq_Model_Basis_Props) EQ TIME_ONLY %then %do;
         %let Freq_Model_Basis_Props = ; /* "make dummies on" will equal freq model basis props, for now */
         %let Freq_Use_Indicators = false;
         %let Freq_Indicator_Properties = &Freq_Model_Basis_Props;  /* temporary */
      %end;
   %end;
   %else %do;
      %let Freq_Model_Basis_Props = INTERNAL_BL_UID; /* "make dummies on" will equal freq model basis props, for now */
      %let Freq_Use_Indicators = true;
      %let Freq_Indicator_Properties = &Freq_Model_Basis_Props;
   %end;
%end;
%else %do;  /* ALL other dataTypes (besides Internal) are not allowed to use Dummy Variables */
   %let Freq_Model_Basis_Props = ; /* "make dummies on" will equal freq model basis props, for now */
   %let Freq_Use_Indicators = false;
   %let Freq_Indicator_Properties = ;
%end;



   %orva_freq_combine_counts( countdata = _ffit_counts,
         instrataprops = &STRATAPROPS,
                               Model_Basis_Props = &Freq_Model_Basis_Props,
         cTool = &cTool,
         dataType = &dataType,
         outCounts = _ffit_countsCombined,
         outnewprops = newpropsds,
                               outStatus = outfreqcombinecountsstatus,
                               debug = &debug
      ); 
   /* NEWPROPS are simply the Freq_Model_Basis_Props. */
   %let NEWPROPS=;
   %let nobs=0;
   %orvu_getObsCount(newpropsds,nobs);
   %if &nobs GT 0 %then %do;
      data _NULL_;
         set newpropsds;
         call symput("NEWPROPS", trim(THENEWPROPS));
      run;
   %end;
   
   %if &debug EQ true %then %do;
      %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Getting combined counts);
   %end;



   /***** Standardize KRIs, if &STANDARDIZE = true. *****/     
   %_standardize_kris( indata = _ffit_countsCombined,
                        outdatacounts = _ffit_counts3,
                        outdatastats = _ffit_stats
   );
      


    /***** Variable Selection                                              *****/
    %orva_freq_selection( data                = _ffit_counts3,
                          selection           = &Freq_Selection_Method_Kris,
                          NUMCCVS             = &NUMCCVS,
                          CCV_NAMES           = &CCV_NAMES,
                          interval            = &interval,
                          ccUid               = &ccUid,
                          Model_Basis_Props   = &NEWPROPS,
                          makeDummiesOn       = &Freq_Indicator_Properties,
                          outCompressedData   = _sel_compressed_data,
                          outParms            = _sel_outparms,
                    outSelDummies       = _sel_outSelDummies,
                    outSelKRIs          = _sel_outSelKRIs,
                          outStatus           = _sel_status,
                          outNotSelected      = _sel_not_selected, 
                          outDummyMap         = _sel_dummyMap,
                          outDroppedDummies   = _sel_droppedDummies,
                          debug               = &debug,
                          dataType            = &datatype
                          %if &scaleData NE %str() %then %do;
                             ,scaleData     =&scaleData
                          %end; 
                          );       
    %if &debug EQ true %then %do;
       %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Freq selection);
    %end;
   data _NULL_;
      set _sel_status;
      call symput('_sel_reason', reason);
      call symput('_sel_status', status);
   run;
   %if &_sel_status %then %do;
      %let reason= &_sel_reason;
      %let status= &_sel_status;
      %goto FFIT_DONE;
   %end;



    /***** Fitting                                                         *****/
    %orva_freq_fit_kri( compressedData   = _sel_compressed_data,
                        notSelected      = _sel_not_selected,
                  selDummies       = _sel_outSelDummies,
                  selKRIs          = _sel_outSelKRIs,
                        parmStatus       = _sel_outparms,
                        droppedDummies   = _sel_droppedDummies,
                        dummyMap         = _sel_dummyMap,
                  useIndicators    = &Freq_Use_Indicators,
                        ffitStats        = _ffit_stats,
                        standardize      = &STANDARDIZE,
                        instrataprops    = &NEWPROPS,
                        ccUid            = &ccUid,
                        NUMCCVS          = &NUMCCVS,
                        useCellOverrides = &useCellOverrides, 
                        interval         = &interval, 
                        avgCOUNTwtdpobs  = &avgCOUNTwtdpobs, 
                        outParm          = _ffit_parm,
                        outBluidEst      = _ffit_blest,
                        outCov           = _ffit_cov, 
                        freqModels       = _ffit_freqModels,
                        parallel         = &parallel,
                        dataType         = &dataType,
                        debug            = &debug
   ); 
    %if &debug EQ true %then %do;
       %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Actual fitting);
    %end;

   /***** SAVE FIT OPTIONS TO OUTPUT DATASET *****/
   data _ffit_parm;
      %_format_freq_fit_options;
      set _ffit_parm;
      %_save_freq_fit_options;
   run;

/***** Finish *****/
%FFIT_DONE:
   %if &abort EQ 1 %then %do;  /* e.g.  instrataprops was empty */
      %put &orvar_error %sysfunc(sasmsg(&orvads, ffit_func_failed_bc_error, noquote, &reason));
   %end;   
   %else %do;
      %if &_ft_abort or &pobsabort EQ 3 or &_sel_status EQ 1 %then %do;   
         /*** Either mean p_obs is < 0.1 OR instrataprops to p_obs macro was empty; 
            or:  in selection:  too many params;
            Freq Fitting WAS NOT PERFORMED ***/
         data _ffit_parm;
            format reason &ORVAR_REASON_FMT distribution $12.;
            %_format_freq_fit_options;
            %let ndist=1;
            %let dist=%scan(&fdists, &ndist, ' ');
            %do %while (&dist NE %str() );
               distribution = "&dist";   
               %if &pobsabort EQ 3 %then %do; reason = "&pobsreason"; %end;
               %else %do;  reason = "&reason"; %end;
               status = 1;
               LOGL = SBC = AIC = .;
               %_save_freq_fit_options;
               output;
               %let ndist=%eval(&ndist+1);
               %let dist=%scan(&fdists, &ndist, ' ');
            %end;
         run;

         data _ffit_cov; 
            set _NULL_; 
         run;

         data _ffit_blest; 
            set _NULL_; 
         run;
      %end;
      %else %if &pobsabort EQ 2 %then %do;  
      /*** one or more strata mean p_obs were less than 0.1; Freq Fit WAS PERFORMED ***/
         data _ffit_parm;
            set _ffit_parm;
            reason = "&pobsreason";
            Count_Yellowflagged_Strata = &Count_Yellowflagged_Strata;
            status = &pobsabort;
         run;
      %end;

       /*** get fit measure options for ranking  ***/
       %if &fitoptsData NE %str() %then %do;
          %orva_readfitopts(  fitoptsdata = &fitoptsData,
                              type        = FITMEASURE,
                              ccUid       = &ccuidForOptions,
                              fittype     = FREQ,
                              datatype    = &datatype
                              );
       %end;
   
       /*** append new fits to output datasets  ***/
       %orva_append_parm_dataset(ccUid           = &ccuid,
                                 sevIteration    = &sevIteration,
                                 sevIterSelected = &sevIterSelected,
                                 measCol         = &Freq_FitMeasures,
                                 selectedMeas    = &Freq_FavFitMeasure,
                                 tmpParm         = _ffit_parm,
                                 finalParm       = &outparm,
                                 tmpCov          = _ffit_cov,
                                 finalCov        = &outcov,
                                 tmpBlEst        = _ffit_blest, 
                                 finalBlEst      = &outBluidEst,
                                 tmpFModels      = _ffit_freqModels, 
                                 finalFModels    = &outModels
                                  %if &iteration NE %str() %then %do;
                                    , 
                                    iteration    = %eval(&iteration - 1)
                                  %end;
                                 );
   %end; 

   %if &debug EQ true %then %do;
      %orvu_timer(op=MARK, tvar=_ffit_cum_timer_, msg=ORVA_FFIT - total macro);
   %end;

   /*- delete unwanted data sets -*/
   %if &debug NE keepdatasets %then %do; 
      proc datasets library=work memtype=data nolist nodetails;
         delete  
               _ffit_: 
               _sel_:
               ;
      quit; 
   %end;
%mend orva_ffit;

%macro _handle_fit_options(fitOptionsRetValM             = ,
                           initDataM                     = ,
                           fitOptsDataM                  = ,
                           inStrataPropsM                = ,
                           STRATAPROPSM                  = ,
                           abortM                        = ,
                           reasonM                       = ,
                           ccuidM                        = ,
                           ccuidForOptionsM              = ,
                           usecelloverridesM             = ,
                           iterationM                    = ,
                           dataTypeM                     = ,
                           Freq_Use_KrisM                = ,
                           Freq_Kris_SuggestedM          = ,
                           Freq_Kris_ForcedM             = ,
                           Freq_Selection_Method_KrisM   = ,
                           Freq_Use_IndicatorsM          = ,
                           Freq_Indicator_PropertiesM    = ,
                           Freq_Model_Basis_PropsM       = ,
                           Freq_Pobserv_SourceM          = ,
                           Freq_Use_Data_ThresholdM      = ,
                           Freq_Missing_DefaultM         = ,
                           Freq_Cut_OffM                 = ,
                           Freq_Keep_Larger_ThresholdsM  = ,
                           Freq_UseStratumMean_DefaultM  = ,
                           Freq_UseStratumMean_UserSpecM = ,
                           Freq_Zero_Count_DefaultM      = ,
                           fdistsM                       = , 
                           SelectedDummiesM              = ,
                           numSelectedDummiesM           =
                           );
    /***** extract the STRATAPROPS *****/
   %let _nobs=0;
   %if &instratapropsM NE %str() %then %do;
       %orvu_getObsCount(&instratapropsM,_nobs);
   %end; 
   %if &_nobs GT 0 %then %do;
      data _NULL_;
         set &instratapropsM;
         call symput("&STRATAPROPSM", trim(THESTRATAPROPS));
      run;
   %end;
   %else %do;
      %let &abortM=1;
      %let &reasonM=%sysfunc(sasmsg(&orvads, ft_isp_ds_empty_warning, noquote));
      %let &fitOptionsRetValM = 1;
   %end;

   /***** For incremental load, all options are in initData dataset *****/
   %if &initDataM NE %str() %then %do;
      data _ffit_fparmds; set &initDataM; 
         where CCUID = &CCUIDM and ITERATION=&iterationM;

         /*- get KRI and Indicator Info -*/
         %if &datatypeM EQ I %then %do;
/*          if ^missing(SelectedKRIs) then do;*/
/*             call symput("&SelectedKRIsM", strip(SelectedKRIs));*/
/*          end;*/
            if ^missing(Freq_Use_Kris) then do;
               call symput("&Freq_Use_KrisM", strip(Freq_Use_Kris));
            end;
            if ^missing(Freq_Kris_Suggested) then do;
               call symput("&Freq_Kris_SuggestedM", strip(Freq_Kris_Suggested));
            end;
            if ^missing(Freq_Kris_Forced)then do;
               call symput("&Freq_Kris_ForcedM", strip(Freq_Kris_Forced));
            end;
            if ^missing(Freq_Selection_Method_Kris) then do;
               call symput("&Freq_Selection_Method_KrisM",strip(Freq_Selection_Method_Kris)); 
            end;
            if ^missing(Freq_Use_Indicators) then do;
               call symput("&Freq_Use_IndicatorsM",strip(Freq_Use_Indicators)); 
            end;
            if ^missing(Freq_Indicator_Properties) then do;
               call symput("&Freq_Indicator_PropertiesM",strip(Freq_Indicator_Properties)); 
            end;
            if ^missing(Freq_Model_Basis_Props) then do;
               call symput("&Freq_Model_Basis_PropsM", strip(Freq_Model_Basis_Props));
            end;
            else do;
               call symput("&Freq_Model_Basis_PropsM", "INTERNAL_BL_UID");
            end;
         %end;
         %else %do;
            call symput("&Freq_Model_Basis_PropsM", "");
         %end;

         /*- get Threshold options -*/
         if ^missing(Freq_Pobserv_Source) then do;
            call symput("&Freq_Pobserv_SourceM", Freq_Pobserv_Source);
         end;
         if ^missing(Freq_Use_Data_Threshold) then do;
            call symput("&Freq_Use_Data_ThresholdM", Freq_Use_Data_Threshold);
         end;
         if ^missing(Freq_Missing_Default) then do;
            call symput("&Freq_Missing_DefaultM", Freq_Missing_Default);
         end;
         if ^missing(Freq_Cut_Off) then do;
            call symput("&Freq_Cut_OffM", Freq_Cut_Off);
         end;
         if ^missing(Freq_Keep_Larger_Thresholds) then do;
            call symput("&Freq_Keep_Larger_ThresholdsM", Freq_Keep_Larger_Thresholds);
         end;
         if ^missing(Freq_UseStratumMean_Default) then do;
            call symput("&Freq_UseStratumMean_DefaultM", Freq_UseStratumMean_Default);
         end;
         if ^missing(Freq_UseStratumMean_UserSpec) then do;
            call symput("&Freq_UseStratumMean_UserSpecM", Freq_UseStratumMean_UserSpec);
         end;
         if ^missing(Freq_Zero_Count_Default) then do;
            call symput("&Freq_Zero_Count_DefaultM", Freq_Zero_Count_Default);
         end;

         /*- get other stuff -*/      
         call symput("&fdistsM", distribution); 
      run;

      /* get selected dummies */
/*    %let &SelectedDummiesM=;*/
/*    %let &numSelectedDummiesM=;*/
/*    %orva_get_selected_dummies(fparmds = _ffit_fparmds);*/
   %end; /* End of Incremental Fit options area */  

   /***** Otherwise, get options from fitopts dataset *****/
   %else %if &fitOptsDataM NE %str() %then %do;
      %if &useCellOverridesM EQ false %then %let &ccuidForOptionsM=.;
      %else %let &ccuidForOptionsM=&ccUidM;

      /*- get Distribution options -*/
      %let &fdistsM=;
      %orva_readfitopts( fitoptsdata = &fitoptsDataM,
                           type = DIST,
                           fittype = FREQ,
                           ccUid = &&&ccuidForOptionsM,
                           datatype = &datatypeM
      );
      %if &fdistsM EQ %str() %then %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, ffit_no_dist_sel_for_fit_warning, noquote));
         %let reason=%sysfunc(sasmsg(&orvads, ffit_no_dist_sel_for_fit_warning, noquote));
         %let abort=1;
         %let &fitOptionsRetValM = 1;
      %end;  

      /*- get KRI Control Options -*/
      %let &Freq_Use_KrisM=false;
      %let &Freq_Selection_Method_KrisM=Stepwise;
      %if &datatypeM EQ I %then %do;
         %orva_readfitopts(fitoptsdata = &fitoptsDataM,
                           type        = KRICONTROL,
                           fittype     = FREQ,
                           ccUid       = &&&ccuidForOptionsM,
                           datatype    = &datatypeM
                           );
      %end;

      /*- get Indicator Control Options -*/
      %if &datatypeM EQ I %then %do;
         %orva_readfitopts(fitoptsdata = &fitoptsDataM,
                           type        = INDICATOR,
                           fittype     = FREQ,
                           ccUid       = &&&ccuidForOptionsM,
                           datatype    = &datatypeM
                           );
      %end;

      /*- get POBSERV options -*/
      %orva_readfitopts(fitoptsData = &fitoptsDataM,
                        fittype     = FREQ,
                        type        = POBSERV,
                        datatype    = &datatypeM,
                        ccUid       = &&&ccuidForOptionsM
                        );
      /*- get Threshold options -*/
      %orva_readfitopts(fitoptsData = &fitoptsDataM, 
                        fittype     = FREQ, 
                        type        = THRESHOLD, 
                        datatype    = &datatypeM, 
                        ccuid       = &&&ccuidForOptionsM
                        );
      /*- get KRIs -*/
      %let &Freq_Kris_SuggestedM=;
      %let &Freq_Kris_ForcedM=;
      %if &&&Freq_Use_KrisM EQ true %then %do;    

         %orva_readfitopts(fitoptsdata = &fitoptsDataM,
                           type        = KRI,
                           fittype     = FREQ,
                           ccUid       = &&&ccuidForOptionsM,
                           datatype    = &datatypeM
                           );
         %if &&&Freq_Kris_SuggestedM EQ %str() AND &&&Freq_Kris_ForcedM EQ %str() %then %do;
            %let &Freq_Use_KrisM=false;
         %end;       
      %end;
   %end;
%mend _handle_fit_options;

/**********************************************************************************/
/*  (Recall: The count dataset &countdata was made by orva_freq_make_counts.)     */
/*    1 - subset to get only the count data we need                               */
/*    2 - Add obs for ibls mapped to this cell, but with no losses in this cell   */
/*    3 - Set up macro variable NAMESOFKRIS (if &Freq_use_Kris = true)            */
/*    4 - Add 6 columns of threshold options to count data set.                   */
/**********************************************************************************/
%macro _count_data_set_prep( _ffit_countsM = , _ffit_lossesM =  );
    /*--- sets macro varibles CELL_WHERE and NUMCCVS ---*/
   %orva_cell_where(ccuid=&ccUid, ccData=&ccData);    
   data &_ffit_countsM(compress=binary);
      set &countdata;
      where &CELL_WHERE;
   run;
   data &_ffit_lossesM (compress=binary);
      set &lossdata;
      where &CELL_WHERE;
   run;
   /* NEXT, add to _ffit_countsM any obs of &countdata with "this" value of CCVUID_x (the dim for ibl) 
     but with no losses in this cell. */
   %if &dataType EQ I %then %do;
      %let ibl_dimension = ;
      %let found=0;
      %do lj=0 %to %eval(&NUMCCVS-1);
         %if %upcase(%scan(&CCV_NAMES, %eval(&lj+1), ' ')) EQ INTERNAL_BL_UID %then %do;
            %let found=1;
            %let ibl_dimension = &lj;
         %end;
      %end;
      %if &found EQ 1 %then %do;
         data _NULL_;
            set &ccData;
            where ccuid=&ccUid;
            call symputX("ibl", CCVUID_&ibl_dimension);
         run;
         /* get ibls with no losses EVER */
/*       data _ffit_extracounts1;*/
/*          set &countdata;*/
/*          where CCVUID_&ibl_dimension EQ &ibl AND missing(CCUID);*/
/*       run;  */

         /* get ibls with losses elsewhere, but no losses in this cell */
         /* get ibls with no losses EVER */
         %_ibls_without_losses_here(ccuid     = &ccUid, 
                                    ccData    = &ccData, 
                                    blcol     = CCVUID_&ibl_dimension, 
                                    blvalue   = &ibl, 
                                    countdata = &countdata,
                                    cTool     = &cTool, 
                                    outdata   = _ffit_extracounts2,
                                    debug     = &debug
                                    );

         data &_ffit_countsM (compress=binary);
            set &_ffit_countsM           /* ibls with losses IN THIS CELL */
              /*_ffit_extracounts1*/   /* ibls with no losses EVER */
                _ffit_extracounts2;    /* ibls without losses here */
            drop adj_length start_of_curr_fitinterval end_of_curr_fitinterval;
         run;

      %end;
      %else %do;  /* IBL is not a CCV.  */
         %_ibls_without_losses_here(ccuid     = &ccUid, 
                                    ccData    = &ccData, 
                                    blcol     = , 
                                    blvalue   = , 
                                    countdata = &countdata,
                                    cTool     = &cTool, 
                                    outdata   = _ffit_extracounts2,
                                    debug     = &debug
                                    );
         data &_ffit_countsM (compress=binary);
            set &_ffit_countsM          /* ibls with losses IN THIS CELL */
            /*_ffit_extracounts1*/ /* ibls with no losses EVER */
            _ffit_extracounts2;    /* ibls without losses here */
            drop adj_length start_of_curr_fitinterval end_of_curr_fitinterval;
         run;
      %end;
   %end;

   data &_ffit_countsM;
      set &_ffit_countsM;
      format Freq_Missing_Default   Freq_Zero_Count_Default Freq_Cut_Off 10.5;
      length Freq_Use_Data_Threshold Freq_UseStratumMean_Default Freq_Keep_Larger_Thresholds $10;

      Freq_Use_Data_Threshold     = "&Freq_Use_Data_Threshold";
      Freq_Missing_Default        = &Freq_Missing_Default;
      Freq_Zero_Count_Default     = &Freq_Zero_Count_Default;
      Freq_UseStratumMean_Default = "&Freq_UseStratumMean_Default";
      Freq_Cut_Off                = &Freq_Cut_Off;
      Freq_Keep_Larger_Thresholds = "&Freq_Keep_Larger_Thresholds"; 

      array prop &STRATAPROPS;
      do i = 1 to dim(prop);
         if missing(prop[i]) then prop[i] = -1;
      end;
   run;  

   /*--- Set NAMESOFKRIS, if needed ---*/
   %if &Freq_Use_Kris = true %then %do; 
      %let STANDARDIZE = true;
   %end;
   %else %do;
      %let STANDARDIZE = false;
   %end;

   %if &STANDARDIZE = true %then %do;
      /*--- set up names of KRIs (macro variable NAMESOFKRIS) ---*/
      data _ffit_counts_justkris (compress=binary); /* just need this to get KRI names */
         set _ffit_counts(keep=KRI_:) ;
      run;

      proc contents noprint data=_ffit_counts_justkris out=_ffit_cont(keep=NAME);
      run;

      proc sql noprint;
         SELECT 
               NAME
         INTO 
               :NAMESOFKRIS separated by " "
         FROM 
               _ffit_cont
         ;        
      quit;
   %end;

   %if &debug EQ true %then %do;
      %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Getting counts);
   %end;
%mend _count_data_set_prep;

%macro _standardize_kris ( indata = ,
                           outdatacounts = ,
                           outdatastats = 
                           );
/* Must be AFTER p_obs calculation, because KRI_LOG (macro variable used to calculate p_obs) 
   needed to use the un-standardized values of the KRIs */
   %if &STANDARDIZE = true %then %do;
      /* Save original KRI values, w/ obs in same order as current _ffit_counts */
      data _ffit_counts_small ; 
         set &indata(keep=KRI_:) ;
      run;

      /* Standardize the KRIs into dataset _ffit_counts2 */    
      ods listing close;
      ods output Standard =  &outdatastats ; 
      proc standard data=&indata mean=0 std=1 print out=_ffit_counts2;
         var &NAMESOFKRIS ;
      run;
      ods listing; 
      proc datasets nolist nodetails; /* Modify names of unstandardized KRIs by adding prefix "un_" */
         MODIFY _ffit_counts_small;  /* Contains original KRI values */ 
         RENAME 
            %let nname=1;
             %let thename=%scan(&NAMESOFKRIS, &nname, ' ');
             %do %while (&thename NE %str() );
                 &thename = un_&thename  
                 %let nname=%eval(&nname+1);
                 %let thename=%scan(&NAMESOFKRIS, &nname, ' ');
             %end;
            %let nname=%eval(&nname-1);
         ;  
      quit;
      /* _ffit_counts3 will have the standarized KRIs (KRI_:) 
         and the unstandardized KRIs (un_KRI_:) */
      data &outdatacounts(compress=binary);
         set _ffit_counts2; set  _ffit_counts_small;
      run;
   %end;
   %else %do;
      data &outdatacounts(compress=binary);
         set &indata;
      run;
      data &outdatastats; set _NULL_; run;
   %end;
   %if &debug EQ true %then %do;
      %orvu_timer(op=MARK_AND_RESET, tvar=_ffit_timer_, msg=ORVA_FFIT - Standardize);
   %end;

%mend _standardize_kris;

%macro _format_freq_fit_options;
   length Freq_Kris_Suggested $ 10000;
   length Freq_Kris_Forced $ 10000;
   length Freq_Pobserv_Source $ 12;
   length Freq_Model_Basis_Props $ 256;
   length Freq_Use_Indicators  $ 32;
   length Freq_Indicator_Properties  $ 256;
   length Freq_Selection_Method_Kris $ 32;
%mend _format_freq_fit_options;

%macro _save_freq_fit_options;
   Freq_Pobserv_Source          = "&Freq_Pobserv_Source";
   Freq_Use_Kris                = "&Freq_Use_Kris";
   Freq_Kris_Suggested          = "&Freq_Kris_Suggested";
   Freq_Kris_Forced             = "&Freq_Kris_Forced";
   Freq_Selection_Method_Kris   = "&Freq_Selection_Method_Kris";
   Freq_Use_Data_Threshold      = "&Freq_Use_Data_Threshold";
   Freq_Keep_Larger_Thresholds  = "&Freq_Keep_Larger_Thresholds";
   Freq_UseStratumMean_Default  = "&Freq_UseStratumMean_Default";
   Freq_UseStratumMean_UserSpec = "&Freq_UseStratumMean_UserSpec";
   Freq_Fit_Interval            = "&interval";
   Freq_Use_Indicators          = "&Freq_Use_Indicators";
   Freq_Indicator_Properties    = "&Freq_Indicator_Properties";
   Freq_Missing_Default         = &Freq_Missing_Default;
   Freq_Cut_Off                 = &Freq_Cut_Off;
   Freq_Zero_Count_Default      = &Freq_Zero_Count_Default;

   %if &Freq_Model_Basis_Props EQ %str() %then %do;
      Freq_Model_Basis_Props    = "TIME_ONLY";
   %end;
   %else %do;
      Freq_Model_Basis_Props    = "&Freq_Model_Basis_Props";
   %end;
%mend _save_freq_fit_options;


%macro _ibls_without_losses_here(ccuid     = , 
                                 ccData    = , 
                                 blcol     = , 
                                 blvalue   = , 
                                 countdata = , 
                                 cTool     = ,
                                 outdata   = ,
                                 debug     = false
                                 );
   data _ibl_counts(keep=%if &cTool EQ true %then %do; OWN %end; 
             INTERNAL_BL_UID interval_start interval_end adj_length CCVUID_:
             KRI_: start_of_curr_fitinterval end_of_curr_fitinterval ); 
      set &countdata; 

      %if &blcol NE %str() and &blvalue NE 0 %then %do;
         where &blcol eq &blvalue; 
      %end;
   run;

   %let CELL_WHERE=;
   %orva_cell_where(ccuid=&ccuid, ccData=&ccData);

   data _ibl_countsInThisCell(keep=INTERNAL_BL_UID);
      set &countdata;
      where &CELL_WHERE;
   run;

   proc sort data=_ibl_countsInThisCell nodup;
      by INTERNAL_BL_UID;
   run;

   data _ibl_counts2(keep=%if &cTool EQ true %then %do; OWN %end;
                  INTERNAL_BL_UID interval_start interval_end adj_length &blcol KRI_: 
                  start_of_curr_fitinterval end_of_curr_fitinterval  ); 
      set _ibl_counts; 
      where (^(&CELL_WHERE)); 
   run;

   proc sort data=_ibl_counts2 nodup; 
      by INTERNAL_BL_UID INTERVAL_START; 
   run;

   data _ibl_counts4;
      merge _ibl_countsInThisCell(in=inThisCell) _ibl_counts2;
      by INTERNAL_BL_UID;
      if (inThisCell) then delete;
   run;

   data &outdata;
      set _ibl_counts4;
      the_count = 0;

      if end_of_curr_fitinterval GT    start_of_curr_fitinterval then do;
         exposure = adj_length/(end_of_curr_fitinterval - start_of_curr_fitinterval);
      end;
      else do; 
         exposure = .;  /* or zero??? */
      end;
   run;

   %if &debug NE keepdatasets %then %do; 
      proc datasets library=work memtype=data nolist nodetails;
         delete  _ibl_counts_:
         ;
      quit; 
   %end;
%mend _ibls_without_losses_here;
