/*****************************************************************************************
 * $Id: orva_freq_selection.sas,v 1.1.2.11.2.26.2.8.2.2.2.1 2013/01/30 14:50:04 stmajo Exp $
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    %orva_freq_selection.sas                                             
 *
 *  PURPOSE: Selects significant regressors to be used later in frequency fitting
 *           
 *  SUPPORT: jariou
 *
 *  PARAMETERS:
 *
 *  data  =  Count dataset outputted by orva_freq_make_counts, with the following variables: OWN, each CCV,
 *           THRESHOLD, pobs, interval_start, interval_end, adj_length, the_count, CCVUID_:, CCUID
 *           possibly KRI_:, possibly? internal_bl_uid (if there were any bl-level KRIs).
 *
 *  selection = the model-selection option to be used to choose the best set of KRIs (regressors).  
 *              This should be a string, provided by the user, and it may take on 
 *              one of the following 5 values: Forward, Backward, Stepwise.  
 *              Note that this is analogous to, but separate from, the parameter that 
 *              orva_sev_fit_kri gets for selecting the best severity KRIs.  
 *              Defaults to FORWARD.
 *
 *  NUMCCVS        =2,
 *
 *  CCV_NAMES      = e.g. INTERNAL_BL_UID INTERNAL_RC_UID,
 *
 *  interval = user-selected fit interval to do the fitting; defaults to YEAR.    
 *             possibles values: YEAR, SEMIYEAR, QTR, MONTH2, MONTH, SEMIMONTH, WEEK2, WEEK
 *
 *  ccUid    = if missing, we are handling the whole matrix at once.  if non-missing, we are
 *             doing one cell.
 *
 *  outCompressedData = Output dataset containing (compressed) frequency counts, plus 
 *          the firm & businessline level KRIs (called KRI_:) and the panel dummy 
 *          variables for CCVUID and possibly OWN (called CCVUID_x_:, where x=0 to NUMCCVs-1).  
 *          Also contains:  OFFSET (log (exposure/FSCALE) )  
 *             Defaults to compressed_data.
 *
 *  outParms  = Parameter dataset output by macro.

 *  outStatus     =  ,
 *
 *  outNotSelected = Column of variables (KRIs and dummies) that weren't selected as 
 *              significant by Proc Reg.  Defaults to not_selected.  

 *  outDroppedDummies = ,
 *
 *  dataType = I (internal data) or C (consortium data).
 *
 *   
 *
 *****************************************************************************************/
%macro ORVA_freq_selection( 
                           data              = ,
                           selection         = FORWARD,
                           NUMCCVS           = 2,
                           CCV_NAMES         = INTERNAL_BL_UID INTERNAL_RC_UID,
                           interval          = YEAR, 
                           ccUid             = ,
                           Model_Basis_Props = ,
                           makeDummiesOn     = ,
                           outCompressedData = compressed_data,
                           outParms          = _sel_outparms,
                           outSelDummies     = _sel_outSelDummies,
                           outSelKRIs        = _sel_outSelKRIs,
                           outStatus         = _sel_status,
                           outNotSelected    = _sel_not_selected, 
                           outDummyMap       = _sel_dummyMap,
                           outDroppedDummies = _sel_droppedDummies,
                           debug             = false,
                           timingFlag        = false,   
                           dataType          = I, 
                           scaleData         = 
                           );                     

   /*  Macro variables that were already set in orva_ffit.sas:
    *
    *  Freq_Kris_Suggested = space-delimited list of the KRIs (bl and firm) suggested for inclusion 
    *              in the model.  The selection heuristic (in proc reg) will choose the significant
    *              KRIs from this list.
    *                  
    *  Freq_Kris_Forced    =  space-delimited list of the KRIs (bl and firm) that MUST be included 
    *              in the model.  These enter the the "INCLUDE=" option on proc reg's 
    *              model statement.
    *              
    *  FREQ_USE_KRIS       = true or false;
    */

   /***** create the status data set up front *****/
   %let reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
   %orva_status_ds( statusds = &outstatus, status = 1, reason = &reason );

   %let status             = 0;
   %let reason             = %sysfunc(sasmsg(&orvads, successful_note, noquote));
   %let abort              = 0;
   %let numDummies         = 0;
   %let dummyNames         = ;
   %let numSelectedKRIs    = 0;  /* initializing */ 
   %let numSelectedDummies = 0;  /* initializing */ 
   %let SelectedKRIs       = ;   /* initializing */
   %let SelectedDummies    = ;   /* initializing */

   /***** clear out out temporary output datasets *****/
   data &outCompressedData; 
      set _NULL_; 
   run;

   data &outNotSelected;
      set _NULL_;
   run;

   data &outParms;
      set _NULL_;
   run;

   data &outDroppedDummies;
      set _NULL_;
   run;

   data &outSelKRIs;
      set _NULL_;
   run;

   data &outSelDummies;
      set _NULL_;
   run;

   data &outDummyMap;
      set _NULL_;
   run;

   /*****  Create offset and log_adjcount to include FSCALE, exposure, pobs. *****/
   %_create_offset_and_logadjcount( 
                                    dataM              = &data,
                                    dataTypeM          = &dataType,
                                    scaleDataM         = &scaleData,
                                    Model_Basis_PropsM = &Model_Basis_Props,
                                    outdata            = _fsel_work1 
                                    );

   /***** Determine the number of Freq_Kris_Forced and Freq_Kris_Suggested *****/
   %let numForcedKRIs    = %orvu_str_getwordcount(strIn = &Freq_Kris_Forced, delimiters = %str( ));
   %let numSuggestedKRIs = %orvu_str_getwordcount(strIn = &Freq_Kris_Suggested, delimiters = %str( ));
      
   /***** Set up dummy variables per ibl.  (10/21/08: dataType == I data only *****/
   /*  initialize immediately before hand rather than all up top*/
   %_create_dummy_variables( 
                              indata             = _fsel_work1, 
                              outdata            = _fsel_work5,
                              dataM              = &data,
                              makeDummiesOnM     = &makeDummiesOn,
                              outDroppedDummiesM = &outDroppedDummies, 
                              outDummyMapM       = &outDummyMap,
                              numDummiesM        = numDummies, 
                              dummyNamesM        = dummyNames
                              );
   %if %upcase(&timingFlag) EQ TRUE %then %do;
      %orvu_timer(op = MARK_AND_RESET, tvar = _ffit_timer_, msg = ORVA_FREQ_SELECTION - Setting up Dummy Variables);
   %end;

   /***** Check whether there are more possible parameters than count observations *****/
   /*  That is: (sum of Forced KRIs + model params + dummies) */
   %let countobs = 0;  
   %orvu_getObsCount(_fsel_work5, countobs);

   %if %upcase(&Freq_Use_Kris) EQ TRUE %then %do; 
      %let param_limit = %eval(&numForcedKris + %sysfunc(min(&numdummies,1)) + 2);

      /*  If the selection method is backwards, include suggested KRIs in the check - 
      PROC reg can produce a garbage fit in this case */
      %if %upcase(&selection) EQ BACKWARD %then %do;
         %let param_limit = %eval(&param_limit + &numSuggestedKris);

         %if %sysevalf(&countobs LE &param_limit, BOOLEAN) %then %do;
            %let reason = %sysfunc(sasmsg(&orvads, fs_too_many_suggested_error, noquote, &countobs));
            %let status = 1;
         %end;
      %end;
      %else %do;
         %if %sysevalf(&countobs LE &param_limit, BOOLEAN) %then %do;
            %let reason = %sysfunc(sasmsg(&orvads, fs_too_many_params_error, noquote, &countobs));
            %let status = 1;
         %end;
      %end;
   %end;
   /* No KRI case */
   %else %if %sysevalf(&countobs LE %sysfunc(min(&numdummies,1)) + 2, BOOLEAN) %then %do;
      %let reason = %sysfunc(sasmsg(&orvads, fs_too_many_params_error, noquote, &countobs));
      %let status = 1;
   %end;

   /* If any errors were found, print error message and abort */
   %if &status. eq 1 %then %do;
      %put &orvar_error &reason.;
      %goto DONE; 
   %end;

   %if %eval(&numDummies + &numSuggestedKRIs) GT 0 %then %do;
      %_orv_select( 
                     indata               = _fsel_work5,
                     Freq_Kris_SuggestedM = &Freq_Kris_Suggested,
                     Freq_Kris_ForcedM    = &Freq_Kris_Forced,
                     Freq_Use_KrisM       = &Freq_Use_Kris,
                     numForcedKrisM       = &numForcedKris,
                     dummyNamesM          = &dummyNames,
                     selectionM           = &selection,
                     numSelectedKrisM     = numSelectedKris,
                     numSelectedDummiesM  = numSelectedDummies,
                     outSelDummiesM       = &outSelDummies,
                     outSelKrisM          = &outSelKris,
                     outNotSelectedM      = &outNotSelected
                     );
      %if %upcase(&timingFlag) EQ TRUE %then %do;
         %orvu_timer(op = MARK_AND_RESET, tvar = _ffit_timer_, msg = ORVA_FREQ_SELECTION - Selection in Proc REG);
      %end; 
   %end;
   %else %do;  /* no need to do selection */
      %let numSelectedKRIs    = &numForcedKRIs;
      %let numSelectedDummies = 0;

      %if &numForcedKRIs GT 0 %then %do;
         data &outSelKRIs;
            length _NAME_ $ 32;
            %let nkri = 1;
            %let kri  = %scan(&Freq_Kris_Forced, &nkri, ' ');

            %do %while (&kri NE %str() );
               _NAME_ = "&kri";
               output;
               %let nkri = %eval(&nkri + 1);
               %let kri  = %scan(&Freq_Kris_Forced, &nkri, ' ');
            %end;
         run; 
      %end;
   %end;

   data &outCompressedData(compress = binary);
      set _fsel_work5;
      drop log_adjcount;
   run;

   data &outParms;
      numSelectedKRIs    = &numSelectedKris;
      numSelectedDummies = &numSelectedDummies;

      if missing(numSelectedKris) then numSelectedKris = 0;
      if missing(numSelectedDummies) then numSelectedDummies = 0;

      Freq_Kris_Suggested = "&Freq_Kris_Suggested";
      Freq_Kris_Forced    = "&Freq_Kris_Forced";
   run;

%DONE:

   %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, fs_abnormal_error, noquote));
      %let status = 1;
   %end;   

   %orva_status_ds( statusds = &outstatus, status = &status, reason = &reason );

   /*- delete unwanted data sets -*/
   %if &debug = keepdatasets %then %do; 
   %end;
   %else %do;
      %orvu_delete_data( data = _fsel_: );
   %end;
%mend ORVA_freq_selection;

%macro _create_offset_and_logadjcount( 
                                       dataM              = ,
                                       dataTypeM          = ,
                                       scaleDataM         = ,
                                       Model_Basis_PropsM = ,
                                       outdata            = 
                                       );
   %let nobs_scale = 0;

   %if &scaleDataM NE %str() %then %do;
      %orvu_getObsCount(&scaleDataM, nobs_scale);
   %end; 

   %if &nobs_scale NE 0 %then %do;
      %if %upcase(&dataTypeM) eq I %then %do; /* INTERNAL */
         %if %upcase(&Model_Basis_PropsM) NE INTERNAL_BL_UID %then %do;
            %put &orvar_warning %sysfunc(sasmsg(&orvads, fs_cant_merge_scale_data_warning, noquote));

            data _fsel_work0;
               set &dataM;
            run;
         %end;
         %else %do;
            /*- merge count data with Scaling data by INTERNAL_BL_UID and YEAR -*/
            proc sql noprint; 
               CREATE TABLE
                              _fsel_work0
               AS SELECT
                              * 
               FROM
                              &dataM AS L 
               LEFT JOIN
                              &scaleDataM( keep = 
                                                   BLUID_SC 
                                                   YEAR 
                                                   FSCALE
                                                   ) AS S
               ON 
                              L.INTERNAL_BL_UID = S.BLUID_SC 
               AND
                              year(L.interval_start) = S.YEAR
                              ;   
            quit;       
         %end;
      %end; /* INTERNAL */
      %else %if %upcase(&dataTypeM) EQ E %then %do; /* EXTERNAL */
         /*- merge count data with Scaling data by YEAR -*/
         proc sql noprint; 
            CREATE TABLE 
                        _fsel_work0
            AS SELECT 
                        * 
            FROM
                        &dataM AS L 
            LEFT JOIN
                        &scaleDataM( keep =
                                             YEAR 
                                             FSCALE
                                    ) as S
            ON
                        year(L.interval_start) = S.YEAR
                        ;   
         quit;       
      %end;  /* EXTERNAL */
   %end;
   %else %do;
      data _fsel_work0;
         set &dataM;
      run;
   %end;

   /***** Drop obs with zero exposure; create log_adjcount for use in Proc Reg variable selection.  *****/
   data &outdata(compress=binary);
      set _fsel_work0; 
      /*    where exposure GT 0;  */

      if missing(FSCALE) then do;
         FSCALEadj = 1; 
      end;
      else do;
         FSCALEadj = FSCALE;
      end;

      %if &nobs_scale ne 0 %then %do;
         %if %upcase(&dataTypeM) eq I %then %do; 
            drop BLUID_SC YEAR; 
         %end;

         %if %upcase(&dataTypeM) eq E %then %do; 
            drop YEAR; 
         %end;
      %end;

      if net_exposure_no_scale LE 0 then do;
         offset = 1;  /* This makes Countreg's output out=dataset compute the correct pred value for such observations. */
         log_adjcount = .;
         the_count = .;
      end;

      /* with FSCALEAdj, the implied_count that is used in %orva_freq_combine is now:
      * implied_count = (the_count*FScaleAdj)/(exposure*pobs)
      * If pobs is zero, then 
      * implied_count = (the_count*FScaleAdj*smallestDouble)/exposure
      * added 0.1 to prevent the the_count=0 case in the logarithm function.
      */
      else do;
         offset = log(net_exposure_no_scale) - log(FSCALEadj); 
         log_adjcount = log(the_count+.1)-offset;
      end;
   run;
%mend _create_offset_and_logadjcount;

%macro _create_dummy_variables( 
                                 indata             = ,
                                 outdata            = ,
                                 dataM              = ,
                                 makeDummiesOnM     = ,
                                 outDroppedDummiesM = , 
                                 outDummyMapM       = ,
                                 numDummiesM        = , 
                                 dummyNamesM        =
                                 );

   %if &makeDummiesOnM NE %str() %then %do;
      data _fsel_designinput;
         set &dataM(keep = &makeDummiesOnM );
      run;

      %orva_designmatrix(  
                           A                  = _fsel_designinput, 
                           design             = _fsel_designoutput, 
                           sparse             = true, 
                           combosonly         = false, 
                           drop               = true, 
                           outDroppedDums     = &outDroppedDummiesM,
                           keepallowndummies  = false, 
                           status_design      = _fsel_statusdesign
                           );

      /* !!!! New thinking (3.3):  Drop the nth dummy for all types of dummy, and use an INT model in countreg. */
      %_make_dummy_map( 
                        props   = &makeDummiesOnM, 
                        thedata = _fsel_designoutput, 
                        themap  = &outDummyMapM
                        );
      
      data _fsel_designoutput;
         set _fsel_designoutput;
         %let nprops = 1;
         %let prop_curr = %scan(&makeDummiesOnM, &nprops, ' ');

         %do %while (&prop_curr ne );
            rename &prop_curr = the_&prop_curr;
            %let nprops = %eval(&nprops + 1);
            %let prop_curr = %scan(&makeDummiesOnM, &nprops, ' ');
         %end;
      run;
      %let nprops = %eval(&nprops - 1);

      /***** add dummy var columns to &indata *****/
      %let prop_1 = %scan(&makeDummiesOnM, 1, ' ');
      proc sql;
         CREATE TABLE
                           _fsel_work4 
         AS SELECT
                           *
         FROM
                           &indata L 
         LEFT OUTER JOIN 
                           _fsel_designoutput R
         ON 
                           (L.&prop_1 EQ R.the_&prop_1 
         %let nprops = 2;
         %let prop_curr = %scan(&makeDummiesOnM, &nprops, ' ');

         %do %while (&prop_curr ne );
            AND 
                           L.&prop_curr EQ R.the_&prop_curr

            %let nprops  = %eval(&nprops + 1);
            %let prop_curr = %scan(&makeDummiesOnM, &nprops, ' ');
         %end;
         );
      quit;
      %let nprops = %eval(&nprops - 1);

      /* After this data step, Dummy_internal_bl_uid_: will be dummies */
      data &outdata; /* we are keeping the original values in internal_bl_uid */
         set _fsel_work4(drop =
                              %let nprops = 1;
                              %let prop_curr = %scan(&makeDummiesOnM, &nprops, ' ');

                              %do %while (&prop_curr ne );
                                 the_&prop_curr 
                                 %let nprops = %eval(&nprops + 1);
                                 %let prop_curr = %scan(&makeDummiesOnM, &nprops, ' ');
                              %end;
                              );  
      run;

      /***** Determine the names of and the number of dummy variables. *****/
      proc contents 
                     data = &outdata 
                     out  = _fsel_content(keep = NAME) 
                     noprint
                     ;
      run;

      data _fsel_content2;
         set _fsel_content;
         if compare(upcase(scan(NAME, 1, '_')), 'DUMMY') EQ 0 then output;
      run;

      %let &numDummiesM = 0;
      %let &dummyNamesM =;    
      %orvu_getObsCount(_fsel_content2, &numDummiesM);

      %if &&&numDummiesM GT 0 %then %do;
         proc sql noprint;
            SELECT 
                     NAME
            INTO 
                     :&dummyNamesM SEPARATED BY " "
            FROM 
                     _fsel_content2
                     ;
         quit;
      %end;
   %end;
   %else %do;
      data &outdata;
         set &indata;
      run;  
   %end;
%mend _create_dummy_variables;

%macro _orv_select( 
                     indata               = ,
                     Freq_Kris_SuggestedM = ,
                     Freq_Kris_ForcedM    = ,
                     Freq_Use_KrisM       = ,
                     numForcedKrisM       = ,
                     dummyNamesM          = ,
                     selectionM           = ,
                     numSelectedKrisM     = ,
                     numSelectedDummiesM  = ,
                     outSelDummiesM       = ,
                     outSelKrisM          = ,
                     outNotSelectedM      = 
                     );

   /* Select the variables with significant impact on the counts */
   proc reg 
            data   = &indata 
            outest = _regResults_a  
            noprint  /* tableout */
            ;
      /* Don't use KRI_: or you'll use the severity KRIs too, which is wrong */
      model 
            log_adjcount = 
                           &Freq_Kris_ForcedM 
                           &Freq_Kris_SuggestedM 
                           &dummyNamesM 
                           / 
                           selection = &selectionM 
                           SLENTRY   = 0.05 
                           SLSTAY    = 0.05

                           %if &numForcedKrisM GT 0 %then %do; 
                              include = &numForcedKrisM  
                           %end;   
                           ;
   run;        

   proc transpose 
                  data = _regResults_a(drop =   _RMSE_ 
                                                Intercept 
                                                log_adjcount
                                       ) 
                                       out = _regResults_
                                       ;
      id _TYPE_; 
   run;

   data _regResults_ 
      &outSelDummiesM(keep = _NAME_)
      &outSelKrisM(keep = _NAME_)
      &outNotSelectedM(rename = (PARMS = Estimate));

      length _NAME_ $ 32;         
      set _regResults_ end = last;

      if PARMS NE . then do;
         if compare(scan(_NAME_, 1, '_'), 'KRI') NE 0 then do;
            numSelectedDummies + 1;
            output &outSelDummiesM;
         end;
         else do;
            numSelectedKRIs + 1;
            output &outSelKrisM;
         end;

         output _regResults_;
      end;
      else do;
         output &outNotSelectedM;
      end;
   
      if last then do;
         call symputX("&numSelectedKrisM", numSelectedKRIs);
         call symputX("&numSelectedDummiesM", numSelectedDummies);
      end;

      drop numSelectedKRIs numSelectedDummies ;
   run;
%mend _orv_select;

/*
 * Purpose: To make a map that reflects the corresponding relations of the 
 * dummy name, the level variable name and the unique value. The relation is
 * represented differently in the input dataset that usually was generated
 * by %orva_designmatrix. For example: The input
 * ------------------------------------------
 * |a |b  | dummy_a_1 |dummy_a_2 |dummy_b_1 |
 * +------+-----------+----------+----------+
 * |1 |10 |   1       |   0      | 1        |
 * |1 |11 |   1       |   0      | 0        |
 * |2 |10 |   0       |   1      | 1        |
 * |2 |11 |   0       |   1      | 0        |
 * |3 |10 |   0       |   0      | 1        |
 * |3 |11 |   0       |   0      | 0        |
 * ------------------------------------------
 * is mapped to:
 * ----------------------------------------
 * |Value |Level_variable_name | _name_   |
 * +------+--------------------+----------+
 * | 1    |a                   |dummy_a_1 |
 * | 2    |a                   |dummy_a_2 |
 * | 10   |b                   |dummy_b_1 |
 * ----------------------------------------
 * 
 * Notice that the value 3(of a) and 11(of b) are dropped as model required.
 * a and b here represent two levels of distinct value seperation.  --xiaoma 
 */
%macro _make_dummy_map( 
                        props   = ,
                        thedata = , 
                        themap  = 
                        );
   data &themap;
      set _NULL_;
   run;

   %let nprops = 1;
   %let prop_curr = %upcase(%scan(&props, &nprops, ' '));

   %do %while (&prop_curr ne );
      %let dsid = 0; 
      %let dumnum = 0;
      %let nobs = 0;
      %let dsid = %sysfunc(open(&thedata,I));                                                                                        

      %if &dsid NE 0 %then %do;                                                                                                      
         %let nobs = %sysfunc(attrn(&dsid,nlobs));
         %let dumnum = %sysfunc(varnum(&dsid,Dummy_&prop_curr._1)); 
         %let rc = %sysfunc(close(&dsid));         
      %end;   

      %if &dumnum GT 0 AND &nobs GT 0 %then %do;
         data _dm1_&nprops;
            set &thedata( keep = &prop_curr Dummy_&prop_curr._:);
         run;

         proc sort 
                  data = _dm1_&nprops 
                  out  = _dm2_&nprops 
                  nodupkey;
            by &prop_curr;
         run;

         %let numLevels = 0;
         %orvu_getObsCount(_dm2_&nprops, numLevels);

         %if &numLevels GT 0 %then %do;
            data _dm3_&nprops(keep = 
                                       _NAME_ 
                                       VALUE 
                                       Level_Variable_Name
                                       );

               set _dm2_&nprops(rename = (&prop_curr = VALUE));
               array Dum{&numLevels} Dummy_&prop_curr._1 - Dummy_&prop_curr._&numLevels;
               Level_Variable_Name = "&prop_curr";

               do j = 1 to &numLevels;
                  if Dum{j} = 1 then _NAME_ = "Dummy_&prop_curr._" || trim(left(j));
               end;

               if missing(_NAME_) then delete;
            run;

            data &themap;
               set 
                     &themap 
                     _dm3_&nprops
                     ;
            run;
         %end;
      %end;

      %let nprops = %eval(&nprops + 1);
      %let prop_curr = %upcase(%scan(&props, &nprops, ' '));
   %end;

   %orvu_delete_data( data = _dm: );
%mend _make_dummy_map;
