/*****************************************************************************************
 *   Copyright (c) 2004 by SAS Institute Inc.  Cary NC 27513  USA
 *
 *   Name: ORVA_freqcorr.sas                                             
 *   Description: Compute the correlation of the frequency of the losses.
 *          The correlations are computed at a smaller interval, interval,
 *          then the interval of the simulation, final_interval,
 *          so that more points can be used to create the covariance matrix. 
 *          A normal copula is used for the joint distribution and smoothing is
 *          used to create the bins.
 *          
 *
 * PARAMETERS
 *
 * intData     = the name of the internal loss data with KRIS.  Must have the following variables:
 *               AMOUNT and CCUID.
 * intSParm    = the name of the internal Severity fitting data set.
 *
 * intFParm    = the name of the internal Frequency fitting data set. 
 *               The data set has one row for each cell containing the FPARM data + CCUID.
 *
 * rrFParm     = the name of the RR Frequency fitting data set. 
 *               The data set has one row for each cell containing the FPARM data + CCUID.
 *
 * ccLibref    =  the name of the libref with the cc data information
 *
 * useWght     = Flag to use a weighting scheme:          
 *                       1 to turn it on 
 * 
 * lambda      = weight to use if usewght = 1             
 *                       0.0001 <= lambda <= 0.9999   
 *
 * GenTotals   = 0 or 1 to indicate whether covariances for total cells are needed.
 *      
 * interval    = SAS interval, defaults to QTR
 *
 * outStatus   = Success or failure - Default is outstatus 
 *
 * outCorr     = Correlation matrix result               
 *                                                      
 ****************************************************************/
 %macro ORVA_freqcorr(
                      intdata    = ,
                      intSParm   = ,
                      intFParm   = ,  
                      rrFParm    = ,
                      useWght    = 0,
                      lambda     = .5,
                      interval   = QTR,
                      CCLibref   = varprj., 
                      GenTotals  = 1,
                      outStatus  = outstatus,
                      outCorr    = outcorr 
                      );

	/*- create the status data set upfront -*/
	%let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
	%orva_status_ds( statusds=&outstatus, status=1, reason=&reason );
	 
   %let status=0;
   %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
   %let abort=0;
   %let width=5;

   /*- error checking -*/
	%if &intdata=%str() %then %do;
		%put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_specify_intdata_error, noquote));
		%let abort=1;
		%let reason=%sysfunc(sasmsg(&orvads, fcorr_IntData_missing_error, noquote));
		%goto DONE;
	%end;

	%if  &usewght = 1 %then %do;
		%if %sysevalf(&lambda < 0.0001, BOOLEAN) or 
		    %sysevalf(&lambda > 0.9999, BOOLEAN) %then %do;
			%put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_lambda_out_of_range_error, noquote));
			%let abort=1;
			%let reason=&orvar_error %sysfunc(sasmsg(&orvads, fcorr_lambda_out_of_range_error, noquote));
			%goto DONE;
		%end;
	%end;

	/* write the pexcess code to a file using the datastep */
	/* This is a series of datasteps that compute the pexcess */
	/* for each distribution ---------------------------------*/
	data _NULL_;
                rc=filename("tmp",,"temp");
	run;

   data _NULL_; set &intSparm;
      file tmp;
         length KRI_LOG $2000;
         array KRI Param_1_: dummy_var;
         put 'proc transpose data=&intSparm out=_parm_; run;';
	      put 'data _null_;';
	      put '	  set _parm_; where upcase(substr(_NAME_,1,6)) eq "PARAM_";';
	      put '	  call symput(_name_, put(col1, BEST20.));';
         put '   %let sdist=' distribution ";";
	      put 'run;';
         put '%orva_set_sev_dist(distIn= &sdist, distOut=sdist );';

         put "run;";
         put '%orvd_&sdist;';
         put 'data _' CCUID '; set &intdata;';
         put '  where CCUID =' CCUID ';';
         put '  if THRESHOLD=0 then do;';
         put '     pExcess=1;';
         put '  end;';
         put '  else  do;';

			kri_count=0;

			do i=1 to dim(KRI)-1;
				if KRI[i] NE . then kri_count=kri_count+1;
			end;

			if kri_count > 0 then do;
				KRI_LOG = "0";

				do i=1 to dim(KRI)-1;
					if KRI[i] NE . then do;
						KRI_LOG = trim(KRI_LOG) || "+" || trim(substrn(vname(KRI(i)),9)) || "*" ||KRI(i);
					end;
				end;

				put "THRESHOLD = THRESHOLD/exp("  KRI_LOG  ");";
			end;

         put '      %Cdf&sdist(THRESHOLD , OneMinus_pExcess)';
         put '      pExcess = 1 - OneMinus_pExcess;';
         put "   end;";
         put "   if fscale ^= . then adjcnt = fscale;";
         put "   else adjcnt = 1;";
         put "   if pexcess ^= . and pexcess ^= 0 then adjcnt = adjcnt/pexcess;";
         put "run;";
   run;

	/* Adjust the cnts for scaling and for thresholds */
	%inc tmp;

	data _NULL_;
                rc=filename("tmp",,"temp");
        run;

	data _NULL_; 
		set &intSparm;
		file tmp;
		put " _" CCUID;
	run;

	data _freqcnts; 
		set 
		%inc tmp;
		;
	run;

	/*- First get FREQUENCY numbers -*/
	proc sort data=_freqcnts; 
		by CCUID FINANCIALDATE;
	run;

	proc timeseries data=_freqcnts out=_freqcnts;
		id FINANCIALDATE interval=&interval acc=total;
		var adjcnt;
		by CCUID;
	run;

        data _freqcnts; set _freqcnts;
            length _newkey $ 32;

            /* Make sure we have a alpha char for the key */
            _newkey = compress('Cell_'|| put(CCUID, &ORVAR_CELLFMT));
            n = adjcnt;

            keep _newkey n FINANCIALDATE;
      run;

        proc sort data=_freqcnts; by FINANCIALDATE;
        run;

       /* make the losses into columns */
        proc transpose data=_freqcnts out=_transpose(drop= _NAME_);
            by FINANCIALDATE;
            id _newkey;
            var n;
        run;

        data _transpose; set _transpose;
         array t[*] _numeric_;
         do i=1 to dim(t); if t[i]= . then t[i]= 0; end;
         drop FINANCIALDATE i;
        run;

      %if ( &usewght = 1 ) %then %do;
          data _NULL_;set _transpose;
                call symputX( 'nobs', _N_);
            run;
        
        data _transpose;
            set _transpose;
            omega = (( &lambda) ** (&nobs - _N_ ) ) * (1 - &lambda);
        run ;
      %end;

      %if ( &usewght = 1 ) %then %do;
        proc corr data= _transpose outk= &outcorr 
                        noprint vardef=weight;
                        weight omega;
        run;
    %end;
    %else %do;
        proc corr data= _transpose outk=&outcorr noprint;
        run;
    %end;

    data &outcorr; set &outcorr;
        where _type_ = "CORR";
                    /* Convert back to a pearson */
        array y[*] _NUMERIC_;
        do i = 1 to dim(y);
           y[i] = sin(constant('pi')*y[i]/2); 
        end;
        output;
    run;


          /* get an identity matrix of the system to make the covariance matrix the same */
          /* size as all of the other generated covariance matrices ---------------------*/
    %ORVA_COMBINE_COV(    ccLibref      =&cclibref,  
                          outCov        =_Identity_,
                          GenTotals     = 0,
                          ostat         =_status_ );
         /* check the status */
    data _NULL_; set _status_;
       call symputX('i_status', status);
    run;
    %if &i_status=1 %then %do;
            %put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_id_mat_create_failed_error, noquote));
            %let abort=1;
            %let reason=%sysfunc(sasmsg(&orvads, fcorr_id_mat_create_failed_error, noquote));
            %goto DONE;
    %end;

         /* combine the two matrices -------------------------------*/
    proc sort data=_Identity_; by _NAME_; run;
    proc sort data=&outcorr; by _NAME_; run;

    data &outcorr; merge _Identity_ &outcorr; by _NAME_; 
      length _TYPE_ $ 8;
      array miss[*] _numeric_;
	   do i = 1 to dim( miss );
	       if miss[i] = . then miss[i] = 0;
		   if miss[i] < 0 then miss[i] = 0;
	   end;
	   _TYPE_='CORR';
    run;

    %if &GenTotals = 1 %then 
       %orva_rollup_cov(cov=&OutCorr, ccLibref=&ccLibref);

%DONE:

      %if &abort %then %do;
            %put &orvar_error %sysfunc(sasmsg(&orvads, fcorr_abnormal_error, noquote));
            %let status=1;
      %end;  
      %orva_status_ds( statusds=&outstatus, status=&status, reason=&reason );
      
      proc datasets library=work memtype=data nolist;
            delete _transpose _freqcnts ;
      quit;

%mend;
