/*****************************************************************************************
 * $Id: orva_getmomts.sas,v 1.1.4.29.2.12.2.9.6.1 2012/05/03 13:08:38 sasdje Exp $
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME:    ORVA_GetMomts.sas                                             
 * PURPOSE: Calculates frequency and severity moments and aggregations for all cells.           
 *          Calculates the aggregates, mean, and count for the Loss Effects Breakdown.
 * SUPPORT: stchri
 *
 * PARAMETERS:
 *
 *   lossData      =  the name of the loss data with the following variables:
 *   dataType      =  (I)nternal, (C)onsortium, (P)ublic.
 *   global_range  =  global range dataset for this project.
 *   missing_range =  name of the missing ranges dataset with the following variables:
 *                    &CCV_NAMES, START_MISSING_DATE, END_MISSING_DATE.
 *   interval      =  DATE interval (YEAR, MONTH ...). Default is YEAR.
 *   ccData        =  Needed for input to orva_CELL_WHERE, to subset losses for this cell.
 *   numCCvs       =  Number of cross-classification variables in project.
 *   CCV_NAMES     =  space-delimited list of cross-classification variables in project.
 *   outSevMomts   =  MEAN, VARIANCE, STD, SKEWNESS, KURTOSIS, COUNT, MINLOSS, MAXLOSS, CCV_n, SUM (for AMOUNT)
 *                    LEGAL_LIABILITY (aggregate), LEGAL_LIABILITY_MEAN, LEGAL_LIABILITY_N (count)
 *                    REGULATORY_ACTION (aggregate), REGULATORY_ACTION_MEAN, REGULATORY_ACTION_N (count)
 *                    RESTITUTION (aggregate), RESTITUTION_MEAN, RESTITUTION_N (count)
 *                    LOSS_TO_ASSETS (aggregate), LOSS_TO_ASSETS_MEAN, LOSS_TO_ASSETS_N (count)
 *                    LOSS_OF_RECOURSE (aggregate), LOSS_OF_RECOURSE_MEAN, LOSS_OF_RECOURSE_N (count)
 *                    WRITE_DOWNS (aggregate), WRITE_DOWNS_MEAN, WRITE_DOWNS_N (count)
 *   outFreqMomts  =  CCUID, MEAN, VARIANCE, CCV_n
 *   outStatus     =  Success or failure - default is status
 *   sharedRollup  =  preform shared loss rollup of data first.
 *   singleEventMult = column that specifies shared loss key.
 *   outprops      =  Output dataset for panel properties.  This only needs to be created 
 *                    once per project per datatype, so it is ignored if sharedRollup=false
 *   outmissprops  =  Output dataset for panel props with missing ranges.  
 *                    This only needs to be created once per project per
 *                    datatype, so it is ignored if sharedRollup=false
 *   outmissrange  =  Updated missing range table.  This only needs to be created 
 *                    once per project per datatype, so it is ignored if sharedRollup=false
 *   scaleData     =  scaling dataset (currently only for internal data)
 *   
 *****************************************************************************************/
 %macro ORVA_GetMomts( lossData       =,
                       dataType       =I,
                       global_range   =,
                       missing_range  =,
                       interval       =YEAR,
                       ccData         =,
                       numCCVs        =2,
                       CCV_NAMES      = INTERNAL_BL_UID INTERNAL_RC_UID , 
                       outSevMomts    =osstat,              
                       outFreqMomts   =ofstat,              
                       outSevMomtsRU  =,              
                       outFreqMomtsRU =,              
                       outStatus      =statusgetmomts,
                       singleEventMult=,
                       outprops       =,
                       outmissprops   =,
                       outmissrange   =,
					   outCountData   =theOutCountData, 
					   outTrendData   =theOutTrendData,
                       ctool          =false,
                       scaleData      =
                     );  

	

    /*- create the status data set up front -*/
    %let reason=%sysfunc( sasmsg(&orvads, unknown_error, noquote) );
    %orva_status_ds( statusds=&outstatus, status=1, reason=&reason );
    
    %let status=0;
    %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
    %let abort=0;

    /*- empty out results -*/
    data &outSevMomts; set _NULL_; run;
    data &outFreqMomts; set _NULL_; run;
    data &outSevMomtsRU; set _NULL_; run;
    data &outFreqMomtsRU; set _NULL_; run;

    /*- error checking -*/
    %if &lossData=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, gmom_specify_global_range_ds_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, gmom_missing_lossData_error, noquote));
        %goto DONE;
    %end;
    %if &global_range=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, gmom_specify_global_range_ds_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, gmom_specify_global_range_ds_error, noquote));
        %goto DONE;
    %end;

    /***** get BYVARS macro variable *****/
    data _NULL_;
       length byvars $100;
       %do i = 0 %to &NUMCCVS-1; 
          byvars = trim(left(byvars)) || " CCVUID_&i";
       %end;    
       call symput("BYCCVS", byvars);
    run;

    /* first handle normal moments */   
    %getmoments_norollup(lossData =&lossData, 
          CCV_NAMES      =&CCV_NAMES,       
          byCCVs         =&BYCCVS, 
          numCCVs        =&numCCvs,
          ccData         =&ccData, 
          interval       =&interval,
          missing_range  =&missing_range,
          outFreqMomts   =&outFreqMomts,
          outSevMomts    =&outSevMomts,
          global_range   =&global_range,
          dataType       =&dataType,
          outCounts      =&outCountData
    );
    
    /* second handle shared loss rollup */
    %if &abort eq 0 %then %do;
        %let sharedLossCount=0;
        %if &singleEventMult ne %str() %then %do;
           PROC SQL noprint;
             select count(*) into :sharedLossCount
             from &lossData
             where &singleEventMult notin (0,.);
           quit;
        %end;
        %if &sharedLossCount eq 0 %then %do;
           data &outSevMomtsRU; set &outSevMomts; run;
           data &outFreqMomtsRU; set &outFreqMomts; run;
        %end;
        %else %do;
           %getmoments_rollup(lossData        =&lossData, 
                              byCCVs          =&BYCCVS, 
                              numCCVs         =&numCCvs,
                              ccData          =&ccData, 
                              singleEventMult =&singleEventMult, 
                              dataType        =&dataType, 
                              outSevMomts     =&outSevMomtsRU, 
                              outFreqMomts    =&outFreqMomtsRU,
                              global_range    =&global_range);	
        %end;
    %end;
  
%DONE:
     
    %if &abort %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, gmom_abnormal_error, noquote));
       %let status=1;
    %end;  
    %orva_status_ds( statusds=&outstatus, status=&status, reason=&reason );


%mend ORVA_GetMomts;


/*********************************************************************************
 * Shared Loss Rollup Moments
 * 1. Use the %orva_rollup() macro to break up the losses into 2 datasets:
 *    'singles' and 'shares'.
 *
 *    The 'shares' contain shared losses that are 'rolled up' for all possible
 *    permutations of CCVs and thus may have duplicates since entries are
 *    created (rolled up) in aggregate cells.
 *
 *    The 'singles' contain non-shared losses and should never have duplicates.
 *
 * 2. The %orva_cell_slices macro creates a dataset of 'slices' - combinations
 *    of cells that constitute permutations of the CCVs.
 *
 * 3. call normal moments macros for each slice.
 *********************************************************************************/
%macro getmoments_rollup(lossData        =, 
                         byCCVs          =, 
                         numCCVs         =, 
                         ccData          =,
                         singleEventMult =, 
                         dataType        =, 
                         outSevMomts     =, 
                         outFreqMomts    =,
                         global_range    =);
    
   
     /***** rollup shared losses in _rollup dataset *****/
     %orva_rollup(lossData=&lossData, byCCVs=&BYCCVS, numCCVs=&numCCvs,
                  ccData=&ccData, singleEventMult=&singleEventMult, 
                  outDataShared=_rollup, outDataSingle=_norollup, dataType=&dataType);     
      
     /**** create slices for each level of aggregation *****/
     %let first= true;    
     %let NUM_SLICES= 0;
     %let NUM_CCVARS= 0;
     %orva_cell_slices(ccData=&ccData,cellKey=CCUID,outData=_slices,guardForTotals=true);
     
     /***** loop over slices *****/
     %do slice=1 %to &NUM_SLICES;

          /****** get slice of losses ******/
          %getdata_byslice( sharedData  =_rollup,
                            singleData  =_norollup,
                            ccData      =&ccData,
                            sliceData   =_slices,
                            sliceNum    =&slice,
                            outData     =_slice);

         /***************************************************************************
          * SEVERITY MOMENTS
         ****************************************************************************/
 
         %getsevmomts(lossData=_slice, outSevMomts=_sliceMomts, 
                      byCCVs=CCUID, numCCVs=&numCCVs, ccData=&ccData, byclass=false);
            
         %orvu_append(base=&outSevMomts,new=_sliceMomts,canUseAppend=true);     

         /***************************************************************************
          * FREQUENCY MOMENTS
          ***************************************************************************/
          %orva_getfreqmomts( 
                    lossData       =_slice,
                    dataType       =&dataType,
                    global_range   =&global_range,
                    interval       =&interval,
					     cTool          =&cTool,
                    missing_range  =&missing_range,
                    CCV_NAMES      =&CCV_NAMES, 
                    NUMCCVS        =&NUMCCVS,
                    outFreqMomts   =_sliceMomts,
                    outStatus      =_ofreqStatus
                  %if &scaleData NE %str() %then %do;
                    ,scaleData     =&scaleData
                  %end;
                   );    
            
         /***** stack up slice *****/
         %if &slice eq 1 %then %do;
              data &outFreqMomts; set _sliceMomts; run;
         %end;
         %else %do;
              proc append base=&outfreqMomts data=_sliceMomts; run;
         %end;
         

     %end; /* looping over aggregation slices */
    
      /***** fix up CCVUID_<num> in moments datasets, 
               while the CCUID is correct the individual
               CCCUIDs are arbitrary at this point *****/
      data _NULL_; set _slices(where=(whereClauseLength=0));
           call symput('hashData',trim(left(keyHash)));
      run;
      data &outSevMomts; 
          set &outSevMomts(drop= %do i = 0 %to &numCCVs-1; CCVUID_&i %end; );
          if _N_ eq 1 then do;
              declare hash ccHash(dataset: "&ccData");
              rc = ccHash.defineKey("CCUID");
              rc = ccHash.defineData(&hashData);
              rc = ccHash.defineDone();
              %do i = 0 %to &numCCVs-1;
                 call missing( CCVUID_&i);
              %end;
          end;
          rc = ccHash.find();
          drop rc;
      run;
      data &outFreqMomts; 
          set &outFreqMomts;
          if _N_ eq 1 then do;
              declare hash ccHash(dataset: "&ccData");
              rc = ccHash.defineKey("CCUID");
              rc = ccHash.defineData(&hashData);
              rc = ccHash.defineDone();
              %do i = 0 %to &numCCVs-1;
                 call missing( CCVUID_&i);
              %end;
          end;
          rc = ccHash.find();
          drop rc;
     run;


     /******************************************************************************
  
     /*- clean up -*/
     proc datasets library=work memtype=data nolist;
        delete _slice: _data _rollup _norollup _ofreqStatus;
     quit; 

%mend getmoments_rollup;



%macro getmoments_norollup(lossData   =, 
                       byCCVs         =, 
                       numCCVs        =, 
                       ccData         =,
                       CCV_NAMES      =, 
                       global_range   =,
                       interval       =,
                       missing_range  =,
                       outSevMomts    =,      
                       outFreqMomts   =,
                       dataType       =,
                       outCounts      =);

      /******************************************************************************
       * SEVERITY MOMENTS
       * handled with one call to %getsevmomts() using class variables
       *******************************************************************************/
       %getsevmomts(lossData=&lossData, outSevMomts=_sliceMomts, 
                    byCCVs=&BYCCVS, numCCVs=&numCCVs, ccData=&ccData);

        /*- get CCUIDs back from ccData -*/
        proc sql noprint;
          create table &outSevMomts as select * from
            _sliceMomts M left join &ccData C on 1
            %do i = 0 %to &numCCVs-1;    
               and M.CCVUID_&i eq C.CCVUID_&i    
            %end;
          ;
        quit;


      /******************************************************************************
        * FREQUENCY MOMENTS
        * handled with one call to %orva_getfreqmomts() for interior cells and
        * one call to %orva_freqmomtsaggr() for aggregate cells.
       *******************************************************************************/ 
        %orva_getfreqmomts( 
                    lossData       =&lossData,
                    dataType       =&dataType,
                    global_range   =&global_range,
                    interval       =&interval,
                    missing_range  =&missing_range,
                    CCV_NAMES      =&CCV_NAMES,               
                    NUMCCVS        =&NUMCCVS,
                    outFreqMomts   =_sliceMomts,
                    outStatus      =&outStatus,
                    outprops       =&outprops,
                    outmissprops   =&outmissprops,
                    outmissrange   =&outmissrange,
                    outCountData   =_outCounts1,
                    ctool          =&ctool
                   );    
                   
        /*- get CCUIDs back from ccData -*/
        proc sql noprint;
          create table _sliceMomts2 as select * from
            _sliceMomts M left join &ccData C on
             M.CCUID eq C.CCUID
          ;
        quit;
		  
        proc sql noprint;
          create table &outCounts as select * from
          _outcounts1 M left join &ccData C on
          M.CCUID eq C.CCUID
        ;
        quit;
		
  
        %orva_cell_slices(ccData=&ccData,cellKey=CCUID,outData=_slices,guardForTotals=true);  
        %orva_freqmomtsaggr(slicesDataset =_slices, 
                            interiorCells =_sliceMomts2,
                            numCCVs       =&NUMCCVS,
                            outDataset    =&outFreqMomts,
                            ccData        =&ccData);
      
		%if &dataType EQ I %then %do;
			%orva_freq_trend_line( dataM        = &lossData,
	                               ccDataM      = &ccData,
	                               NUMCCVS      = &NUMCCVS,
                                   interval     = &interval,
								   global_range = &global_range,
	                               outdata      = &outTrendData,
	                               trendStatus  = outTrendStatus
			);
		%end;

        /*- clean up -*/
        proc datasets library=work memtype=data nolist;
          delete _sliceMomts _sliceMomts2 _slices;
        quit; 

%mend getmoments_norollup;



/**********************************************************************************
  getsevmomts: calculate SEVERITY moments using proc summary
 **********************************************************************************/

%macro getsevmomts(lossData=, outSevMomts=, byCCVs=, numCCVs=, ccData=, byclass=true );
   
    /*- make sure loss dataset is sorted -*/
           
    proc sort data=&lossData  out  = _sortedData( keep=&byCCVs amount LEGAL_LIABILITY RESTITUTION LOSS_TO_ASSETS
                              LOSS_OF_RECOURSE WRITE_DOWNS NON_INSURANCE_RECOVERIES) ; 
    by &byCCVs; run;

     /*- handle moments for AMOUNT -*/
    proc summary data=_sortedData  	vardef=n noprint;
        output out=_SevMoments1(drop=_type_ _freq_) 
               mean=MEAN var=VARIANCE skew=SKEWNESS 
               kurt=KURTOSIS n=COUNT min=MINLOSS max=MAXLOSS sum=SUM;
        var AMOUNT; 
        %if &byclass eq true %then %do;
           class &byCCVs;
        %end;
        %else %do;
           by &byCCVs;
        %end;
    run;

    /*for loss categories: legal_liability, regulatory_action, etc, set the zero-valued loss to
      missing, before doing PROC SUMMARY, to get the no zero 
      count of the losses
     */


    data _sortedData;
        set _sortedData;
        if LEGAL_LIABILITY =0 then 
        LEGAL_LIABILITY = .;
        if REGULATORY_ACTION = 0 then 
        REGULATORY_ACTION = .;
        if RESTITUTION = 0 then
        RESTITUTION = .;
        if LOSS_TO_ASSETS = 0 then
        LOSS_TO_ASSETS = .;
        if LOSS_OF_RECOURSE =0 then
        LOSS_OF_RECOURSE = .;
        if WRITE_DOWNS = 0 then
        WRITE_DOWNS = .;
        if NON_INSURANCE_RECOVERIES = 0 then
        NON_INSURANCE_RECOVERIES = .;
    run;

    /*- do this separately because different stats are needed for this group -*/
    proc summary data=_sortedData;
      
        output mean= LEGAL_LIABILITY_MEAN REGULATORY_ACTION_MEAN RESTITUTION_MEAN 
                     LOSS_TO_ASSETS_MEAN LOSS_OF_RECOURSE_MEAN WRITE_DOWNS_MEAN 
                     NON_INSURANCE_RECOVERIES_MEAN
               sum= LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
                    LOSS_OF_RECOURSE WRITE_DOWNS NON_INSURANCE_RECOVERIES
               n= LEGAL_LIABILITY_N REGULATORY_ACTION_N RESTITUTION_N 
                  LOSS_TO_ASSETS_N LOSS_OF_RECOURSE_N WRITE_DOWNS_N 
                  NON_INSURANCE_RECOVERIES_N
               out=_SevMoments2(drop=_type_ _freq_);

        var LEGAL_LIABILITY REGULATORY_ACTION RESTITUTION LOSS_TO_ASSETS 
            LOSS_OF_RECOURSE WRITE_DOWNS NON_INSURANCE_RECOVERIES;

         %if &byclass eq true %then %do;
           class &byCCVs;
        %end;
        %else %do;
           by &byCCVs;
        %end;
    run;

     /*- merge moments datasets together -*/
    proc sort data=_SevMoments1; by &byCCVs; run; 
    proc sort data=_SevMoments2; by &byCCVs; run; 
    data &outSevMomts; 
        merge _sevMoments1 _sevMoments2;
        by &byCCVs;
        STD = sqrt(VARIANCE);
        Kurtosis=Kurtosis+3;
         %do i = 0 %to &numCCVs-1;    
           if CCVUID_&i eq . then CCVUID_&i = 0;    
          %end;
      run;
    
    
    /*- cleanup our temporary datasets -*/
     proc datasets library=work memtype=data nolist;
        delete _sorteddata _SevMoments1 _SevMoments2;
     quit;


%mend;

%macro getdata_byslice( sharedData =,
                        singleData =,
                          ccData     =,
                        sliceData  =,
                        sliceNum   =,
                        outData    =) ;

    data _NULL_; set &sliceData; where index eq &sliceNum;
        call symputX("ccWhereNum", whereClauseLength);
        call symput("ccWhere", trim(left(whereClause)));
        call symput("guardedWhere", trim(left(guardedWhereClause)));
        call symput("ccKeep", trim(left(keepClause)));
        call symputX("ccKeepNum", keepClauseLength);
        call symput("keyHash", keyHash);
    run;
    
    /*** non-Aggregate Slice:
          need all non aggregate entries from both datasets ***/  
     %if &ccWhereNum eq 0 %then %do;
        data _DAT&sliceNum.; 
           set &sharedData(where=(&guardedWhere)) 
               &singleData(where=(&ccWhere));
       run;
    %end;
    
    /*** GRAND-Aggregate slice:
         Need all entries from &singleData
         GuardedWhereClause works on &sharedData
         need to set all CCUIDs to grand aggregate cell ***/  
    %else %if &ccKeepNum eq 0 %then %do;
        /*- get CCUID for the grand aggregate cell -*/
        data _NULL_; set &ccData; 
           where 1 %do i = 0 %to &numCCVs-1;    
                      and CCVUID_&i eq 0    
                   %end;
           ;
           call symputX('CCUID',CCUID); 
        run;   
        data _DAT&sliceNum.; 
           set &sharedData(where=(&guardedWhere)) 
               &singleData;
           CCUID=&CCUID;
        run;
    %end;
    
    /*** simple-Aggregate slice (along 1 dimension):
         Need all entries from &singleData 
         GuardedWhereClause works on &sharedData
         use hashmap to get corrected CCUID ***/ 
    %else %do;
        data _cc; set &ccData; where &ccWhere; keep CCUID &ccKeep.; run;
       data _DAT&sliceNum.; set &singleData;
           if _N_ eq 1 then do;
             declare hash ccHash(dataset: '_cc', hashexp: 6);
            rc = ccHash.defineKey(&keyHash.);
             rc = ccHash.defineData("CCUID");
             rc = ccHash.defineDone();
           end;
          rc = ccHash.find();
       run;
       data _DAT&sliceNum.; 
           set &sharedData(where=(&guardedWhere)) 
               _DAT&sliceNum.;          
       run;
    %end;
   
   data &outData; set _DAT&sliceNum.; run;

   /*- cleanup our temporary datasets -*/
    proc datasets library=work memtype=data nolist;
        delete _DAT&sliceNum. _cc;
    quit;
  
%mend getdata_byslice;




/****************
%orvsetup(version=3.3);
libname VARPRJ "C:\VarTablesNew\ver_1\components\projects\bank41813";
options mprint symbolgen notes;
options nomprint nosymbolgen nonotes;

%orva_getmomts(lossData=VARPRJ.lossi,dataType=I,interval=QTR,global_range=VARPRJ.global_range_i,ccData=VARPRJ.cc,numCCVs=2,
 missing_range=VARPRJ.missing_range,ccv_names=INTERNAL_BL_UID INTERNAL_RC_UID ,sharedRollup=false,
 singleEventMult=SHARED_LOSS_GROUP_UID,outSevMomts=sevmomts_i,outFreqMomts=freqmomts_i,outStatus=WORK.status);

%orva_getmomts(lossData=VARPRJ.lossi,dataType=I,interval=QTR,global_range=VARPRJ.global_range_i,ccData=VARPRJ.cc,numCCVs=2,
 missing_range=VARPRJ.missing_range,ccv_names=INTERNAL_BL_UID INTERNAL_RC_UID ,sharedRollup=true,
 singleEventMult=SHARED_LOSS_GROUP_UID,outSevMomts=sevmomtsrollup_i,outFreqMomts=freqmomtsrollup_i,outStatus=WORK.status);
 
libname VARPRJ "C:\Documents and Settings\stchri\My Documents\My SAS Files\9.2\var_project4";
%orva_getmomts(lossData=VARPRJ.lossp,dataType=E,interval=YEAR,global_range=VARPRJ.global_range_p,
               ccData=VARPRJ.cc,numCCVs=2,missing_range=,ccv_names=BASEL_BL_UID BASEL_ER_UID ,
               sharedRollup=false,singleEventMult=,outSevMomts=VARPRJ.sevmomts_p,
               outFreqMomts=freqmomts_p,outStatus=WORK.status);
%orva_getmomts(lossData=VARPRJ.lossc,dataType=C,interval=YEAR,global_range=VARPRJ.global_range_c,
               ccData=VARPRJ.cc,numCCVs=2,missing_range=,ccv_names=BASEL_BL_UID BASEL_ER_UID ,
               sharedRollup=true,singleEventMult=SHARED_LOSS_GROUP_UID,outSevMomts=VARPRJ.sevmomts_rollup_c,
               outFreqMomts=freqmomts_rollup_c,outStatus=WORK.status);
 ****************/
