/*****************************************************************************************
 * $Id: orva_kri_forecast.sas,v 1.1.2.9.2.12.8.1 2013/01/15 22:11:06 sasdje Exp $
 *  Copyright (C), 2007
 *  SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_kri_forecast.sas                                             
 *  PURPOSE: Takes the output dataset from orva_interval_datasets
 *           plus forecasting options for KRIs, and then forecasts the KRIs
 *           at least up until "untildate".
 *            
 *           This macro is called during data versioning by the %orva_freq_preprocess macro.
 * 
 *  SUPPORT: lajack
 *
 *  PARAMETERS:
 *
 *  data = REQUIRED.  Output dataset from orva_interval_datasets.
 *
 *  foreout = name of the output dataset. 
 *
 *  status_KRIforecast
 *
 *  KRIoptions = REQUIRED.  Dataset with KRI options.
 *
 *  allccvnames = REQUIRED.
 *
 *  lastccv = REQUIRED.
 *
 *  untildate  = REQUIRED.  Date to which the KRIs should be forecasted.  
 *              Used to compute "lead" for Proc HPF.
 *               
 *  interval = Time interval size for which the analysis needs to be done, defaults to QTR.    
 *             possibles values: YEAR, SEMIYEAR, QTR, MONTH2, MONTH, SEMIMONTH, WEEK2, WEEK                                                                                                                       
 *      
 *****************************************************************************************/
 %macro orva_kri_forecast( data               = ,
                          foreout            = ,
                          status_KRIforecast = ,
                          KRIoptions         = ,
                          allccvnames        = ,
                          lastccv            = ,
                          untildate          = ,
                          interval           = qtr);

    %put %sysfunc(sasmsg(&orvads, status_note, noquote)) 
         %sysfunc(sasmsg(&orvads, krif_start_status, noquote, &interval));

    /**** create the status data set upfront with unknown status ****/
    %let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
    %orva_status_ds( statusds=&status_KRIforecast, status=1, reason=&reason );

    %let status=0;
    %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
    %let abort=0;

    /**** test that necessary parameters aren't empty ****/
    %if &data=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krif_specify_data_intv_ds_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, krif_missing_DATA_error, noquote));
        %goto DONE;
    %end;

    %if &KRIoptions=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krif_specify_kriopts_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, krif_missing_KRIoptions_error, noquote));
        %goto DONE;
    %end;

    %if &allccvnames=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krif_specify_ccvs_list_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, krif_missing_allccvnames_error, noquote));
        %goto DONE;
    %end;

    %if &lastccv=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krif_specify_lastccv_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, krif_missing_lastCCV_error, noquote));
        %goto DONE;
    %end;
   
    %if &untildate=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krif_specify_untildate_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, krif_missing_untildate_error, noquote));
        %goto DONE;
    %end;

   /********** compute LEAD *************/
   /**** Detect the earliest latest obs interval_start, across all bygroups.  ****/
   /**** Then compute lead = intck(&interval,earliest end data,untildate) ****/

   %let earliestlastdate = '04jan2119'd;  
   %let lead=0;
   data _null_;
        set &data end=last;
        by &allccvnames interval_start;
        retain datetobeat 0;    
        if _n_ = 1 then datetobeat = &earliestlastdate;
        if (last.&lastccv AND interval_start < datetobeat) then do;
           datetobeat = interval_start;
           lead = intck("&interval",datetobeat,&untildate); 
           call symputX("lead", lead);
        end;
   run;
   %if (&lead < 1) %then %do;
        %put &orvar_note %sysfunc(sasmsg(&orvads, krif_kris_already_exist_error, noquote, &untildate));
        %let abort=0;
   %let status=0;
        %let reason=%sysfunc(sasmsg(&orvads, krif_kris_already_exist_error, noquote, &untildate));
        %goto DONE;
   %end;

   /********* process forecasting options accumulate and model *********/
   %let NROWS=0;
   %let dsid=%sysfunc(open(&KRIoptions,I));
   %if &dsid NE 0 %then %do;
      %let NROWS=%sysfunc(attrn(&dsid,nlobs));
      %let rc=%sysfunc(close(&dsid));
   %end;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krif_cant_compute_rows_in_kri_forecast_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, krif_cant_compute_rows_in_kri_forecast_error, noquote));
        %goto DONE;
   %end;

    proc SQL noprint;
            select 
                    kri , 
                    accumulate , 
                    model ,
                    lower_bound format best20. ,
                    upper_bound format best20.
            into 
                    :kri_1 - :kri_&NROWS , 
                    :accumulate_1 - :accumulate_&NROWS , 
                    :model_1 - :model_&NROWS ,
                    :lbound_1 - :lbound_&NROWS ,
                    :ubound_1 - :ubound_&NROWS
            from 
                    &KRIoptions
            ;
    quit;

    /***** actually forecast new values *****/
    proc hpf data=&data out=_kf_hpfout lead=&lead nooutall outprocinfo=_kf_procinfo;
         id interval_start INTERVAL=&interval SETMISSING=average;
         by &allccvnames;
         %do i=1 %to &NROWS;
            forecast &&kri_&i / accumulate=&&accumulate_&i model=&&model_&i;
         %end;
    run;
    
    /***** adjust for lower and upper bounds *****/
    data _kf_hpfout; set _kf_hpfout;      
       %do i=1 %to &NROWS;
         %if &&lbound_&i ^= . %then %do;
            if &&kri_&i LT &&lbound_&i then &&kri_&i = &&lbound_&i;
         %end;
         %if  &&ubound_&i ^= . %then %do;
            if &&kri_&i GT &&ubound_&i then &&kri_&i = &&ubound_&i;
         %end;
       %end;
    run;
    
    data &foreout;
      set &data _kf_hpfout;
      by &allccvnames;      
    run;
    data &foreout; set &foreout;
      if missing(interval_end) then do;
        interval_end=intnx("&interval",interval_start,0,'end');
        interval_length = interval_end - interval_start + 1;
      end;
    run;

%DONE:

   %if &abort eq 1 %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krif_abnormal_error, noquote));
        %let status=1;
   %end;   
   %orva_status_ds( statusds=&status_KRIforecast, status=&status, reason=&reason );

    /*- delete unwanted data sets  -*/
    proc datasets library=work memtype=data nolist nodetails;
       delete _kf_: ;
    quit; 
%mend  orva_kri_forecast;




/*data KRIoptions;*/
/* kri='kri_1'; accumulate='average'; model='simple'; output;*/
/* kri='kri_2'; accumulate='average'; model='simple'; output;*/
/* kri='kri_3'; accumulate='average'; model='simple'; output;*/
/**/
/*run;*/
