/*****************************************************************************************
 *  Copyright (C), 2004
 *  SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orva_krimerge.sas                                             
 *  PURPOSE: Merges the Project's Loss dataset with KRI datasets.
 *           It is called in two different stages:
 *              1. For project creation losses are merged with the KRI datasets
 *                 from the META folder by DATE. 
 *              2. At simulation time losses are merged with the Simulation KRI datasets.
 *
 *  SUPPORT: stchri
 *
 *  PARAMETERS:
 *
 *    data    =  Name of the project's loss dataset with the following variables:
 *               INTERNAL_BL_UID, FIRST_DATE_OF_EVENT
 *    dataOut =  Name of the output dataset. Not touched if an error occurs.
 *    kriBl   =  Name of the BusinessLine KRI dataset with the following variables:
 *                  BLUID_KRI, any number of KRIs named KRI_<uid>
 *                  if merging by date:
 *                    INTERVAL_START, INTERVAL_END, INTERVAL_LENGTH
 *               If empty, &KRIlib..kribl_&interval will be used.
 *    kriFirm =  Name of the Firm KRI dataset with the following variables:
 *                  any number of KRIs named KRI_<uid>                  
 *                  if merging by date:
 *                    INTERVAL_START, INTERVAL_END, INTERVAL_LENGTH
 *               If empty, &KRIlib..krifirm_&interval will be used.
 *    cpiData  = NOT USED. 
 *    interval = project's interval.
 *    dateMerge= true means use interval dates when merging, false don't.
 *               (false is used when this is called at simulation time)
 *    renameFirm = add _F suffix to firm kri variables names.
 *    krilib  = libname for the META directory.
 *    byCCUID = merge both KRI dataset by CCUID, this is used for KRI datasets
 *              imported from the CTool to enhance the CTool models.
 *    ostat   = name of the out status dataset.  Will contain the following:
 *                status = 0 on success, 1 otherwise.
 *                reason = printable message.
 *    
 * NOTES: see sample code at end of file.
 *      
 *****************************************************************************************/
 %macro orva_krimerge( data=        ,
                       dataOut=     ,
                       kriBl=       ,
                       kriFirm=     , 
                       cpiData=     ,    
                       interval=    MONTH,
                       dateMerge=   true,
                       renameFirm=  true,
                       krilib=      ,
                       datatype=    I,
                       ostat=       status,
                       debug=       false,
                       CTool=       false);

    /*- create the status dataset upfront -*/
    %let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
    %orva_status_ds( statusds=&ostat, status=1, reason=&reason );

    %let status=0;
    %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
    %let abort=0;

    /*- check input (loss) dataset -*/
    %if &data=%str() %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, krim_specify_data_error, noquote));
        %let abort=1;
        %let reason=%sysfunc(sasmsg(&orvads, krim_missing_DATA_error, noquote));
        %goto DONE;
    %end;
             /* get the number of ccvs ------*/
    data _NULL_; 
       set &data(obs = 1);
       array ccv[*] ccvuid_:;
       call symput("numccvs", dim(ccv));
    run;
    
    /*- choose correct kriBl from META folder if not passed -*/
    %if &kriBl=%str() and &interval ne NONE %then %do;
       data _NULL_;
          kriBl= lowcase("&KRIlib..kribl_&interval");
          call symput('kriBl', kriBl);
       run;
    %end;
    /*- choose correct kriFirm from META folder if not passed -*/
    %if &kriFirm=%str() and &interval ne NONE %then %do;
       data _NULL_;
          kriFirm= lowcase("&KRIlib..krifirm_&interval");
          call symput('kriFirm', kriFirm);
       run;
    %end;
    
    /*- Do we have any Businessline Kris? -*/
    %let nobs_bl=0;
    %if &kriBl ^= %str() %then %do;
       %let dsid=%sysfunc(open(&kribl,I)); 
       %if &dsid ^= 0 %then %do;        
           %let nobs_bl=%sysfunc(attrn(&dsid,nlobs));  
           %let rc=%sysfunc(close(&dsid)); 
       %end;      
    %end; 
    %if &nobs_bl GT 0 %then %do;
       proc contents data= &kribl out=_c noprint; run;
       %let kri_cols_found= 0;
       data _NULL_; set _c end=last;
          retain found 0;   
          if upcase(substr(NAME,1,4)) eq 'KRI_' then do;
               found = found+1;                           
          end;
          if last then do;
            call symputX('kri_cols_found',found);
          end;
       run;
       %if &kri_cols_found eq 0 %then %do;
          %let nobs_bl=0;
       %end;
    %end;

    /*- Do we have any Firm level Kris? -*/
    %let nobs_firm=0;
    %if &kriFirm ^= %str() %then %do;
       %let dsid=%sysfunc(open(&kriFirm,I));
       %if &dsid ^= 0 %then %do;        
          %let nobs_firm=%sysfunc(attrn(&dsid,nlobs)); 
          %let rc=%sysfunc(close(&dsid));
       %end;  
    %end;
    %if &nobs_firm GT 0 %then %do;
       proc contents data= &kriFirm out=_c noprint; run;
       %let kri_cols_found= 0;
       data _NULL_; set _c end=last;
          retain found 0;   
          if upcase(substr(NAME,1,4)) eq 'KRI_' then do;
               found = found+1;                           
          end;
          if last then do;
               call symputX('kri_cols_found',found);
          end;
       run;
       %if &kri_cols_found eq 0 %then %do;
          %let nobs_firm=0;
       %end;
   %end;

    
    /*- make sure we get rid of old values -*/
    proc contents data=&data out=_losskri noprint; run;
    %let ALREADY_HAS_KRIS=false;
    data _null_; set _losskri;
      if (upcase(substr(NAME,1,4)) eq "KRI_") then do;
        call symput('ALREADY_HAS_KRIS','true');
        stop;
      end;
    run;
    data _losskri (compress=binary); set &data;
      %if &ALREADY_HAS_KRIS eq true %then %do;
        drop KRI_:;
      %end;
    run;


   /*- Do we have anything to do? -*/
   %if &nobs_bl EQ 0 and &nobs_firm EQ 0 %then %do;
      %let reason=%sysfunc(sasmsg(&orvads, krim_no_kris_error, noquote));

      /*- make sure final losses are sort correctly -*/
      proc sort 
                data = _losskri
                out  = &dataOut(compress=binary index=(ccvuid_0 ccvuid_1 %if &numCCVs > 2 %then ccvuid_2;) ); 
         by 
            CCUID 
            FIRST_DATE_OF_EVENT; 
      run;

      %goto DONE;
   %end;
     
    /***** Rename Firm KRIs (this used to be done in orva_kri_subset) *****/
    %if &renameFirm EQ true AND &nobs_firm GT 0 %then %do;
    
           /*** create macro variable with all firm kri names in it ***/
           %let kriNames= ;
           proc contents data=&kriFirm out=_c noprint; run;
           data _NULL_; set _c;
                  retain krivars;   
                  length krivars $ 1024;
                  if upcase(substr(NAME,1,4)) eq 'KRI_' then do;
                          found = 1;
                          krivars = trim(krivars) || ' ' || trim(name);
                          call symput( 'kriNames', krivars );      
                  end;
           run;
            
           /*** rename kris for firm with _F ***/
           data _krifirm; set &kriFirm
                 (rename=(
                         %let nkris=1;
                         %let kri=%scan(&kriNames, &nkris, ' ');
                         %do %while (&kri ne ); 
                                 &kri=&kri._F
                                 %let nkris=%eval(&nkris+1);
                                 %let kri=%scan(&kriNames, &nkris, ' ');
                         %end;   
                     ));
           run;
    %end;
    %else %if &nobs_firm GT 0 %then %do;
         data _krifirm;  set &kriFirm; 
         run;
    %end;
    
    /*** drop extraneous columns from kribl dataset ***/
    %if &nobs_bl GT 0 %then %do;     
         data _kribl;  set &kriBl;  
           /* get rid of imputed columns */
           keep %if &CTool eq true %then %do;
              OWN
                %end;
              BLUID_KRI 
              %if &dateMerge eq true %then %do;
                INTERVAL_START INTERVAL_END INTERVAL_LENGTH 
              %end;
              KRI_: ;
         run;
    %end;
    
    %let COL_BL=INTERNAL_BL_UID;
    
    /*- merge with both BusinessLine and Firm KRI values -*/
    %if &nobs_bl GT 0 and &nobs_firm GT 0 %then %do;
        
          proc sql noprint; 
               create table _workbl_ (compress=binary)
               as select * from _losskri as L left join _kribl as K
               on L.&COL_BL=K.BLUID_KRI 
               %if &dateMerge eq true %then %do;
                 and L.FIRST_DATE_OF_EVENT LE K.interval_end  
                 and L.FIRST_DATE_OF_EVENT GE K.interval_start  
               %end;   
               %if &CTool eq true %then %do;
                 and ( L.own EQ K.own OR K.own eq 0 )
               %end;
               ;   
          quit;
          data _workbl_(compress=binary); set _workbl_; 
               drop 
                  %if &dateMerge eq true %then %do;
                     INTERVAL_START INTERVAL_END INTERVAL_LENGTH
                  %end; 
               BLUID_KRI; 
          run;
          %if &dateMerge eq false and &CTool eq false %then %do;
            data _workf_(compress=binary); set _workbl_;
              if _N_ eq 1 then do; set _krifirm; end;
            run;
          %end;
          %else %do;
             proc sql noprint; 
                  create table _workf_(compress=binary)
                  as select * from _workbl_ as L left join _krifirm as K on
                       1
                    %if &dateMerge eq true %then %do;
                        and L.FIRST_DATE_OF_EVENT LE K.interval_end 
                        and L.FIRST_DATE_OF_EVENT GE K.interval_start 
                    %end;   
                    %if &CTool eq true %then %do;
                        and ( L.own EQ K.own OR K.own eq 0 );   
                    %end;
                    ;
             quit;
          %end;
    %end;
 
    /*- merge with BusinessLine KRI values only -*/
    %else %if &nobs_bl GT 0 %then %do;
         proc sql noprint; 
               create table _workf_ (compress=binary)
               as select * from _losskri as L left join _kribl as K
               on L.&COL_BL=K.BLUID_KRI 
                 %if &dateMerge eq true %then %do;
                    and L.FIRST_DATE_OF_EVENT LE K.interval_end 
                    and L.FIRST_DATE_OF_EVENT GE K.interval_start
                 %end;   
                 %if &CTool eq true %then %do;
                    and (L.own EQ K.own OR K.own eq 0)     
                 %end;
                ;   
         quit;
         data _workf_(compress=binary); set _workf_ (drop=BLUID_KRI); run;
    %end;
        
    /*- merge with Firm KRI values only -*/
    %else %if &nobs_firm GT 0 %then %do;
          proc sql noprint; 
               create table _workf_ (compress=binary)
               as select * from _losskri as L left join _krifirm as K on
                      1 
                 %if &dateMerge eq true %then %do;
                      and L.FIRST_DATE_OF_EVENT LE K.interval_end  
                      and L.FIRST_DATE_OF_EVENT GE K.interval_start  
                 %end; 
                 %if &CTool eq true %then %do;
                      and (L.own EQ K.own OR K.own eq 0)     
                 %end;
               ;                 
          quit;
    %end;
    
    data _losskri (compress=binary); set _workf_;    
       %if &dateMerge eq true %then %do; 
          drop interval_start interval_end interval_length;
       %end;
    run; 

     /*- make sure final losses are sorted correctly -*/
     proc sort 
                data = _losskri 
                out  = &dataOut(compress=binary index=(ccvuid_0 ccvuid_1 %if &numCCVs > 2 %then ccvuid_2;) )
                ; 
            
        by 
            CCUID 
            FIRST_DATE_OF_EVENT; 
    run;
 
%DONE:
    %if &abort %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, krim_abnormal_error, noquote));
      %let status = 1;
    %end;
        
    %orva_status_ds( statusds=&ostat, status=&status, reason=&reason );

    /*- delete unwanted data sets -*/
    %if &debug eq false %then %do;
       proc datasets library=work memtype=data nolist nodetails;
          delete _c _workbl_ _workf_ _kribl _krifirm _losskri;
       quit;
    %end;
 
%mend orva_krimerge;

/***************************************
options mprint;
%orva_krimerge( data= VARPRJ.lossi,
               dataOut= losskri,
               kriBl=  kriblprj,
               kriFirm= krifirmprj, 
               ostat= status );
****************************************/
