/*****************************************************************************************
 * Copyright (C), 2008
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    ORVA_rare_scen_sim.sas                                             
 *  PURPOSE: 
 *
 *  SUPPORT: jariou
 *
 *  PARAMETERS:
 *      nreps      = number of scenario replications
 *      scenUids   = space delimited list of Rare Event Scenarios to use.
 *      freqNames  = Dataset with frequency distributions information
 *      sevNames   = Dataset with severity distributions information
 *      outScenSim = output simulation dataset for Rare Event Scenarios.
 *      debug      = true or false (defaulted to false) Indicates whether the user gets extra info
 *
 *  EXAMPLE:
 *        %orva_rare_scen_sim(scenUids=1034 2304, outScenSim=VARPRJ.rareScenSim);
 *
 *  Enumerations:
 *                EVENT_TYPE     
 *                               Certain
 *                               Independent
 *                               Correlated
 *
 *                SCENARIO_SPEC 
 *                               ModelBased
 *                               BucketBased          
 *****************************************************************************************/
%macro ORVA_rare_scen_sim( 
                           nreps      = 1000,
                           seedData   = ,
                           scenUids   = ,
                           freqNames  = ,
                           sevNames   = ,
                           outScenSim = ,
                           debug      = false
                         );

   data &outScenSim;
      set _NULL_;
   run;

   /*****************************************/
   /* Get the seed information: BEGIN       */
   /*****************************************/
   /* seedData must be specified */
   %if "&seedData" EQ "" %then %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, losim_provide_seedData_error, noquote));
      %let abort  = 1;
      %let reason = %sysfunc(sasmsg(&orvads, losim_missing_seedData_error, noquote));
      %goto DONE;
   %end;
   /*****************************************/
   /* Get the seed information: END         */
   /*****************************************/
 
   %let nscen = 1;
   %let scen  = %scan(&scenUids, &nscen, ' ');
   
   %do %while(&scen NE %str()); /* Level = 1 */
      proc sql noprint;
         SELECT 
                  put( Seed, BEST20. )
         INTO 
                  :Seed_1-:Seed_3
         FROM 
                  &seedData
         WHERE 
                  ScenarioUID = &scen 
         AND 
                  Host = 1
                  ;
      quit;

      /* These macros will be filled in by orv_scenarioIdentifyType */
      %let scenData              = %str();
      %let scenGroupData         = %str();
      %let scenPlotBreakdownData = %str();
      %let scenEventType         = %str();
      %let scenSpecificationType = %str();

      %orv_scenarioIdentifyType(
                                   inScenarioID                 = &scen,
                                   outScenarioData              = scenData,
                                   outScenarioGroupData         = scenGroupData,
                                   outScenarioPlotBreakdownData = scenPlotBreakdownData,
                                   outScenarioEventType         = scenEventType,
                                   outScenarioSpecificationType = scenSpecificationType
                                   );

      %if %upcase(&debug) EQ TRUE %then %do;
         %orva_scenarioSimDisplayDebugInfo(
                                               Scenario      = &scen, 
                                               Data          = &scenData , 
                                               Group         = &scenGroupData, 
                                               PlotBreakdown = &scenPlotBreakdownData, 
                                               EventType     = &scenEventType, 
                                               Specification = &scenSpecificationType
                                               );
      %end; /* justCheckScenarios = TRUE */

      %orv_scenarioSimulate(
                              ScenarioID                = &scen,
                              ScenarioData              = &scenData,
                              ScenarioGroupData         = &scenGroupData,
                              ScenarioPlotBreakdownData = &scenPlotBreakdownData,
                              ScenarioEventType         = &scenEventType,
                              ScenarioSpecificationType = &scenSpecificationType
                              ); 

      %orvu_append(
                     base = &outScenSim, 
                     new  = _orss_&scen
                     );                   

      %let nscen = %eval(&nscen + 1);
      %let scen  = %scan(&scenUids, &nscen, ' ');
   %end; /* Level = 1 */       
%DONE:

   %orvu_delete_data( data = _orss_: );
%mend ORVA_rare_scen_sim;

%macro orv_scenarioSimIndependentModel(
                                       simCount = , 
                                       dataIn   = , 
                                       seed     = , 
                                       dataOut  =
                                       );
   data &dataOut;
      set &dataIn;   
        call streaminit(&SEED.);

      do _rep_ = 1 to &simCount;
         /* First, simulate number of events */
         Event_Count = FreqSim(Frequency_Dist, &theFreqSimCallParams);

         /* Now see how many losses will be caused by those events */
         if PROBABILITY = 1 then do;
            Loss_Count  = Event_Count;
         end;
         else if (PROBABILITY = 0) OR (Event_Count = 0) then do;
            Loss_Count = 0;
         end;
         else do;
            Loss_Count = rand("BINOMIAL",PROBABILITY,Event_Count);
         end;
         
         /* Now accumulate the total amount of money those losses will cost */
         Loss_Amount = 0;  

         do i = 1 to Loss_Count;
            Loss_Amount = Loss_Amount + SevSim(Severity_Dist,&theSevSimCallParams );
         end;
         
         output;
      end;

      keep  
            SCENARIO_UID 
            SCENARIO_EVENT_UID 
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT 
            ;
   run; 

   /***********************************************************/
   /* For the same CCUID, many SCENARIO_EVENT_UID's may exist */
   /* We need to combine those losses caused by those events  */
   /* into a single number per CCUID                          */
   /***********************************************************/

   /* First sort in the proper order to be manipulated */
   proc sort 
            data = &dataOut;
      BY 
         CCUID 
         _rep_ 
         ;
   run;

   /* Do the accumulation per CCUID */
   data &dataOut;
      set &dataOut;
      BY 
         CCUID 
         _rep_
         ;
      retain 
            Total_Event_Count 0 
            Total_Loss_Count  0 
            Total_Loss_Amount 0
            ;

      if first._rep_ then do;
         Total_Event_Count = 0;
         Total_Loss_Count = 0;
         Total_Loss_Amount = 0;
      end;
      
      Total_Event_Count = Total_Event_Count  + Event_Count ;
      Total_Loss_Count = Total_Loss_Count  + Loss_Count ;
      Total_Loss_Amount = Total_Loss_Amount  + Loss_Amount ;

      if last._rep_ then do;
         Event_Count = Total_Event_Count ;
         Loss_Count  = Total_Loss_Count;
         Loss_Amount = Total_Loss_Amount;
         output;
      end;

      keep  
            SCENARIO_UID 
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT
            ;
   run;
%mend orv_scenarioSimIndependentModel;

%macro orv_scenarioSimIndependentBucket(
                                          simCount = , 
                                          dataIn   = , 
                                          seed     = , 
                                          dataOut  =
                                          );
   /* I need to figure out the size fo the array*/
   data _NULL_;
      set &dataIn ;
      retain maxBucketCount 0;
      array buckWeights bucketwgt:;
      maxBucketCount = max( maxBucketCount , nBuckets);
      call symputX( "buckWeightsSize", maxBucketCount );
      call symputX( "buckSize", maxBucketCount + 1 );
   run;   

   data &dataOut;
      set &dataIn;   

      array    BucketWeights[*] bucketwgt1-bucketwgt&buckWeightsSize;
      array tmpBucketWeights[&buckWeightsSize] _temporary_;

      array    Buckets[*] bucket0-bucket&buckWeightsSize;
      array tmpBuckets[&buckSize] _temporary_;

      do i = 1 to &buckWeightsSize;
         tmpBucketWeights[i] = BucketWeights[i];
         tmpBuckets[i]       = Buckets[i];
      end;

      tmpBuckets[&buckSize]  = Buckets[&buckSize];
      call NormalizeWeights( tmpBucketWeights, nBuckets );

      call BucketFreqSimDef(  
                              Frequency_Mean,
                              Frequency_Min,
                              Frequency_Max,
                              Frequency_dist,
                              FParam_1,
                              FParam_2
                              );
      call streaminit(&SEED.);

      do _rep_ = 1 to &simCount;
         /* First, simulate number of events */
         Event_Count = FreqSim(
                                 Frequency_Dist, 
                                 &theFreqSimCallParams
                                 );

         /* Now see how many losses will be caused by those events */
         if PROBABILITY = 1 then do;
            Loss_Count = Event_Count;
         end;
         else if (Event_Count = 0) OR (PROBABILITY = 0) then do;
            Loss_Count = 0;
         end;
         else do;
            Loss_Count = rand(
                              "BINOMIAL",
                              PROBABILITY,
                              Event_Count
                              );
         end;

         /* Now accumulate the total amount of money those losses will cost */
         Loss_Amount = 0;  

         do i = 1 to Loss_Count;
            Loss_Amount = Loss_Amount + BucketSevSim( 
                                                      &seed,
                                                      tmpBuckets,
                                                      tmpBucketWeights,
                                                      nBuckets
                                                      );
         end;
         
         output;
      end;

      keep  
            SCENARIO_UID 
            SCENARIO_EVENT_UID 
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT 
            ;
   run; 

   /***********************************************************/
   /* For the same CCUID, many SCENARIO_EVENT_UID's may exist */
   /* We need to combine those losses caused by those events  */
   /* into a single number per CCUID                          */
   /***********************************************************/

   /* First sort in the proper order to be manipulated */
   proc sort 
            data = &dataOut;
      BY 
         CCUID 
         _rep_
         ;
   run;

   /* Do the accumulation per CCUID */
   data &dataOut;
      set &dataOut;
      BY 
         CCUID 
         _rep_
         ;
      retain 
               Total_Event_Count 0 
               Total_Loss_Count  0 
               Total_Loss_Amount 0
               ;
      if first._rep_ then do;
         Total_Event_Count = 0;
         Total_Loss_Count = 0;
         Total_Loss_Amount = 0;
      end;
      
      Total_Event_Count = Total_Event_Count  + Event_Count ;
      Total_Loss_Count = Total_Loss_Count  + Loss_Count ;
      Total_Loss_Amount = Total_Loss_Amount  + Loss_Amount ;

      if last._rep_ then do;
         Event_Count = Total_Event_Count;
         Loss_Count  = Total_Loss_Count;
         Loss_Amount = Total_Loss_Amount;
         output;
      end;

      keep  
            SCENARIO_UID 
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT
            ;
   run;
%mend orv_scenarioSimIndependentBucket;

%macro orv_scenarioSimCorrelatedModel(
                                       simCount = , 
                                       dataIn   = , 
                                       groups   = , 
                                       seed1    = , 
                                       seed2    = , 
                                       seed3    = , 
                                       dataOut  =
                                       );
   /* Keep only one copy of SENARIO_EVENT_UID and the correponding frequency information */ 
   proc sort 
            data = &dataIn 
            out  = _orss_event_ids(keep = 
                                          SCENARIO_UID
                                          SCENARIO_EVENT_UID 
                                          FREQUENCY_DIST
                                          FPARAM_:
                                    )
            nodupkey
            ;
      by SCENARIO_EVENT_UID; 
   run;

   /* Simulate the number of events for each Event_ID */
   data _orss_Events;
      set _orss_event_ids;
      call streaminit(&SEED1.);

      do _rep_ = 1 to &simCount;
         Event_Count = FreqSim(
                               Frequency_Dist, 
                               &theFreqSimCallParams
                               );
         output;
      end;

      keep 
            Scenario_Event_UID 
            Scenario_UID 
            _rep_ 
            Event_Count
            ;
   run;

   /* Get a unique copy of all the groups with their probabilities*/
   proc sort 
            data = &groups 
            out  = _orss_group_ids 
            nodupkey
            ;
      BY 
         Event_ID 
         Group_ID
         ;
   run;

   /* Merge simulated event count with group information */
   proc sql noprint;
      CREATE TABLE
                     _orss_events_with_group_info( keep = 
                                                            Event_ID 
                                                            Scenario_UID 
                                                            _rep_ 
                                                            Event_Count 
                                                            Group_ID 
                                                            Group_Prob
                                                            )
      AS 
      SELECT      
                     * 
      FROM
                     _orss_events a, 
                     _orss_group_ids b
      WHERE 
                     a.SCENARIO_EVENT_UID = b.EVENT_ID
                     ;
   quit;


   /* SCENARIO_UID EVENT_ID _REP_ EVENT_COUNT GROUP_ID GROUP_PROB */
   proc sort   
            data = _orss_events_with_group_info
            out  = _orss_events_sorted
            ;
      BY 
         Event_ID 
         _rep_ 
         Group_ID 
         ;
   run;
      
   /* Simulate how many events create a loss in any given Group */
   /* SCENARIO_UID EVENT_ID _REP_ EVENT_COUNT GROUP_ID GROUP_PROB LOSS_COUNT */
   data _orss_group_loss_counts;
      set _orss_Events_sorted;
      BY 
         Event_ID 
         _rep_
         ;
      retain 
            left_Event_Count 0
            left_Prob        1
            ;
        call streaminit(&SEED2.);

      if first._rep_ then do;
         left_Event_Count = Event_Count;
         left_Prob        = 1;
      end;

      if left_Event_Count > 0 then do;
         Loss_Count       = rand("BINOMIAL", min(Group_Prob / left_Prob, 1), left_Event_Count);
         left_Event_Count = max(left_Event_Count - Loss_Count, 0);
         left_Prob        = max(left_Prob - Group_Prob, 0);
      end;
      else do;
         Loss_Count = 0;
      end;

      drop 
            left_event_count
            left_Prob
            ;
   run;

   /* Add the CCUID info from the groups dataset */
   proc sql noprint;
      CREATE TABLE 
                     _orss_ccuid_loss_count( keep = 
                                             SCENARIO_UID
                                             EVENT_ID 
                                             GROUP_ID 
                                             _rep_ 
                                             EVENT_COUNT 
                                             LOSS_COUNT 
                                             CCUID
                                             )
      AS 
      SELECT 
                     * 
      FROM
                     _orss_group_loss_counts a, 
                     &groups b
      WHERE 
                     a.GROUP_ID = b.GROUP_ID
                     ;
   quit;

   /* Incorporate the severity information for each CCUID */
   proc sql noprint;
      CREATE TABLE 
                     _orss_counts_with_sev( keep = 
                                                   SCENARIO_UID
                                                   EVENT_ID 
                                                   GROUP_ID 
                                                   _rep_ 
                                                   EVENT_COUNT 
                                                   LOSS_COUNT 
                                                   CCUID
                                                   SEVERITY_DIST
                                                   PARAM_:
                                                   )
      AS 
      SELECT 
                     * 
      FROM  
                     _orss_ccuid_loss_count a, 
                     &dataIn b
      WHERE 
                     (a.EVENT_ID = b.SCENARIO_EVENT_UID)
            AND
            (a.CCUID = b.CCUID)
            ;
   quit;

   /* Sort so the accumulation by CCUID can be done */
   proc sort 
            data  = _orss_counts_with_sev 
             out  = _orss_sev;
      BY 
         CCUID 
         _REP_ 
         EVENT_ID 
         ;
   run;

   /* Accumulate the total count by CCCUID                     */
   /* Then simulate and accumulate that many individual losses */
   data &dataOut;
      set _orss_sev;
      BY 
         CCUID 
         _rep_ 
         EVENT_ID
         ;
      retain 
            Total_Event_Count 0 
            Total_Loss_Count  0
            Partial_Loss_Count 0
            Loss_Amount        0
            ;
      call streaminit(&SEED3.);

      if first._rep_ then do;
         Total_Event_Count = 0;
         Total_Loss_Count  = 0;
         Loss_Amount       = 0;
      end;
      
      if first.event_id then do;
         Partial_Loss_Count  = 0;
      end;

      Partial_Loss_Count = Partial_Loss_Count  + Loss_Count ;

      if last.event_id then do;
         do j = 1 to Partial_Loss_Count;
            Loss_Amount = Loss_Amount + SevSim(Severity_Dist,&theSevSimCallParams );
         end;

         Total_Loss_Count = Total_Loss_Count + Partial_Loss_Count;
      end;

      if last._rep_ then do;
         Loss_Count = Total_Loss_Count;         
         output;
      end;

      keep 
            SCENARIO_UID 
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT 
            ;
   run;
%mend orv_scenarioSimCorrelatedModel;

%macro orv_scenarioSimCorrelatedBucket(
                                       simcount = , 
                                       dataIn   = , 
                                       groups   = , 
                                       seed1    = , 
                                       seed2    = , 
                                       seed3    = , 
                                       dataOut  =
                                       );
   /* I need to figure out the size fo the array*/
   data _NULL_;
      set &dataIn ;
      retain maxBucketCount 0;
      array buckWeights bucketwgt:;
      maxBucketCount = max( maxBucketCount , nBuckets);
      call symputX( "buckWeightsSize", maxBucketCount );
      call symputX( "buckSize", maxBucketCount + 1 );
   run;

   /* Keep only one copy of SENARIO_EVENT_UID and the correponding frequency information */ 
   proc sort 
            data = &dataIn 
            out  = _orss_event_ids(keep = 
                                          SCENARIO_UID
                                          SCENARIO_EVENT_UID 
                                          FREQUENCY_:
                                    )
            nodupkey
            ;
      BY 
         SCENARIO_EVENT_UID; 
   run;

   /* Simulate the number of events for each Event_ID */
   data _orss_Events;
      set _orss_event_ids;
      retain 
            FParam_1 0
            FParam_2 0
            ;
      call streaminit(&SEED1.);

      call BucketFreqSimDef(  
                              Frequency_Mean,
                              Frequency_Min,
                              Frequency_Max,
                              Frequency_dist,
                              FParam_1,
                              FParam_2
                              );

      do _rep_ = 1 to &simCount;
         Event_Count = FreqSim(
                               Frequency_Dist, 
                               &theFreqSimCallParams
                               );
         output;
      end;

      keep 
            Scenario_Event_UID 
            Scenario_UID 
            _rep_ 
            Event_Count
            ;
   run;

   /* Get a unique copy of all the groups with their probabilities*/
   proc sort 
            data = &groups 
            out  = _orss_group_ids(keep = 
                                          Event_ID 
                                          Group_ID
                                          Group_Prob
                                          )
            nodupkey
            ;
      by 
         Event_ID 
         Group_ID
            ;
   run;

   /* Merge simulated event count with group information */
   proc sql noprint;
      CREATE TABLE 
                     _orss_events_with_group_info (keep = 
                                                            Event_ID 
                                                            Scenario_UID 
                                                            _rep_ 
                                                            Event_Count 
                                                            Group_ID 
                                                            Group_Prob
                                                            )
      AS
      SELECT
                     * 
      FROM
                     _orss_events a, 
                     _orss_group_ids b
      WHERE 
                     a.SCENARIO_EVENT_UID = b.EVENT_ID
                     ;
   quit;


   /* SCENARIO_UID EVENT_ID _REP_ EVENT_COUNT GROUP_ID GROUP_PROB */
   proc sort   
            data = _orss_events_with_group_info
            out  = _orss_events_sorted
            ;
      by 
         Event_ID 
         _rep_ 
         Group_ID 
         ;
   run;
      
   /* Simulate how many events create a loss in any given Group */
   /* I need to simulate a multinomial distribution where the   */
   /* number of trials is Event_Count and the probabilities are */
   /* the group probabilities. This is accomplished by keeping  */ 
   /* a running count of how many events are left and what is   */
   /* the remaining probability left, i.e. I simulate a series  */
   /* of conditional Binomial distributions                     */
   /* SCENARIO_UID EVENT_ID _REP_ EVENT_COUNT GROUP_ID GROUP_PROB LOSS_COUNT */
   data _orss_group_loss_counts;
      set _orss_Events_sorted;
      by 
         Event_ID 
         _rep_
         ;
      retain 
            left_Event_Count 0
            left_Prob        1
            ;
      call streaminit(&SEED2.);

      if first._rep_ then do;
         left_Event_Count = Event_Count;
         left_Prob        = 1;
      end;

      if left_Event_Count > 0 then do;
         Loss_Count       = rand("BINOMIAL", min(Group_Prob / left_Prob, 1), left_Event_Count);
         left_Event_Count = max(left_Event_Count - Loss_Count, 0);
         left_Prob        = max(left_Prob - Group_Prob, 0);
      end;
      else do;
         Loss_Count = 0;
      end;

      drop 
           left_event_count
           left_Prob
           ;
   run;

   /* Add the CCUID info from the groups dataset */
   proc sql noprint;
      CREATE TABLE 
                     _orss_ccuid_loss_count(keep= 
                                             SCENARIO_UID
                                             EVENT_ID 
                                             GROUP_ID 
                                             _rep_ 
                                             EVENT_COUNT 
                                             LOSS_COUNT 
                                             CCUID
                                             )
      AS 
      SELECT 
                  * 
      FROM
                  _orss_group_loss_counts a, 
                  &groups b
      WHERE 
                  a.GROUP_ID = b.GROUP_ID
                  ;
   quit;

   /* Incorporate the severity information for each CCUID */
   proc sql noprint;
      CREATE TABLE 
                     _orss_counts_with_bucket( keep = 
                                                      CCUID
                                                   SCENARIO_UID
                                                   EVENT_ID 
                                                   GROUP_ID 
                                                   _rep_ 
                                                   EVENT_COUNT 
                                                   LOSS_COUNT 
                                                      nbuckets
                                                   BUCKET:
                                                   )
      AS 
      SELECT 
                     * 
      FROM
                     _orss_ccuid_loss_count a, 
                     &dataIn b
      WHERE 
                     ( a.EVENT_ID = b.SCENARIO_EVENT_UID )
            AND
            (a.CCUID = b.CCUID)
            ;
   quit;

   /* Sort so the accumulation by CCUID can be done */
   proc sort 
            data = _orss_counts_with_bucket 
            out  = _orss_sev
            ;
      by 
         CCUID 
         _REP_ 
         EVENT_ID 
         ;
   run;

   /* Accumulate the total count by CCCUID                     */
   /* Then simulate and accumulate that many individual losses */
   data &dataOut;
      set _orss_sev;
      BY 
         CCUID 
         _rep_ 
         EVENT_ID
         ;
      retain 
               Total_Event_Count 0 
               Total_Loss_Count  0
               Partial_Loss_Count 0
               Loss_Amount        0
               ;

      array    BucketWeights[*] bucketwgt1-bucketwgt&buckWeightsSize;
      array tmpBucketWeights[&buckWeightsSize] _temporary_;

      array    Buckets[*] bucket0-bucket&buckWeightsSize;
      array tmpBuckets[&buckSize] _temporary_;

      do i = 1 to &buckWeightsSize;
         tmpBucketWeights[i] = BucketWeights[i];
         tmpBuckets[i]       = Buckets[i];
      end;

      tmpBuckets[&buckSize]  = Buckets[&buckSize];
      call NormalizeWeights( tmpBucketWeights, nBuckets );
      call streaminit(&SEED3.);

      if first._rep_ then do;
         Total_Event_Count = 0;
         Total_Loss_Count  = 0;
         Loss_Amount       = 0;
      end;
      
      if first.event_id then do;
         Partial_Loss_Count  = 0;
      end;

      Partial_Loss_Count = Partial_Loss_Count  + Loss_Count ;

      if last.event_id then do;
         do j = 1 to Partial_Loss_Count;
               Loss_Amount = Loss_Amount + BucketSevSim( 
                                                         &SEED3.,
                                                         tmpBuckets,
                                                         tmpBucketWeights,
                                                         nBuckets
                                                         );
         end;

         Total_Loss_Count = Total_Loss_Count + Partial_Loss_Count;
      end;

      if last._rep_ then do;
         Loss_Count  = Total_Loss_Count;
         output;
      end;

      keep 
            SCENARIO_UID 
            CCUID 
            _REP_ 
            EVENT_COUNT 
            LOSS_COUNT 
            LOSS_AMOUNT 
            ;
   run;
%mend orv_scenarioSimCorrelatedBucket;

%macro orv_scenarioIdentifyType(
                                 inScenarioID                 = ,
                                 outScenarioData              = ,
                                 outScenarioGroupData         = ,
                                 outScenarioPlotBreakdownData = ,
                                 outScenarioEventType         = ,
                                 outScenarioSpecificationType = 
                                 );
   %orvu_find_dataset(  
                        nameIn  = Scenario_&inScenarioID, 
                        libs    = VARPRJ VARPRJ_M,
                        nameOut = &outScenarioData
                        );

   %orvu_find_dataset( 
                        nameIn  = ScenarioGroup_&inScenarioID, 
                        libs    = VARPRJ VARPRJ_M,
                        nameOut = &outScenarioGroupData
                        );   

   %orvu_find_dataset( 
                        nameIn  = Scenario_PlotBrkDwn_&inScenarioID, 
                        libs    = VARPRJ VARPRJ_M,
                        nameOut = &outScenarioPlotBreakdownData
                                                      );

   %if &outScenarioData NE %str() %then %do; 
      data _null_;
         set &&&outScenarioData; 
         if _N_ = 1;

         if EVENT_TYPE = 0 then do;
            call symput ("&outScenarioEventType", "Certain");
         end;

         if EVENT_TYPE = 1 then do;
            call symput ("&outScenarioEventType", "Independent");
         end;

         if EVENT_TYPE = 2 then do;
            call symput ("&outScenarioEventType", "Correlated");
         end;
      run;
   %end;
   %else %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, rare_scen_no_exist_no_sim_error, noquote, &inScenarioID));
   %end;

   %if &&&outScenarioPlotBreakdownData = %str() %then %do; 
      %let &outScenarioSpecificationType. = ModelBased;
   %end;
   %else %do;
      %let &outScenarioSpecificationType. = BucketBased;
   %end; 
%mend orv_scenarioIdentifyType;

%macro orva_scenarioSimDisplayDebugInfo(
                                          Scenario      = ,
                                          Data          = ,
                                          Group         = ,
                                          PlotBreakdown = ,
                                          EventType     = ,
                                          Specification = 
                                          );
   %put ;
   %put |      *************************************************;
   %put |      *;
   %put |      *  Scenario           =  &Scenario;
   %put |      *;
   %put |      *  ----------------------------------------------;
   %put |      *;
   %put |      *  ScenarioData       =  &Data;
   %put |      *  ScenarioGroup      =  &Group;
   %put |      *  ScenarioPlotBrkDwn =  &PlotBreakdown;
   %put |      *;
   %put |      *  ----------------------------------------------;
   %put |      *;
   %put |      *  EVENT_TYPE         =  &EventTyper;
   %put |      *  SCENARIO_SPEC      =  &Specification;
   %put |      *;
   %put |      *************************************************;
   %put ;
%mend orva_scenarioSimDisplayDebugInfo;

%macro orv_scenarioSimulate(
                              ScenarioID                = ,
                              ScenarioData              = ,
                              ScenarioGroupData         = ,
                              ScenarioPlotBreakdownData = ,
                              ScenarioEventType         = ,
                              ScenarioSpecificationType = 
                              );

   %if  &scenarioEventType = Certain %then %do; 
      %orva_sim_scenario_certain(
                                    simCount = &nReps, 
                                    dataIn   = &scenarioData,
                                    scenSpec = %substr(&scenarioSpecificationType, 1, 1), 
                                    seed     = &seed_1, 
                                    scenKind = RARE,
                                    dataOut  = _orss_&scenarioID
                                    ); 
   %end; /* EVENT_TYPE = Certain */
   %else %if &scenarioEventType = Independent %then %do; 
      %if &scenarioSpecificationType = ModelBased %then %do; 
         %orv_scenarioSimIndependentModel( 
                                             simCount = &nReps,
                                             dataIn   = &scenarioData,
                                             Seed     = &Seed_1,
                                             dataOut  = _orss_&scenarioID
                                             );
      %end; /* SCENARIO_SPEC = ModelBased */
      %else %if &scenarioSpecificationType = BucketBased %then %do; 
         %orv_scenarioSimIndependentBucket(   
                                               simCount = &nReps,
                                               dataIn   = &scenarioPlotBreakdownData,
                                               Seed     = &Seed_1,
                                               dataOut  = _orss_&scenarioID
                                               );
      %end; /* SCENARIO_SPEC = BucketBased */
      %else %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, rare_scen_spec_missing_no_sim_error, noquote, &scenarioID));
      %end;
   %end; /* EVENT_TYPE = Independent */
   %else %if  &scenarioEventType = Correlated %then %do; 
      %if &scenarioSpecificationType = ModelBased %then %do; 
         %orv_scenarioSimCorrelatedModel(  
                                            simCount = &nReps,
                                            dataIn   = &scenarioData,
                                            groups   = &scenarioGroupData,
                                            Seed1    = &Seed_1,
                                            Seed2    = &Seed_2,
                                            Seed3    = &Seed_3,
                                            dataOut  = _orss_&scenarioID
                                            );
      %end; /* SCENARIO_SPEC = ModelBased */
      %else %if &scenarioSpecificationType = BucketBased %then %do; 
         %orv_scenarioSimCorrelatedBucket( 
                                             simCount = &nReps,
                                             dataIn   = &scenarioPlotBreakdownData,
                                             groups   = &scenarioGroupData,
                                             Seed1    = &Seed_1,
                                             Seed2    = &Seed_2,
                                             Seed3    = &Seed_3,
                                             dataOut  = _orss_&scenarioID
                                             );
      %end; /* SCENARIO_SPEC = BucketBased */
      %else %do;
         %put &orvar_error %sysfunc(sasmsg(&orvads, rare_scen_spec_missing_no_sim_error, noquote, &scenarioID));
      %end;
   %end; /* EVENT_TYPE = Correlated */
   %else %do;
      %put &orvar_error %sysfunc(sasmsg(&orvads, rare_scen_event_type_missing_no_sim_error, noquote, &scenarioID));
   %end;
%mend orv_scenarioSimulate;
