/*************************************************************************************
 * Copyright (C), 2007
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME   : orva_readfitopts                                             
 * PURPOSE: Read fitting options from dataset           
 * SUPPORT: stchri
 *
 * PARAMETERS :
 *
 *   fitoptsData = fitting options dataset.
 *   fittype     = SEV,FREQ
 *   type        = THRESHOLD, ADVANCED, DIST, FITMEASURE, SEVUSERINPUT, KRI
 *   datatype    = I,C,P
 *   ccUid       = ccuid for cell we are fitting. If ccUid is missing, only project-level 
 *                 options are used to set the requisite macro variables.
 *                 Otherwise, this macro attempts to find cell-level options for the value of ccUid.  
 *                 (if none are found, then the project-level options are used.)
 *
 * OUTPUT (macro variables): 
 *   type=THRESHOLD:
 *        for fittype=SEV the following macro variables are set:
 *           threshold
 *           gofThreshold
 *           defaultThreshold
 *        for fittype=FREQ and type=THRESHOLD the following macro variables are set:
 *           Freq_Use_Data_Threshold 
 *           Freq_Missing_Default 
 *           Freq_Zero_Count_Default  
 *           Freq_UseStratumMean_Default 
 *           Freq_UseStratumMean_UserSpec
 *           Freq_Cut_Off 
 *           Freq_Keep_Larger_Thresholds
 * 
 *  type=DIST:
 *       for fittype=SEV the following macro variable is set:
 *           dists
 *       for fittype=FREQ the following macro variable is set:
 *           fdists
 *  type=KRI:
 *       for fittype=SEV the following macro variable is set:
 *           Sev_Kris_Suggested
 *           Sev_Kris_Forced
 *       for fittype=FREQ the following macro variable is set:
 *           Freq_Kris_Suggested
 *           Freq_Kris_Forced
 *
 *  type=POBSERV: (both fittypes)
 *       Probability_Observed
 *       Source_Method
 *       Source_Type
 *       Source_Uid           
 *
 *  type=FITMEASURE:
 *      for fittype=SEV the following macro variable is set:
 *        Sev_FitMeasures=;
 *        Sev_FavFitMeasure=;
 *      for fittype=FREQ the following macro variable is set:
 *        Freq_FitMeasures=;
 *        Freq_FavFitMeasure=;
 **************************************************************************************/
%macro orva_readfitopts( fitoptsData=, fittype=SEV, type=THRESHOLD, datatype=I, ccUid= );

    /* Set CCUID to MISSING for project level options */
    %if &ccuid EQ %str() %then %do;
        %let ccuid=.;
    %end;
    
    /* 
     * Note that THRESHOLD and POBSERV options are datatype (I,C,P) dependent
     * when we are dealing with project level options, but the rest are not.
     * In this second case set datatype to EMPTY.
     */
    %if &ccuid EQ . %then %do;
       %if &type ne THRESHOLD and &type ne POBSERV %then %do;
          %let datatype = %str();
       %end;
    %end;
    
    /* Get our optiont prefix based on fitting TYPE */
    %if &fittype EQ SEV %then %do;
        %let prefix=Sev_;
        %let prefixL=4;
    %end;
    %else %do;
        %let prefix=Freq_;
        %let prefixL=5;
    %end;   
    
    /* Finally, handle option type requested */
    %if &type EQ THRESHOLD %then %do;
        %read_threshold_options;
    %end;
    %else %if &type EQ POBSERV %then %do;
        %read_pobserv_options;
    %end;
    %else %if &type EQ DIST %then %do;
        %read_dist_options;
    %end;
    %else %if &type EQ SEVUSERINPUT %then %do;
        %read_sui_options;
    %end;
    %else %if &type EQ KRI %then %do;
        %read_kri_options;
    %end;
    %else %if &type EQ INDICATOR %then %do;
        %read_indicator_options;
    %end;
    %else %if &type EQ KRICONTROL %then %do;
        %read_kricontrol_options;
    %end;
    %else %if &type EQ FITMEASURE %then %do;
        %read_fitmeasure_options;
    %end;
    %else %if &type EQ ADVANCED %then %do;
        %read_advanced_options;
    %end;

%mend orva_readfitopts;

/**
 * Generic macro to retrieve fit options by
 *   type, ccuid, datatype, and prefix
 * datatype is ignored if ccuid = .
 */
%macro get_options(flag=options_exist,type=,datatype=,ccuid=,usePrefix=true);
   data _NULL_; set &fitoptsData; 
         where CCUID = &ccuid AND type EQ "&type" 
            %if &datatype ne %str() %then %do;
              AND datatype EQ "&datatype"
            %end;
            %if &usePrefix EQ true %then %do;
              AND substr(name,1,&prefixL)="&prefix"
            %end;
         ;
         if cvalue NE "" then do;
            call symput(name,trim(left(cvalue)));
         end;
         else do;
            call symput(name,trim(left(put(nvalue,BEST.))));
         end;
         call symput("&flag", "true");
    run;
%mend get_options;

%macro read_kricontrol_options;
      %if &fittype EQ SEV %then %do;
        %let Sev_Use_Kris=false;
        %let Sev_Selection_Method_Kris=Stepwise;
      %end;
      %else %do;
        %let Freq_Use_Kris=false;
        %let Freq_Selection_Method_Kris=Stepwise;
      %end;
      /*** read options as specified by ccUid ***/
      %let options_exist=false;
      %get_options(flag=options_exist,type=&type,datatype=&datatype,ccuid=&ccuid,usePrefix=true);
      /*** if that failed and ccUid was specified, try project-level instead ***/ 
      %if &options_exist EQ false AND &ccuid NE . %then %do;
         %get_options(flag=options_exist,type=&type,ccuid=.,usePrefix=true);
      %end;
%mend read_kricontrol_options;

%macro read_sui_options;
      %let SevUserInput= true;
      %let TailBoundaryPercent = 95;
      %let TailBoundaryValue   = 1000000;
      %let TailBoundaryChoice  = TailBoundaryPercent;
      /*** read options as specified by ccUid ***/
      %let options_exist=false;
      %get_options(flag=options_exist,type=&type,datatype=&datatype,ccuid=&ccuid,usePrefix=false);
      /*** if that failed and ccUid was specified, try project-level instead ***/ 
      %if &options_exist EQ false AND &ccuid NE . %then %do;
         %get_options(flag=options_exist,type=&type,ccuid=.,usePrefix=false);
      %end;
%mend read_sui_options;

/*************************************************************
 * Macro to read the POBSERV options 
 ************************************************************/
%macro read_pobserv_options;    
    /*** default values we expect to find ****/
    %let Probability_Observed=.;
    %let Source_Method=.;
    %let Source_Type=.;
    %let Source_Uid=0;           
    
    /*** get probability of observability using ccuid specified ***/
    %let options_exist=false;
    %get_pobserv(flag=options_exist,datatype=&datatype,ccuid=&ccuid);
    
    /*** if not found and cell specified try project-level ***/
    %if &options_exist EQ false AND &ccuid NE . %then %do;
       %get_pobserv(flag=options_exist,datatype=&datatype,ccuid=.);
    %end;
%mend read_pobserv_options;
%macro get_pobserv(flag=options_exist,datatype=,ccuid=);
   data _NULL_; set &fitoptsData; 
         where CCUID = &ccuid AND type EQ "&type" AND datatype EQ "&datatype"  
               AND substr(name,1,&prefixL)="&prefix";
         if name EQ "Sev_PObserv_Source" then do;
            call symput('Source_Method','Infer');
            call symput('Source_Type',trim(left(cvalue)));
            call symput('Source_Uid',trim(left(nvalue)));
         end;
         else if name EQ "Freq_PObserv_Source" then do;
            call symput('Freq_PObserv_Source',trim(left(cvalue)));
         end;
         else if name EQ "Sev_Probability_Observed" then do;
            call symputX('Probability_Observed',put(nvalue, BEST20.));
         end;
         else if name EQ "Sev_PObserv_Method" then do;
            call symput('Source_Method',trim(left(cvalue)));
         end;  
         call symput("&flag", "true");
    run;
%mend get_pobserv;


/*************************************************************
 * Macro to read the THRESHOLD options 
 ************************************************************
 *   Frequency Options
 *
 *       Freq_Cut_Off                 : NUMERIC
 *       Freq_Keep_Larger_Thresholds  : TRUE or FALSE
 *       Freq_Missing_Default         : NUMERIC
 *       Freq_UseStratumMean_Default  : TRUE or FALSE
 *       Freq_Use_Data_Threshold      : TRUE or FALSE
 *       Freq_Zero_Count_Default      : NUMERIC
 *
 *   Severity Options
 *
 *       Sev_Cut_Off                  : NUMERIC
 *       Sev_GOF_Threshold            : NUMERIC
 *       Sev_Keep_Larger_Thresholds   : TRUE or FALSE
 *       Sev_Missing_Default          : NUMERIC
 *       Sev_Use_Data_Threshold       : TRUE or FALSE
 *       Sev_Use_GOF_Threshold        : TRUE or FALSE
 ************************************************************/

%macro read_threshold_options;

     /*** default values we expect to find ****/
     %if &fittype EQ SEV %then %do;
         %let threshold = .;
         %let gofThreshold = .;
         %let defaultThreshold = .;
         %let Sev_Use_Data_Threshold=true;
         %let Sev_Use_GOF_Threshold=false;
         %let Sev_Missing_Default=0;
         %let Sev_Cut_Off=10000;
         %let Sev_GOF_Threshold=.;
         %let Sev_Keep_Larger_Thresholds=.;
     %end;
     %else %do;
          %let Freq_Use_Data_Threshold = true;
          %let Freq_Missing_Default = 0;
          %let Freq_Zero_Count_Default =0;
          %let Freq_UseStratumMean_Default = false;
          %let Freq_UseStratumMean_UserSpec = false;
          %let Freq_Cut_Off = . ;
          %let Freq_Keep_Larger_Thresholds = true;
     %end;           
     
     /*** read options as specified by ccUid ***/
     %let options_exist=false;
     %get_options(flag=options_exist,type=&type,datatype=&datatype,ccuid=&ccuid);
     
     /*** if that failed and ccUid was specified, try project-level instead ***/ 
     %if &options_exist EQ false AND &ccuid NE . %then %do;
         %get_options(flag=options_exist,type=&type,datatype=&datatype,ccuid=.);
     %end;
     
     %if &fittype EQ SEV %then %do;
        %if %upcase(&Sev_Use_Data_Threshold) NE TRUE %then %do;
            %let threshold = &Sev_Cut_Off;
        %end;
        %if %upcase(&Sev_Use_GOF_Threshold) EQ TRUE %then %do;
            %let gofThreshold = &Sev_GOF_Threshold;
        %end;
        %let defaultThreshold = &Sev_Missing_Default;
     %end;     
%mend read_threshold_options;

/*************************************************************
 * Macro to read the distribution options into a
 * space-delimited string, placed in either dists or fdists
 ************************************************************/
%macro read_dist_options;
     /*** initialize our macro variables ***/
     %if &fittype EQ SEV %then %do;
       %let dists=;
     %end;
     %else %do;
       %let fdists=;
     %end;
     /*** look for distributions as requested, and concatenate ***/
     %let options_exist=false;
     %get_dists(flag=options_exist,ccuid=&ccUid,datatype=&datatype);
     /*** if we attempted cell-level and it failed, try project-level ***/
     %if &options_exist eq false AND &ccUid NE . %then %do;
       %get_dists(flag=options_exist,ccuid=.,datatype=&datatype);   
     %end;
%mend read_dist_options;

%macro get_dists(flag=options_exist,ccuid=,datatype=);
    data _NULL_; set &fitoptsData end=LAST; 
         where CCUID = &ccuid AND type EQ "DIST" AND substr(name,1,&prefixL)="&prefix"
            %if &ccuid ne . %then %do; AND datatype = "&datatype" %end;
            ;
         length dists $512;
         retain dists;
         dists = trim(left(dists)) || " " || trim(left(cvalue)) ;
         if LAST then do;
            %if &fittype EQ SEV %then %do;
               call symput('dists',trim(left(dists)));
            %end;
            %else %do;
               call symput('fdists',trim(left(dists)));
            %end;
            call symput("&flag", "true");
         end;
      run;
%mend get_dists;

/*************************************************************
 * Macro to read the FITMEASURE options 
 ************************************************************/
%macro read_fitmeasure_options;
     /*** initialize our macro variables ***/
     %if &fittype EQ SEV %then %do;
       %let Sev_FitMeasures=;
       %let Sev_FavFitMeasure=;
     %end;
     %else %do;
       %let Freq_FitMeasures=;
       %let Freq_FavFitMeasure=;
     %end;
     /*** look for distributions as requested, and concatenate ***/
     %let options_exist=false;
 
     %get_options(flag=options_exist,type=&type,ccuid=&ccUid,usePrefix=true,datatype=&datatype);
     /*** if we attempted cell-level and it failed, try project-level ***/
     %if &options_exist eq false AND &ccUid NE . %then %do;
       %get_options(flag=options_exist,type=&type,ccuid=.,usePrefix=true);   
     %end;
%mend read_fitmeasure_options;

/**************************************************************
 * Macro to read the KRI options into a
 * space-delimited string, placed in the following macro vars:
 *   Sev_Kris_Suggested, Sev_Kris_Forced
 *     or
 *   Freq_Kris_Suggested, Freq_Kris_Forced
 *************************************************************/
%macro read_kri_options;
     /*** initialize our macro variables ***/
     %if &fittype EQ SEV %then %do;
       %let Sev_Kris_Suggested=;
       %let Sev_Kris_Forced=;
     %end;
     %else %do;
       %let Freq_Kris_Suggested=;
       %let Freq_Kris_Forced=;
     %end;     
     /*** look for kris as requested, and concatenate ***/
     %let options_exist=false;
     %get_kris(flag=options_exist,ccuid=&ccUid,datatype=&datatype);

    /* Don't attempt if cell level KRICONTROL is set to true ------*/
     %local Sev_Use_Kris Freq_Use_Kris Sev_Selection_Method_Kris Freq_Selection_Method_Kris;
     %let Sev_Use_Kris=false;
     %let Freq_Use_Kris=false;
     data _NULL_; set &fitoptsData; 
         where CCUID = &ccuid AND type EQ "KRICONTROL"  ;
         call symput(name,compress(cvalue));
     run;
     /*** if we attempted cell-level and it failed, try project-level ***/
     %if (&Sev_Use_Kris=false and &fittype=SEV) or
        (&Freq_Use_Kris=false and &fittype=FREQ) %then %do;
         /*** if we attempted cell-level and it failed, try project-level ***/
         %if &options_exist eq false AND &ccUid NE . %then %do;
           %get_kris(flag=options_exist,ccuid=.,datatype=&datatype);   
    %end;
     %end;
%mend read_kri_options;

/**************************************************************
 * Macro to read the INDICATOR options into a
 *  placed in the following macro vars:
 *************************************************************/
%macro read_indicator_options;
     /*** initialize our macro variables ***/
     %if &fittype EQ FREQ %then %do;
       %let Freq_Use_Indicators=;
       %let Freq_Model_Basis_Props=;
      %let Freq_Indicator_Properties=;
     %end;     
     /*** look for  ***/
     %let options_exist=false;
    %get_options(flag=options_exist,type=INDICATOR,ccuid=&ccUid,datatype=&datatype, usePrefix=false);
     /*** if we attempted cell-level and it failed, try project-level ***/
     %if &options_exist eq false AND &ccUid NE . %then %do;
      %get_options(flag=options_exist,type=INDICATOR,ccuid=.,usePrefix=false);
     %end;
%mend read_indicator_options;


/**************************************************************
 * Macro to read the ADVANCED options into a
 *  placed in the following macro vars:
 *************************************************************/
%macro read_advanced_options;
     /*** initialize our macro variables ***/
     %if &fittype EQ SEV %then %do;
         %let Min_Fitting_Count = ;
     %end;     
     /*** look for  ***/
     %let options_exist=false;
     %get_options(flag=options_exist,type=ADVANCED,ccuid=&ccUid,datatype=&datatype, usePrefix=false);
     /*** if we attempted cell-level and it failed, try project-level ***/
     %if &options_exist eq false AND &ccUid NE . %then %do;
        %get_options(flag=options_exist,type=ADVANCED,ccuid=.,usePrefix=false);
     %end;
%mend read_advanced_options;


%macro get_kris(flag=options_exist,ccuid=,datatype=);

   proc sql noprint;
      /* Suggested */
      SELECT distinct
                        name
      INTO
                        :&fittype._Kris_Suggested separated by " "
      FROM
                        &fitoptsData.
      WHERE
                        CCUID = &ccuid 
                        AND 
                        type EQ "KRI" 
                        AND 
                        cvalue EQ "&fittype"
                        %if &ccuid NE . %then %do; 
                           AND 
                           datatype = "&datatype" 
                        %end;
                        AND
                        nvalue = 1
      ;
      /* Forced */
      SELECT distinct
                        name
      INTO
                        :&fittype._Kris_Forced separated by " "
      FROM
                        &fitoptsData.
      WHERE
                        CCUID = &ccuid 
                        AND 
                        type EQ "KRI" 
                        AND 
                        cvalue EQ "&fittype"
                        %if &ccuid NE . %then %do; 
                           AND 
                           datatype = "&datatype" 
                        %end;
                        AND
                        nvalue = 2
      ;
   quit;

   %let &flag. = true;

%mend get_kris;

/************* example/test code ***************
libname proj "C:\VarTablesNew\ver_1\components\projects\Bank11782";
options mprint symbolgen;

%let threshold = .;
%let gofThreshold = .;
%let defaultThreshold = .;
%orva_readfitopts(fitoptsdata=proj.fitopts,type=THRESHOLD,fittype=SEV,datatype=I,ccUid=7);
%put "threshold=&threshold";
%put "gofThreshold=&gofThreshold";
%put "defaultThreshold=&defaultThreshold";

%let Freq_Use_Data_Threshold = .;
%let Freq_Missing_Default = .;
%let Freq_Zero_Count_Default = .;
%let Freq_UseStratumMean_Default = .;
%let Freq_UseStratumMean_UserSpec = .;
%let Freq_Cut_Off = .;
%let Freq_Keep_Larger_Thresholds = .;
%orva_readfitopts(fitoptsdata=proj.fitopts,type=THRESHOLD,fittype=FREQ,datatype=I,ccUid=7);
%put "Freq_Use_Data_Threshold=&Freq_Use_Data_Threshold";
%put "Freq_Missing_Default=&Freq_Missing_Default";
%put "Freq_UseStratumMean_Default=&Freq_UseStratumMean_Default";
%put "Freq_UseStratumMean_UserSpec=&Freq_UseStratumMean_UserSpec";
%put "Freq_Zero_Count_Default=&Freq_Zero_Count_Default";
%put "Freq_Cut_Off=&Freq_Cut_Off";
%put "Freq_Keep_Larger_Thresholds=&Freq_Keep_Larger_Thresholds";


%let dists=;
%orva_readfitopts(fitoptsdata=proj.fitopts,type=DIST,fittype=SEV,datatype=I,ccUid=.);
%put "dists=&dists";

%let TailBoundaryPercent=.;
%let TailBoundaryValue=.;
%let TailBoundaryChoice=.;
%orva_readfitopts(fitoptsdata=proj.fitopts,type=SEVUSERINPUT,ccUid=23);
%put "TailBoundaryPercent=&TailBoundaryPercent";
%put "TailBoundaryValue=&TailBoundaryValue";
%put "TailBoundaryChoice=&TailBoundaryChoice";

***********************************************/

/*
data ljfitopts;
   NAME="Freq_missing_default"; NVALUE=1000; CVALUE=.;TYPE="THRESHOLD"; CCUID=.; dataType="I";output;
   NAME="Freq_missing_default"; NVALUE=5000; CVALUE=.;TYPE="THRESHOLD"; CCUID=3; dataType="I";output;

run;
*/

/*%let freq_missing_default = .;*/
/*%let gofThreshold = .;*/
/*%let defaultThreshold = .;*/
/*options mprint symbolgen;*/
/*%orva_readfitopts(fitoptsdata=ljfitopts,fittype=FREQ,datatype=I, ccuid=);*/
/*%put " freq_missing_default = &freq_missing_default";*/
/*%put "gofThreshold=&gofThreshold";*/
/*%put "defaultThreshold=&defaultThreshold";*/

/*
options mprint symbolgen;
%orva_readfitopts(fitoptsdata=ljfitopts,fittype=FREQ,datatype=I, ccuid=3);
%put " freq_missing_default = &freq_missing_default";

options mprint symbolgen;
%orva_readfitopts(fitoptsdata=ljfitopts,fittype=FREQ,datatype=I, ccuid=7);
%put " freq_missing_default = &freq_missing_default";
*/

/**/
/*%let Freq_Use_Data_Threshold = .;*/
/*%let Freq_Missing_Default = .;*/
/*%let Freq_Zero_Count_Default = .;*/
/*%let Freq_Use_Stratum_Mean = .;*/
/*%let Freq_Cut_Off = .;*/
/*%let Freq_Keep_Larger_Thresholds = .;*/
/*options mprint symbolgen;*/
/*%orva_readfitopts(fitoptsdata=proj.fitopts,fittype=FREQ,datatype=I,ccUid=7);*/
/*%put "Freq_Use_Data_Threshold=&Freq_Use_Data_Threshold";*/
/*%put "Freq_Missing_Default=&Freq_Missing_Default";*/
/*%put "Freq_Use_Stratum_Mean=&Freq_Use_Stratum_Mean";*/
/*%put "Freq_Zero_Count_Default=&Freq_Zero_Count_Default";*/
/*%put "Freq_Cut_Off=&Freq_Cut_Off";*/
/*%put "Freq_Keep_Larger_Thresholds=&Freq_Keep_Larger_Thresholds";*/
