/************************************************************************************* 
 *
 * Copyright (C), 2007
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 * NAME   : orva_reset_sparm_preserve                                             
 * PURPOSE: Reset the fitting dataset, preserving the complete collection of scenario fitting related iterations where there
 *          has been any type of modification (in the collection). Any iterations that are found to be orphans must be cleaned
 *          up as well. This macro only preserves the sparm iterations
 * SUPPORT: tomlon
 *
 * PARAMETERS :
 *   sparm = sparm dataset name
 *   sparm = fparm dataset name
 *   statusds = status data set
 *
 * OUTPUT: 
 *   None
 * 
 * Logic is as follows:
 *   1) Find all modified frequency iterations and get their severity iteration parent.
 *   2) Add to this collection all referenced severity iterations.
 *   3) Add to this all consortium iteration parents.
 *   4) Any iteration in the above list must exist in the severity iteration dataset.
 *   5) Delete all frequency iterations that do not have severity iteration parents 
 *      in the above severity iteration list.
 **************************************************************************************/
%macro orva_reset_sparm_preserve( sparm=, fparm=, statusds= );

   /*** Initialize the status dataset ***/
    %let reason = %sysfunc(sasmsg(&orvads, unknown_error, noquote));
    %let status=1;
    %orva_status_ds( statusds=&statusds, status=&status, reason=&reason );

   /*** Sort sparm by required fields ***/
   proc sort data=&sparm out=sortedsparm;
      by ccuid iteration;
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_sort_sparm, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_sort_sparm, noquote));
       %goto LOCAL_EXIT;
   %end;
   
   
   
   /*- check for empty dataset -*/
   %let dsid=%sysfunc(open(&fparm,I));
   %let nfreqs=%sysfunc(attrn(&dsid,nlobs));
   %let rc=%sysfunc(close(&dsid));
   %if &nfreqs>0 %then %do;
      /*** Find all modified frequency iterations and get their severity iteration parents ***/
      data tmpmodifiedfreqseviters;
         set &fparm;
         keep ccuid severityiteration;
         length severityiteration 8;
         if esttype NE '.';
         severityiteration = seviter;
      run;     
      %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
          %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_pull_modified_freqs, noquote));
          %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_pull_modified_freqs, noquote));
          %goto LOCAL_EXIT;
      %end;
      proc sql; 
           create table uniquetmpmodifiedfreqseviters as
           select distinct ccuid,severityiteration
           from tmpmodifiedfreqseviters;
      quit;
      %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
          %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_pull_modified_unique, noquote));
          %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_pull_modified_unique, noquote));
          %goto LOCAL_EXIT;
      %end;
      data modifiedfreqseviters;
         set uniquetmpmodifiedfreqseviters;
         keep ccuid iteration;
         length iteration 8;
         iteration = severityiteration;
      run;
      %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
          %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_rename_modfreq_column, noquote));
          %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_rename_modfreq_column, noquote));
          %goto LOCAL_EXIT;
      %end;
   %end;
      
   /*** Find all modified severity iterations ***/
   data modifiedseviters;
      set sortedsparm;
      keep ccuid iteration;
      if esttype NE '.';
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_detect_mod_sev_iters, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_detect_mod_sev_iters, noquote));
       %goto LOCAL_EXIT;
   %end;
   
   /*** Out of the modified freq sev iterations, keep only those that exist in the sparm dataset ***/
   
   %if &nfreqs>0 %then %do;
      data modifiedseviterstokeep;
         merge modifiedfreqseviters(in=inModFreq)
               sortedsparm(in=inSParm);
         by ccuid iteration;
         keep ccuid iteration;
         if inSParm=1;
         if inModFreq=1;
      run;
      %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
          %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_freq, noquote));
          %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_freq, noquote));
          %goto LOCAL_EXIT;
      %end;
   %end;
   %else %do;
      data modifiedseviterstokeep;
         merge sortedsparm(in=inSParm);
         by ccuid iteration;
         keep ccuid iteration;
         if inSParm=0;
      run;
      %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
          %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_freq, noquote));
          %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_freq, noquote));
          %goto LOCAL_EXIT;
      %end;
   %end;
   
   data combinedmodifiedseviterstokeep;
      merge modifiedseviterstokeep
            modifiedseviters;
      by ccuid iteration;
      keep ccuid iteration;
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_ref, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_ref, noquote));
       %goto LOCAL_EXIT;
   %end;
   
   /*** For the above iterations to keep, also record any referenced severity iteration ***/
   data tmpreferencedseviterstokeep;
      merge combinedmodifiedseviterstokeep
            sortedsparm;
      by ccuid iteration;
      keep ccuid tmpiteration;
      length tmpiteration 8;
      if reference ne .;
      tmpiteration = reference;
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_ref_iters, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_merge_sev_ref_iters, noquote));
       %goto LOCAL_EXIT;
   %end;
   proc sql; 
        create table uniquetmprefseviterstokeep as
        select distinct ccuid, tmpiteration
        from tmpreferencedseviterstokeep;
   quit;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_determine_unique_ref_iters, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_determine_unique_ref_iters, noquote));
       %goto LOCAL_EXIT;
   %end;
   data referencedseviterstokeep;
      set uniquetmprefseviterstokeep;
      keep ccuid iteration;
      length iteration 8;
      iteration = tmpiteration;
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_rename_ref_iters, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_rename_ref_iters, noquote));
       %goto LOCAL_EXIT;
   %end;
      
   /*** get the superset of all modified severity iterations and their reference non-scenario iterations ***/
   data tmpallseviterationstokeep;
      merge referencedseviterstokeep
            combinedmodifiedseviterstokeep;
      by ccuid iteration;
      keep ccuid iteration;
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_sev_iter_superset, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_sev_iter_superset, noquote));
       %goto LOCAL_EXIT;
   %end;
   proc sql; 
        create table allseviterationstokeep as
        select distinct ccuid, iteration
        from tmpallseviterationstokeep;
   quit;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_sev_iter_superset_unique, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_sev_iter_superset_unique, noquote));
       %goto LOCAL_EXIT;
   %end;
      
   /*** Create the new Sparm dataset, only preserving those iterations belonging to the above collections ***/
   data &sparm;
      merge sortedsparm(in=inSParm)
            allseviterationstokeep(in=insevtokeep);
      by ccuid iteration;
      if inSParm=1;
      if insevtokeep=1;
   run;
   %if &SYSERR^=0 & &SYSERR ^= 4 %then %do;
       %put &orvar_error %sysfunc(sasmsg(&orvads, reset_iterations_cant_regen_parmfile, noquote));
       %let reason=%sysfunc(sasmsg(&orvads, reset_iterations_cant_regen_parmfile, noquote));
       %goto LOCAL_EXIT;
   %end;
   
   /*** all is OK - reflect this in the status dataset ***/
   %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
   %let status=0;

%LOCAL_EXIT:

   /*** Write out the status and any reason ***/
   %orva_status_ds( statusds=&statusds, status=&status, reason=&reason );
   
   /*** Delete unwanted data sets ***/
   proc datasets library=work memtype=data nolist nodetails;
      delete sortedsparm tmpmodifiedfreqseviters uniquetmpmodifiedfreqseviters modifiedfreqseviters modifiedseviters modifiedseviterstokeep combinedmodifiedseviterstokeep tmpreferencedseviterstokeep uniquetmprefseviterstokeep referencedseviterstokeep tmpallseviterationstokeep allseviterationstokeep;
   quit;
                                  
%mend orva_reset_sparm_preserve;
