/*****************************************************************************************
 * Copyright (C), 2005
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved. 
 *
 * NAME    : orva_shared_corr.sas  
 * PURPOSE : Create Covariance matrix for the shared losses          
 * SUPPORT : sasdje         
 *
 * PARAMETERS :
 *
 *    data            =  the name of the loss dataset with the following variables:
 *                       CCUID
 *                       FINANCIALDATE
 *                       singleEventMult
 *                       FSCALE      
 *    global_range    =  Global range dataset for internal losses, must contain the following
 *                       variables: START_DATE, END_DATE 
 *    ccLibref        =  the name of the libref with the cc data information
 *
 *    GenTotals       = 0 or 1 to indicate whether covariances for total cells are needed.
 *      
 *    singleEventMult =  Single event multiple loss variable name, defaults to SHARED_LOSS_GROUP_UID
 *    outCorr         =  Empirical correlation matrix, defaults to work.outCorr
 *    outStatus       =
 *    freqCounts      =  freqcounts_[i|c] dataset
 *
 *    TODO: may use first_date_of_event instead of financialdate.
 *****************************************************************************************/
%macro orva_shared_corr( 
                        data            = ,                         
                        interval        = ,
                        global_range    = ,
                        ccLibref        = , 
                        singleEventMult = SHARED_LOSS_GROUP_UID,
                        GenTotals       = 1,
                        OutCorr         = outCorr, 
                        outStatus       = ostatus,
                        freqCounts      =
                        );
    /*- create the status data set upfront -*/
    /*suppress the warnings of the long(>262) quoted string may have a unbalanced quote. 
      In 3D data, The macro needs a 441 bit string to represent the shared loss info.*/
    options noquotelenmax;
    %let reason=%sysfunc(sasmsg(&orvads, unknown_error, noquote));
    %orva_status_ds( statusds=&outstatus, status=1, reason=&reason );

    %let status=0;
    %let reason=%sysfunc(sasmsg(&orvads, successful_note, noquote));
    %let abort=0;

    /*- Determine the number of dimensions ------------*/
    %let NUMCCVS=0;

    proc contents noprint data=&ccLibref.cc out=_cc; run;

    %let dsid=%sysfunc(open(_cc,I));
    %if &dsid ^= 0 %then %do;
        %let NUMCCVS=%sysfunc(attrn(&dsid,nlobs));
        %let rc=%sysfunc(close(&dsid));
    %end;
    %if &NUMCCVS eq 0 %then %return;
    %let NUMCCVS= %eval(&numccvs-1);

    /*- create macro variables for all interior cells names -*/
    proc sort data=&ccLibref.cc out=_cc;by ccuid;run;
    %let NVARS=0;
    data ccmap; set _cc;
	  length nstr str $32;
      retain num 0 total 0;

      total +1;
      if 1 %do i = 0 %to &NUMCCVS-1;
              and CCVUID_&i ne 0
           %end;
      then do;
        num+1;
        nstr = "NAME_" || strip(put(num,best.));
        /* macros NAME_1 - NAME_<nvars> are linked to Cell_<CCUID> */
		str = "Cell_" || strip(put(CCUID,&ORVAR_CELLFMT));
        call symput(nstr, strip(str));
        /* macros Cell_<CCUID> are linked to their nvar */
        call symputX("NVARS",num);
        call symputX("NN",num*num);
      end;
      call symputX("All_NVARS",total);
	  
	  keep ccuid num;
    run;
    /* this sorting order is important in constructing the bitmap 
       in creating dataset _cov_with_group1 */
    proc sort data=ccmap;
       by num;
    run;

    /*Create a map that maps the num to ccuid*/
    data ccmapReverse; 
      set ccmap;
      by num;
      if(first.num) then output;
    run;
 
    /*Create an orderred var list: share1, share2,..., note the number is not 
      consecutive. The holes are reserved to the totals*/
    proc sql noprint;
        select "shared"||strip(put(ccuid, best.)) into :varOrder separated by " "
        from ccmapReverse;
    quit;

    %if &NVARS eq 0 %then %return;

    /*- merge with the cc data set ------------------------------*/
    data _datain_; set &data;
      count = 1;
      keep CCUID &SingleEventMult FSCALE FINANCIALDATE count;
    run; 

    /*for each ccuid, there are nperiods(calculate below) of intervals, each has a corresponding
      exposure and interval definition
     */
    proc sql noprint;
        create table _datain_freq_ as 
        select distinct ccuid, start_of_curr_fitinterval, end_of_curr_fitinterval, exposure 
        from &freqCounts;
    quit;

    /*Bring in the time exposure and fitting intercal information*/
    proc sql noprint;
       create table _datain2_ as 
       select * from _datain_, _datain_freq_ 
       where _datain_.ccuid=_datain_freq_.ccuid;
    quit;

    /* create a count variable to indicate the occurence of shared loss event. This count
       will be used later to get the shared and un-shared loss count for each shared loss group.
       if the shared loss event happened within the fitting interval, count=1. Otherwise 0.
       The puupose of doing this is to only get the count from shared loss event, and not for the 
       rows that used to calculate the exposure across the time intervals within global range. 
       We should make sure that the number of none zero COUNT value records should match the
       total number of records in &data(the loss dataset). 
     */ 
    data _datain_;
       set _datain2_;
       if financialdate GE start_of_curr_fitinterval AND 
          financialdate LT end_of_curr_fitinterval then do;
           count = 1*fscale;
       end;
       else do;
           count = 0;
       end;
    run;

    proc sql noprint; 
      create table _nototal_ as select * from _datain_ L, &ccLibref.cc C
           where L.CCUID=C.CCUID;
    quit;

    /*- Shared losses all have variance of 1/nperiods --------*/
    data _NULL_; set &global_range;
      nperiods = intck( "&interval", start_date, end_date ) +1;
      call symput( "nperiods", nperiods);
	  call symput("start_date",trim(put(start_date,date9.)));
      call symput("end_date", trim(put(end_date,date9.)));
    run;
    /*create time buckets for each interval when the event happens*/
    data _shared_; set _nototal_;
      start = "&start_date"d;
      thisInterval=intck( "&interval", start, (start_of_curr_fitinterval+end_of_curr_fitinterval)/2 );
      keep CCUID &SingleEventMult fscale FINANCIALDATE thisInterval exposure count;
      drop start_of_curr_fitinterval end_of_curr_fitinterval;
    run;

    /*get the shared(unshared) count for shared loss events*/
    proc sort data=_shared_; 
        by &SingleEventMult thisInterval CCUID; 
    run;

    proc summary data=_shared_; by &SingleEventMult thisInterval CCUID ;
        var count;
        output out=_shared_Sum sum=countSum;
    run;

    proc transpose data=_shared_Sum out=_shared_t_ prefix=shared;
        by &SingleEventMult thisInterval;
        id CCUID;
        var countSum;
    run;

    /*given the shared loss count info, find the minimum for each cell set combination
      (via the bitmap). bitmap (eg 0011) is a representation of a set of cells that 
      share the loss for that shared loss group id. bit 1 means the ith cell share the 
      loss. 0011 means the 1st and 2nd cell share the loss.  
     */
    data  _cov_with_group1;
        set _shared_t_(rename=(shared_loss_group_uid=the_shared_loss_group_uid)); 
        array c[*] &varOrder;
        array cc(&nvars); 
        /*cc is used to hold the value of loss count and later is used to construct unshared 
          loss count when shared loss group uid is missing. if the data is:
          +-----------------------+---------+--------+--------+
          | shared_loss_group_uid | cell_1  | Cell_2 | Cell_3 |
          +-----------------------+---------+--------+--------+
          |                      .|     20  |    11  |      9 |
          +-----------------------+---------+--------+--------+
        it is transformed to:
          +-----------------------+---------+--------+--------+-------+
          | shared_loss_group_uid | cell_1  | Cell_2 | Cell_3 |bitmap |
          +-----------------------+---------+--------+--------+-------+
          |                      .|     20  |    0   |      0 | 001   |
          +-----------------------+---------+--------+--------+-------+
          |                      .|      0  |    11  |      0 | 010   |
          +-----------------------+---------+--------+--------+-------+
          |                      .|      0  |    0   |      9 | 100   |
          +-----------------------+---------+--------+--------+-------+
        */
        do j = 1 to &nvars;
            cc[j] = c[j];
        end;
        length bitmap $&nvars.;
        format bitmap $&nvars.. ;
        BITS1 = repeat("1", (&nvars-1)); /*create a string of 1s of length nvars*/
        BITS0 = repeat("0", (&nvars-1));
        bitmap = BITS1; /*init to 1, when bitmap = 0, stop the iteration*/
        /*Continue the loop unitl the bitmap is all zero*/
        do while(bitmap ne BITS0); /*&nvars iterations*/
            bitmap = BITS0;
            minV = 10e6; /*init to a large number, this is a unlike number for shared lossed event number that happens in a year: the longest interval the Var handles*/
            minInd = 0; /*index of the which bit in the bitmap that is 1*/
            if (missing(the_shared_loss_group_uid)) then do;   /*losses with missing shared loss group uid are treated as unshared loss*/
                do i =1 to &nvars;
                    if missing(cc[i]) then cc[i]=0;
                    if missing(c[i]) then c[i]=0;
                    if(cc[i] GE 1) then do;
                        /* if shared_loss_group_uid is missing, then treat this loss as
                           unshared loss. So create bipmap for each non-zero c[i] */
                        /* move the loss count into diagnal */
                        do ii=1 to &nvars;
                            if ii eq i then do;
                                c[ii]=cc[i];
                            end;
                            else do;
                                c[ii]=0;
                            end;
                        end;
                        /*set the corresponding bitmap to 1 at ith positon counting from right*/
                        if i eq 1 then do;
                            bitmap=repeat("0",&nvars-2)||"1";
                        end;
                        else if i eq &nvars then do;
                            bitmap="1"||repeat("0",&nvars-2);
                        end;
                        else do;
                            bitmap = repeat("0",(&nvars-i-1))||"1"||repeat("0", (i-2));
                        end;
                        minV = cc[i]; 
                        minInd=i;
                        output;
                        bitmap=BITS0;
                        continue;
                    end;
                end; /*end do i loop*/
            end; /*end of missing guid case*/
            else do; /*guid is defined*/ 
                do i =1 to &nvars;
                    if missing(c[i]) then c[i]=0;
                    if(c[i] GE 1) then do;
                        /*in binary representation, this corresponds to set ith bit to 1 counting from right*/
                        if(i eq &nvars) then do;
                            bitmap="1"||substr(bitmap, 2, &nvars-1);
                        end;
                        else if (i eq 1) then do;
                            bitmap=substr(bitmap, 1, &nvars-1)||"1";
                        end;
                        else do;
                            bitmap = substr(bitmap, 1, &nvars-i)||"1"||substr(bitmap, &nvars-i+2, i-1);
                        end;
                    end; 
                    if c[i] LT minV and c[i] GT 0 then do;
                        /*record the minimum count value aross cells*/
                        minV = c[i]; 
                        minInd=i;
                    end;
                end; /*end do i loop*/ 
                /*   if bitmap eq 0 then break;*/
                if bitmap ne BITS0 then do;
                    output;
                end;
                /*subtract the minimal count value from the shared loss count then start over again*/
                do i = 1 to &nvars;
                    if c[i] gt 0 then do;
                        c[i]=c[i]-minV;
                    end;
                end;
            end; /*end of guid is defined*/
        end;/*while loop*/
        drop cc: i ii j BITS:;
    run;

    /* handle the time exposure */
    proc sql noprint;
        create table _exposure_data as 
        select distinct ccuid, exposure, thisinterval from _shared_
        order by ccuid, thisinterval;
    quit;

    %local n_interval;
    %let n_interval = %eval(&nperiods-1);

    proc sql noprint;
        create table _bitmap as select distinct bitmap from _cov_with_group1;
    quit;

    /*This data step is to calculate the sum of time exposure over all time intervals 
      for each bitmap, note that the time exposure for each time interval is the mininum
      time exposure in the cell group
     */
    data _cov_with_group2;
        if (0) then set ccmapReverse;
        if (0) then set _exposure_data;
        if _N_ = 1 then do;
	        declare hash ccHash(dataset: "ccmapReverse");
             rc = ccHash.defineKey("num");
             rc = ccHash.defineData("CCUID");
             rc = ccHash.defineDone();

             declare hash exposureHash(dataset: "_exposure_data");
             rc = exposureHash.defineKey("CCUID", "thisinterval");
             rc = exposureHash.defineData("exposure");
             rc = exposureHash.defineDone();
        end;
        set _bitmap;
        /*from bitmap, figured out which bit(that is the mapped ccuid number) contributes exposure*/
        retain mexp 0;
        mexp=0;
        do thisinterval = 0 to &n_interval; /*to summarize the exposures*/
            minexp = 1; /*start from max*/
            do i = 1 to &nvars;
                if (substr(bitmap, &nvars-i+1, 1) eq "1") then do;
                    num=i;
                    rc = ccHash.find();
                    if rc=0 then do; /*get CCUID*/
                        rc = exposureHash.find();
                        if rc = 0 then do;/*get exposure*/
                            if (exposure lt minexp) then do;
                                /*keep the smallest exposure across cells in this interval*/
                                minexp=exposure;
                            end;
                        end; 
                    end;
                end;
            end; /* i loop*/
            mexp = mexp+minexp; /*accumulate the exposures for this interval*/
        end; /*end thisinterval*/
        drop rc ccuid num i exposure thisinterval minexp;
    run;

    proc sort data=_cov_with_group1;
        by bitmap;
    run;

    /*now we have the reduced shared loss count per bitmap*/
    proc summary data=_cov_with_group1 ;
        output out=_cov_with_group3(drop=_freq_ _type_)  sum=minVSum;
        var minV;
        by bitmap;
    run;  

    /*create the table that collects bitmap and corrsponding loss counts(summed over loss group and intervals)
      and exposure(summed over intervals: minimum of the cell exposure among shared group )*/
    proc sql noprint;
        create table _shared_lost_table as select nk.bitmap, nk.minvsum, e.mexp from 
           _cov_with_group3 as nk, _cov_with_group2 as e 
        where nk.bitmap=e.bitmap;
    quit;

    /*calculate the covariance*/
    data _shared_lost_table;
        set _shared_lost_table;
        cov = minvsum/mexp;
    run;

    proc sql noprint;
        select strip(put(count(bitmap),best.)) into :nbm from _shared_lost_table;
        select bitmap into :bm1 - :bm%left(%trim(&nbm)) from  _shared_lost_table;
    quit;

    /*calculate covariance matrix */
    data outCov;
        array covArray[&NVARS,&NVARS];
        array bm[&nbm] $&nvars.;
        array cell_[&NVARS] %do i=1 %to &NVARS;
                                &&NAME_&i.
                            %end;
                            ;
        if (0) then set _shared_lost_table;

        %do i = 1 %to &nbm;
            bm[&i]="&&bm&i.";
        %end;

        do i = 1 to &nvars;
            do j = 1 to &nvars;
                covArray[i,j] = 0;/*init to 0, otherwise missing*/
            end;
        end;
        _type_ = "COV";
        if _N_ = 1 then do;
	        declare hash shlHash(dataset: "_shared_lost_table");
            rc = shlHash.defineKey("bitmap");
            rc = shlHash.defineData("cov");
            rc = shlHash.defineDone();
        end;
        /*sum up the cov of c(i,j) by summing up the covs corresponding to bitmaps with none zero
          bit at position ith and jth*/
        do i= 1 to &NVARS;
            do j = 1 to &NVARS;
                do k = 1 to &nbm; 
                    if (substr(bm[k], &nvars-j+1, 1) eq "1") then do;
                        if (substr(bm[k], &nvars-i+1, 1) eq "1") then do;
                            bitmap = bm[k];
                            rc = shlHash.find();
                            /*both bit position has one, add the value to the cov[i,j]*/
                            covArray[i,j] = covArray[i,j] + cov;
                        end;
                    end;  
                end; /* loop k */
            end; /*loop j*/
        end; /*loop i*/

        /*- write out the covariance matrix -*/
        do i= 1 to &NVARS;
            do j = 1 to &NVARS;
                cell_[j] = covArray[i,j];
            end;

            _name_ = vname(cell_[i]);
            output;
        end;
        keep _name_ _type_  cell_:;
    run;

   %if &GenTotals = 1 %then %do;
      %orva_rollup_cov( 
                       cov = OutCov, 
                       ccLibref = &ccLibref
                       );
   %end;

   /* In outCov data, some diagnal is zero since there are no loss(no count) for the cell. */
   /* To make it friendly to orva_cov2corr, set these zeros to 1                           */
   data outCov; 
      set outCov; 
      array diag[*] cell_:; 

      if diag[_n_] eq 0 or diag[_n_] eq . then do; 
         diag[_n_] = 1; 
      end; 

      do i = 1 to dim(diag); 
         if diag[i] < 0 then do; 
            diag[i] = 0; 
         end; 
      end;     
   run; 

   %orva_cov2corr(
                  inCov   = outCov, 
                  outCorr = &outCorr, 
                  nvars   = &ALL_NVARS
                  );  
    options quotelenmax;
%DONE:
    %if &abort %then %do;
        %put &orvar_error %sysfunc(sasmsg(&orvads, scorr_abnormal_error, noquote));
        %let status=1;
    %end;  
    %orva_status_ds( statusds=&outstatus, status=&status, reason=&reason );


    proc datasets library=work memtype=data nolist;
        delete ccmap: _shared_: _shared_t_ _datain_ _datain2_ _nototal_ outCov _shared_lost_table
               _cov_with_group: _exposure_data _bitmap;
    quit;
%mend orva_shared_corr;
