/******************************************************************************
 * $Id: orva_var_gradient.sas,v 1.1.2.23.2.41.2.4.2.1 2012/08/06 14:01:59 jogwhi Exp $
 * Copyright (C), 2008
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 *
 * NAME:     orva_var_gradient.sas
 * PURPOSE:  Calculate the gradient of the VaR values with respect to the parameters. 
 *           Called by %orva_conf_interval().
 * SUPPORT:  stmajo
 *
 * Parameters:
 *
 *   ccUid         = key for this cell.
 *   spriority     = Name of the input data set containing severity parameter estimates.
 *   fpriority     = Name of the input data set containing frequency parameter estimates.
 *   fdraws        = Frequency draws from simulation
 *   SimInterval   = Interval used for frequency simulation
 *   varstat       = Used to get the var numbers for the cell 
 *   loss          = the losses for the model used in this cell
 *   freq_gradient = The name of the output dataset (for frequency) for the ccUid.
 *   sev_gradient  = The name of the output dataset (for severity) for the ccUid.
 *
 * NOTE: This is, for intents and purposes, a submacro of orva_conf_interval.  It is not 
 *       designed to be used as a stand alone macro.  Any errors are passed through to 
 *       orva_conf_interval for handling.  Error messages should not be written to the 
 *       log by this macro.
 ***********************************************************************************/
%macro orva_var_gradient( ccUid         =,
                          spriority     =,
                          fpriority     =,
                          fdraws        =, 
                          SimInterval   =,
                          varstat       =,
                          loss          =,
                          freq_gradient =,
                          sev_gradient  =,
                          abort_flag    =,
                          debug         =false
                         );

   /* Status macros are already defined in orva_conf_interval.  
      Initialize abort_flag at 1and only set to 0 if no error occurs */
   %let &abort_flag = 1;

   /* Set up the CELL_WHERE macro */
   %let CELL_WHERE=;
   %orva_cell_where(ccuid=&ccuid,ccData=&ccData);

   /* Define "infinity" to check against later */
   %let INFINITY = %sysfunc(exp(700));


   /*--- Get the list of evaluation points  -----------------*/
   %let p_1=none;
   data _vg_var_; set &varstat;
      where CCUID=&CCUID;
      call symputX("P_1", put(p_1, BEST20.));
      call symputX("P_2", put(p_2, BEST20.));
      call symputX("P_3", put(p_3, BEST20.));
      var_value = p_1;
      quantile = pctl1;
      output;
      var_value = p_2;
      quantile = pctl2;
      output;
      var_value = p_3;
      quantile = pctl3;
      output;
      keep var_value quantile;
   run;

   /*--- Must have VaR Values -------------*/
   %if &p_1 eq none %then %do;  
      %let reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_var_values_error, noquote, &ccuid));
      %goto DONE;
   %end;
   /*****************************************************************/
   /*           Get the Severity Distribution Information           */
   /*****************************************************************/

   /* Get datasets where information is stored */
   %let sev_dsname = none;
   data _NULL_;
      set &spriority;
      where CCUID=&CCUID and SOURCE_TYPE = "I";
      /* Dataset with  iteration, distribution and parameters, selected KRIs and coefficients, # of KRIs  ( selsparmi ) */
      call symput("sev_dsname",dataset_name);
   run;

   /*--- Must have severity distribution info -------------*/
   %if &sev_dsname eq none %then %do;  
      %let reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_sev_info_error, noquote, &ccuid));
      %goto DONE;
   %end;
   %orvl_ds_exists(&sev_dsname, "orva_var_gradient", false);
   %if &ds_nobs eq 0 %then %do;
      %let reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_sev_info_error, noquote, &ccuid));
      %goto DONE;
   %end; 

   /* Pull out the distribution and KRI information */
   data _NULL_;
      set &sev_dsname;
      where CCUID=&CCUID;
      call symputx("siteration",iteration);
      sdist = strip(distribution);
      call symputx("sdist", sdist);
      if count_kri = . then count_kri =0;
      call symputx("sev_KRI_count",count_kri);
      call symputx("sparamcount",paramcount);
      call symputx("sev_kri_model",KRI_MODEL);
   run;

   %let SEV_RC_TYPE = INTERNAL_BL_UID;
   data _NULL_;
      set &sev_dsname;
      where CCUID=&CCUID;
      %do i=1 %to &sparamcount;
         call symputx("SParam_&i",put(Param_&i,BEST20.));
      %end;
   run;
            
   /***load the severity distribution macro ***/
   %orva_set_sev_dist(distIn=&sdist, distOut=sdist);
   %orvd_&sdist;

   /* Get the severity parameters used in simulation */
   /* If KRIs present, first need to get the severity scale parameter across all 
      cell subunits (risk categories e.g. internal BLs).  We will need a 
      Dparam_1_# for each opf these. If no KRIs, we have all we need */
   %if %sysevalf(&SEV_KRI_COUNT >0, boolean) %then %do;

      /* Need to distinguish between firm level and risk category level KRIs */
      data _null_;
         set &sev_dsname;
         where CCUID=&CCUID;
         length firm_kri_list $2000;
         length rc_kri_list $2000;
         n_f_kris = 0;
         n_rc_kris = 0;
         array kri[*] Param_1_KRI_:;
         do i=1 to dim(kri);
            if kri[i] ne . then do;
               y = trim(vname(kri[i]));
               z = length(y);
               if z eq 16 then do;
                  rc_kri_list = trim(rc_kri_list) || "  " || trim(substr(vname(KRI(i)),9,8));
                  n_rc_kris = n_rc_kris + 1;
               end;
               else do;
                  firm_kri_list = trim(firm_kri_list) || "  " || trim(substr(vname(KRI(i)),9,10));
                  n_f_kris = n_f_kris + 1;
               end;
            end;
         end;
         call symputx( "SEV_N_F_KRIs", n_f_kris);
         call symputx( "SEV_FIRM_KRI_LIST", firm_kri_list);
         call symputx( "SEV_N_RC_KRIs", n_rc_kris);
         call symputx( "SEV_RC_KRI_LIST", RC_KRI_LIST);
      run;
      %if &SEV_N_RC_KRIs gt 0 %then %do;
         /* Now for each severity risk category, have a different Param_1, 
         so figure out how many there are and then grab the risk category IDS 
         (e.g. internal_bl_uid's).  Start by getting a list of the risk categories */
 
         /* Need Weights associated with each risk category.  Also get the list
            of the loss grouping IDs */
         data _vg_sev_kridata;
            set &lossi; where &CELL_WHERE;
         run;
         proc sort data=_vg_sev_kridata; 
            by &SEV_RC_TYPE; 
         run;
         proc summary data= _vg_sev_kridata;
            by &SEV_RC_TYPE;
            output out=_vg_sev_krisum;
         run;
         /* Get the total and the # of risk categories */
         data _NULL_;
            set _vg_sev_krisum end=last;
            retain sum 0;
            sum = sum+_freq_;
            if last then do;
               call symputx("nlosses",sum);
               call symputx("nsev_mix",_N_);
            end;
         run;

         /* Adjust the frequencies in the summary data so that they sum to 1 (these are
            the weights) */
         data _vg_sev_krisum;
            set _vg_sev_krisum;
            weight = _FREQ_ / &nlosses;
         run;

         /* Put the weights and the loss grouping IDs into macros */
         proc sql noprint; 
            SELECT 
                     &Sev_rc_type, 
                     Weight
            INTO 
                     :Sev_rc_ids SEPARATED BY " ",
                     : Sparam1_weights SEPARATED BY " "
            FROM 
                     _vg_sev_krisum
         quit;

         /* Get the severity scale parameters for each loss group.  Do this by 
            combining kribl_sim and krifirm_sim and then constructing the parameters
            'from scratch' */
         %let rc_kri_sims = %str();
         %ORVU_find_dataset( namein=kribl_sim, libs = VARPRJ VARPRJ_M, nameout=rc_kri_sims);
         data _vg_sev_rc_kris;
            set &rc_kri_sims;
            where BLUID_KRI in ( &SEV_RC_IDS );
         run;
      %end;
      %else %do;
         /* This will be used later- will have a scale parameter that is not SParam_1
            because there are firm KRIs */
         %let SEV_RC_IDS = 0;
         %let SPARAM1_WEIGHTS = 1;
         %let nsev_mix = 1;
      %end;
      %if &SEV_N_F_KRIs gt 0 %then %do;
         %let firm_kri_sims = %str();
         %ORVU_find_dataset( namein=krifirm_sim, libs = VARPRJ VARPRJ_M, nameout=firm_kri_sims);
         data _vg_sev_firm_kris;
            set
            /* When we merge the firm and rc kri datasets, they need to be the same size */
            %do i=1 %to &nsev_mix.;
               &firm_kri_sims
            %end; ;
         run;
      %end;

      /* Make sure variables are sorted before merging */
      %if &SEV_N_RC_KRIs gt 0 %then %do;
         proc sort data = _vg_sev_rc_kris;
            by RISK_FACTOR_VARIABLE_DATE;
         run;
      %end;
      %if &SEV_N_F_KRIs gt 0 %then %do;
         proc sort data = _vg_sev_firm_kris;
            by RISK_FACTOR_VARIABLE_DATE;
         run;
      %end;

      data _vg_sev_kris_combined;
         MERGE  
         %if &SEV_N_RC_KRIs gt 0 %then %do;
            _vg_sev_rc_kris 
         %end;
         %if &SEV_N_F_KRIs gt 0 %then %do;
            _vg_sev_firm_kris
         %end; ;
         BY RISK_FACTOR_VARIABLE_DATE; 
      run;

      /* Now go through and compute the scale for each business line */
      %let found__error = false;
      data _vg_sev_kris_combined;
         set _vg_sev_kris_combined end=last;
         retain max_theta 0;
         %if %sysevalf(&SCALE_TRANSFORM eq ID, boolean) %then %do; 
            /* Check that we can compute exponential */
            if log(&SParam_1) + &sev_kri_model gt log(&INFINITY) then do;
               call symput("found__error","true");
            end;
            else do;
               theta = &SParam_1.*exp(&sev_kri_model.);
            end;
         %end;
         %else %do;
            /* Check that we can compute exponential */
            if &SParam_1 + &sev_kri_model gt log(&INFINITY) then do;
               call symput("found__error","true");
            end;
            else do;
               theta = &SParam_1.+&sev_kri_model.;
            end;
         %end;
         max_theta = max(theta,max_theta);
         if last then do;
            call symputx("MAX_P1",put(max_theta,BEST20.));
         end;
      run;
      %if &found__error eq true %then %do;
         %let reason = &orvar_warning %sysfunc(sasmsg(&orvads, vargrad_scale_too_large_error, noquote, &ccuid)));
         %goto DONE;
      %end;

      /* Put the scales into a macro */
      proc sql noprint; 
         SELECT 
                  theta
         INTO 
                  :SEV_SCALES SEPARATED BY " "
         FROM 
                  _vg_sev_kris_combined
      quit;
   %end;
   /* If there are severity  risk category KRIs, then the severity distribution is a 
      mixture over risk categories. To avoid continually conditioning on KRIs, 
      treat the case with no KRIs like there are KRIs, but only 1 risk category. */
   %else %do;
         %let SEV_N_RC_KRIs   = 0;
         %let SEV_SCALES      = &Sparam_1;
         %let MAX_P1          = &Sparam_1;
         %let nsev_mix        = 1;
         %let SPARAM1_WEIGHTS = 1;
   %end;
   /*****************************************************************/
   /*         Get the Frequency Distribution Information            */
   /*****************************************************************/

   /* Get datasets where information is stored */
   %let freq_dsname = none;
   %let freq_rc_dsname = none;
   data _NULL_;
      set &fpriority;
      where CCUID=&CCUID and SOURCE_TYPE = "I";
      /* Dataset with  iteration, distribution and parameters, selected KRIs and coefficients, # of KRIs  ( selfparmi ) */
      call symput("freq_dsname",dataset_name);
      /* Dataset with loss grouping specific KRI coeficients ( fparm_mdli ) */
      call symput("freq_rc_dsname", fmodels_ds);
   run;

   /*--- Must have frequency distribution info -------------*/
   %if &freq_dsname eq none %then %do;  
      %let reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_freq_info_error, noquote, &ccuid));
      %goto DONE;
   %end;
   %orvl_ds_exists(&freq_dsname, "orva_var_gradient", false);
   %if &ds_nobs eq 0 %then %do;
      %let reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_freq_info_error, noquote, &ccuid));
      %goto DONE;
   %end; 

   /*--- Must have frequency KRI info -------------*/
   %if &freq_rc_dsname eq none %then %do;  
      %let reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_freq_kri_info_error, noquote, &ccuid));
      %goto DONE;
   %end;
   %orvl_ds_exists(&freq_rc_dsname, "orva_var_gradient", false);
   %if &ds_nobs eq 0 %then %do;
      %let reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_freq_kri_info_error, noquote, &ccuid));
      %goto DONE;
   %end; 

   /* Pull out the distribution and KRI information */
   data _NULL_;
      set &freq_dsname;
      where ( (CCUID=&CCUID) and (seviter=&siteration) );
      call symputx("fiteration",iteration);
      fdist = strip(distribution);
      call symput("fdist", fdist);
      if count_kri = . then count_kri=0;
      call symputx("freq_KRI_count",count_kri);
      freq_KRIs = strip(SelectedKRIs);
      call symput("freq_KRIs",freq_KRIs);
      FREQ_RC_TYPE = strip(Freq_Model_Basis_Props);
      call symput("FREQ_RC_TYPE ",FREQ_RC_TYPE);
      if COUNT_DUMMIES = . then COUNT_DUMMIES=0;
      call symputx("freq_dummy_count",COUNT_DUMMIES);
      call symputx("fparamcount",paramcount);
      Freq_Fit_Interval = strip(freq_fit_interval);
      call symput("Freq_Fit_Interval",Freq_Fit_Interval);
   run;

   /* Load the distribution macros */
   %orvd_&fdist;

   %let status = 0;
   %let fit_sim_ratio = 1;
   /* Get the model ratio */
   data _NULL_;
      fit_sim_ratio = intck("&Freq_Fit_Interval", 0, 366) / intck("&SimInterval", 0, 366);
      call symput('fit_sim_ratio', put(fit_sim_ratio, BEST20.));
   run;

   /* Get the parameters used in simulation- stored in feffects.  Adjust for fit
      interval simulation interval diffs if necessary */
   %let freq_parm_ds = %str();
   %ORVU_find_dataset( namein=feffects, libs = VARPRJ VARPRJ_M, nameout=freq_parm_ds);
   data _NULL_;
      set &freq_parm_ds;
      where CCUID=&CCUID;

      if &fit_sim_ratio NE 1 then do;
         /* adjust for interval */
         call ChangeInterval(distribution, &fit_sim_ratio, &theFreqSimCallParams);
      end;

      %do i=1 %to &fparamcount.;
         call symputx("FPARAM&i.",put(FPARAM_&i.,BEST20.));
      %end;
   run;


   %if ( (&FREQ_KRI_COUNT gt 0) or (&freq_dummy_count gt 0) ) %then %do;
   
      /* Need to distinguish between firm level and risk category level KRIs */
      data _vg_selfparm;
         set &freq_dsname;
         where ( (CCUID=&CCUID) and (seviter=&siteration) );
         drop kri_intercept;
      run;
      data _null_;
         set _vg_selfparm;
         length firm_kri_list $2000;
         length rc_kri_list $2000;
         n_f_kris = 0;
         n_rc_kris = 0;
                        %if &freq_KRI_count gt 0 %then %do;
         array kri[*] KRI_:;
         do i=1 to dim(kri);
            if kri[i] ne . then do;
               y = trim(vname(kri[i]));
               z = length(y);
               if z eq 8 then do;
                  rc_kri_list = trim(rc_kri_list) || "  " || trim(vname(KRI(i)));
                  n_rc_kris = n_rc_kris + 1;
               end;
               else do;
                  firm_kri_list = trim(firm_kri_list) || "  " || trim(vname(KRI(i)));
                  n_f_kris = n_f_kris + 1;
               end;
            end;
         end;
                        %end;
         call symputx( "FREQ_N_F_KRIs", n_f_kris);
         call symputx( "FREQ_FIRM_KRI_LIST", firm_kri_list);
         call symputx( "FREQ_N_RC_KRIs", n_rc_kris);
         call symputx( "FREQ_RC_KRI_LIST", RC_KRI_LIST);
      run;

      /* Now for each frequency risk category, have a different mean, which will
         be needed later when we take the derivative wrt KRI parameters.  Compute
         them now */
       
      /* If there are risk category level KRIs or dummies, grab the stratum model (mean_i = ... )  */
      %if ( (&FREQ_N_RC_KRIs gt 0) or (&freq_dummy_count gt 0) ) %then %do;
         data _vg_modeldata;
            set &freq_rc_dsname  end=last;
            where ( (CCUID=&CCUID) and (iteration=&fiteration) );
            StratumModel = compress(StratumModel);
            if last then call symputx("nfreq_groups",_N_);
         run;

         proc sql noprint;
            select &FREQ_RC_TYPE, StratumModel 
            into: FREQ_RC_IDS separated by ' ',: freq_kri_model separated by ' '
            from _vg_modeldata
         quit;
         /* If there are risk category level KRIs, get their values in simualtion */
         %if &FREQ_N_RC_KRIs gt 0 %then %do;
            %let rc_kri_sims = %str();
            %ORVU_find_dataset( namein=kribl_sim, libs = VARPRJ VARPRJ_M, nameout=rc_kri_sims);
            data _vg_freq_rc_kris;
               set &rc_kri_sims;
               where BLUID_KRI in ( &FREQ_RC_IDS );
               drop RISK_FACTOR_VARIABLE_DATE;
            run;
         %end;
      %end;
      %else %do;
         %let nfreq_groups = 0;
      %end;
      %if &FREQ_N_F_KRIs gt 0 %then %do;
         %let firm_kri_sims = %str();
         %ORVU_find_dataset( namein=krifirm_sim, libs = VARPRJ VARPRJ_M, nameout=firm_kri_sims);
         data _vg_freq_firm_kris;
            %if ( (&FREQ_N_RC_KRIs gt 0) or (&freq_dummy_count gt 0) ) %then %do;
               set
               %do i=1 %to &nfreq_groups.;
                  &firm_kri_sims
               %end; ;
            %end;
            %else %do;
               set  &firm_kri_sims;
            %end;
            drop RISK_FACTOR_VARIABLE_DATE;
         run;
         /* Add in the bluids- they are assumed to be present in freq_kri_grad macro*/
         data _vg_bluids;
            %do i=1 %to &nfreq_groups;
               %let id = %scan(&FREQ_RC_IDS,&i,' ');
               BLUID_KRI = &ID;
               output;
            %end;
         run;
         data _vg_freq_firm_kris;
            set _vg_freq_firm_kris;
            set _vg_bluids;
         run;
      %end;
      /* Dataset on which to compute the stratum means */
      %if ( (&FREQ_N_RC_KRIs eq 0) and (&FREQ_N_F_KRIs eq 0) ) %then %do;
      /* There are no kri's simulated, so create a dummy dataset.  The stratum means
         will have equations like: lambda_i = exp(int + dummy) */
            data _vg_freq_kris_combined;
               %do i=1 %to &nfreq_groups;
                  %let ID = %scan(&FREQ_RC_IDS,&i,' ');
                  BLUID_KRI = &ID.;
                  output;                  
               %end;
            run;      
      %end;
      %else %do;
      /* Pull the KRI data together to compute the stratum means.  The stratum means
         will hva equations like: lambda_i = exp(int + beta*KRI_#### + dummy) */
         data _vg_freq_kris_combined; 
            %if &FREQ_N_RC_KRIs gt 0 %then %do;
               set _vg_freq_rc_kris;
            %end;
            %if &FREQ_N_F_KRIs gt 0 %then %do;
               set _vg_freq_firm_kris;
            %end;;
         RUN;
      %end;
      %if ( (&FREQ_N_RC_KRIs gt 0) or (&freq_dummy_count gt 0) ) %then %do;
         /* Compute the stratum mean for each Risk Category */
         %do i=1 %to &nfreq_groups;
            %let model = %scan(&freq_kri_model,&i,' ');
            data _vg_freq_kris_combined;
               set _vg_freq_kris_combined;
               if _N_ = &i then do;
                  stratum_mean = &model;
               end;
            run;
         %end;

         /* Store them in a macro */
         proc sql noprint;
            select Stratum_Mean 
            into: FREQ_STRAT_MEANS separated by ' '
            from _vg_freq_kris_combined
         quit;
      %end;
      %else %do;
         data _NULL_;
            %do i=1 %to &FPARAMCOUNT;
               FParam_&i = &&FPARAM&i.;
            %end;
            %Moments&FDIST();
            call symputx("FREQ_STRAT_MEANS" , mean);
      %end;

      %if &freq_dummy_count gt 0 %then %do;
         /* Get a list of the dummies, their values, and the associated risk category IDs */
         data _null_;
            set &freq_dsname;
            where ccuid = &ccuid.;
            if _N_ = 1 then do;
               array  DUMMY DUMMY_:;
               length DUMMY_LIST $2000;
               length DUMMY_VALS $2000;
               length DUMMY_RCS $2000;
      
               do i=1 to dim(DUMMY);
                  if DUMMY[i] NE . then do;
                     DUMMY_LIST = trim(DUMMY_LIST) || " " || trim(vname(DUMMY(i)));
                     DUMMY_VALS = trim(DUMMY_VALS) || " " || DUMMY(i);
                     DUMMY_RCS  = trim(DUMMY_RCS)  || " " || trim( compress(lowcase(vname(DUMMY(i))),'abcdefghijklmnopqrstuvwxyz.-_ '));
                  end;
               end;
               call symputx( "FREQ_DUMMY_LIST", DUMMY_LIST );
               call symputx( "FREQ_DUMMY_VALS", DUMMY_VALS );
               call symputx( "FREQ_DUMMY_RCS", DUMMY_RCS );
            end;
         run;
      %end;
   %end;
   %else %do;
      %let FREQ_N_F_KRIs = 0;
      %let FREQ_N_RC_KRIs = 0;
      data _null_;
         %do i=1 %to &fparamcount;
            fparam_&i = &&fparam&i.;
         %end;
         %Moments&fdist.();
         call symputx("FREQ_STRAT_MEANS",put(mean,BEST20.));
         call symputx( "FREQ_FIRM_KRI_LIST", " ");
         call symputx( "FREQ_RC_KRI_LIST", " ");
      run;
   %end;

   /**************************************************/
   /*             Now get the gradient               */
   /**************************************************/

   /* Determine the largest VaR value */
   data _NULL_;
      V = max(&P_1, &P_2, &P_3);
      call symput("Var_Max", put(V,Best20.));
   run;
   /* If all VaR are 0 or missing, then skip to the end */
   %if %sysevalf(&Var_Max le 0, boolean) %then %do;
      %let reason = &orvar_warning %sysfunc(sasmsg(&orvads, vargrad_no_var_gt_zero_error, noquote, &ccuid));
      %goto DONE;
   %end;

   /* Check the largest VaR value against 99.9% quantile of the severity distribution.  Use the larger of the 2
      to determine the region over which to perform the discretization.  This becomes important when the severity
      scale is so large that the maxloss truncates most observations. */
   data _temp_;
      threshold = 0;
      ecdf = .999;
   run;
   %let param_1 = &MAX_P1.;
   %do i=2 %to &SPARAMCOUNT.;
      %let PARAM_&i. = &&Sparam_&i.;
   %end;
   %QUANT&SDIST();
   data _NULL_;
      set _temp_;
      call symputx("sev_upper_bnd",put(quant,BEST20.));
   run;  

   %let upper_bnd = %sysfunc(max(&sev_upper_bnd , &Var_Max));
   /* Will have 1 scale parameter for each risk category, so compute total # of severity parameters in the mixture 
      model */
   %let SPARM_TOT = %eval(&SPARAMCOUNT + &NSEV_MIX -1);

   /* total number of columns we will need in results */
   %let tot_cols = %eval(&SPARM_TOT + &FPARAMCOUNT + 2);

   /* Determine the number of intervals to use in discretization */
   /* use 2**n points. */
   %let pwr = 9;
   %let npoints = %eval(2**&pwr);
   %let h = %sysevalf((&upper_bnd/(&npoints/2)));

   /* For Body/Tail distributions, we need to be careful.  It is possible to have h (computed above) be larger than 
      the body/tail boundary point.  In this case, no points of the discretized severity distribution (computed later)
      or of derivs_indata (computed next) will lie in the body of the distribution, and errors will occur. So, check 
      if this happens.  If it does, we need to use more discretization points and possibly not use as much padding 
      (the standard padding has 75% of the discretized points lying past the largest VaR value) */
   %if %trim(%upcase(&SDIST)) eq MIXDIST %then %do;
   /* Loop over the severity distributions (may have a mixture of several when there are KRIs) and find the 
      smallest body/tail boundary value. */
      %let P1 = %scan(&SEV_SCALES, 1, ' ');
      /* It is possible that with the projected KRIs, the severity scale is so large that the exponential
         function of it is too big for the machine to handle, e.g. exp(P1) for P1 > 800 is a problem.  So,
         need to check that this isn't going to be the case.  If it is, we won't be able to properly take 
         derivatives of the severity cdf, so abort. */
      %let Min_Tail_Boundary = &INFINITY;
      %do w=1 %to &NSEV_MIX;
         %let P1 = %scan(&SEV_SCALES, &w, ' ');
         %if &SCALE_TRANSFORM = ID %then %do;
            %let Min_Tail_Boundary = %sysfunc(min(%sysevalf(&Param_xLimit * &P1),&Min_Tail_Boundary));
         %end;
         %else %do;
            %if %sysevalf( %sysfunc(log(&Param_xLimit)) + &P1. gt %sysfunc(log(&INFINITY)),BOOLEAN) %then %do;
               /* We will need to abort */
               %let reason = &orvar_warning %sysfunc(sasmsg(&orvads, vargrad_scale_too_large_error, noquote, &ccuid));
               %goto DONE;
            %end;
            %else %do;
               %let Min_Tail_Boundary=%sysfunc(min(%sysevalf(&Param_xLimit*%sysfunc(exp(&P1))),&Min_Tail_Boundary));
            %end;
         %end;
      %end;
   %end;
   /* For other distributions, we want at least 5 points to lie in the 'body' - below the  95%
      quantile.  Otherwise derivatives will be very inaccurate */
   %else %do;
      data _temp_;
         threshold = 0;
         ecdf = .95;
      run;
      %QUANT&SDIST();
      data _NULL_;
         set _temp_;
         call symputx("min_tail_boundary",quant);
      run;
   %end;
   %if %upcase(&debug) ne FALSE %then %do;
      %put MIN TAIL BOUNDARY IS: &Min_Tail_Boundary;
      %put INTERVAL WIDTH WITH POWER = &PWR IS: &h;
      %let rat = %sysevalf(&Min_Tail_Boundary/&h);
      %put RATIO is: &rat.;
   %end;
   %if %sysevalf(5*&h > &Min_Tail_Boundary,BOOLEAN) %then %do;
      /* There are too few discretization points in the body of the distribution. See how many discretization 
         points will be needed to get at least 5 in the body.  If too many, use another method */
      %let pwr = %sysfunc(ceil(%sysevalf(%sysfunc(log(10*&upper_bnd/&Min_Tail_Boundary))/%sysfunc(log(2)))));
   %end;

   /* Status macro for fft_param_grad and normal_param_grad */
   %let abort_pg = 1;

   %if %sysevalf(&pwr gt 11, BOOLEAN) %then %do;
   /* In this case, can not get enough points in the body without using a LOT of points which
      is too computationally expensive (and sometimes impossible), so use the normal approximation method */
      %if %upcase(&debug) ne FALSE %then %do;
         %put RESORTING TO NORMAL APPROXIMATION OF AGGREGATE FOR PARAMETER VARIANCE CI CALCULATION;
      %end;
      %normal_param_grad(varpct=_vg_var_, 
                         outgrad=_vg_pgrad,
                         abort_npg=abort_pg,
                         abort_reason=reason );
   %end;
   %else %do;
      /* Compute the number of discretization points and the interval width */
      %let npoints = %sysevalf(2**&pwr);
      %if %sysevalf(5*&h > &Min_Tail_Boundary,BOOLEAN) %then %do;
         /* DEBUG INFO */
         %if %upcase(&debug) ne FALSE %then %do;
            %put ---------------- CALLING FFT_PARAM_GRAD -----------------;
            %put DIST IS: &SDIST;
            %put POWER IS: &PWR (npoints is %sysevalf(2**&pwr));
         %end;
      %end;
      %let h = %sysevalf((&upper_bnd/(&npoints/2)));
      
      %fft_param_grad(varpct=_vg_var_, 
                      n_disc_points= &npoints,
                      disc_int_width= &h,
                      outgrad=_vg_pgrad,
                      abort_fpg = abort_pg,
                      abort_reason = reason );
   %end;

   /* Check for errors */
   %if &abort_pg %then %do;
      %goto DONE; 
   %end;

   /* Add in the gradient effects of the KRIs */
   %if %sysevalf(&SEV_KRI_COUNT >0, boolean) %then %do;
      %sev_kri_gradient( firm_kri_data=_vg_sev_firm_kris, 
                         N_F_KRIs=&SEV_N_F_KRIs,
                         firm_kri_list=&SEV_FIRM_KRI_LIST,
                         rc_kri_data=_vg_sev_rc_kris, 
                         N_RC_KRIs=&SEV_N_RC_KRIs,
                         RC_KRI_LIST=&SEV_RC_KRI_LIST,
                         prefix=DS_, 
                         scales=&sev_scales, 
                         ingrad=_vg_pgrad, 
                         outgrad=_vg_pgrad
                         );
   %end;
   %freq_kri_gradient( firm_kri_data=_vg_freq_firm_kris,
                       N_F_KRIs=&FREQ_N_F_KRIs,
                       firm_kri_list=&FREQ_FIRM_KRI_LIST,
                       N_RC_KRIs=&FREQ_N_RC_KRIs,
                       RC_KRI_LIST=&FREQ_RC_KRI_LIST,
                       N_DUMMIES=&FREQ_DUMMY_COUNT,
                       combined_kri_data=_vg_freq_kris_combined, 
                       prefix=DF_, 
                       means=&FREQ_STRAT_MEANS, 
                       ingrad=_VG_pgrad, 
                       outgrad=_vg_pgrad
                     );
   data &sev_gradient;
      set _vg_pgrad;
      drop 
      %if ( (&SEV_N_RC_KRIs eq 0) and (&sev_KRI_count gt 0) ) %then %do; 
        DSparam_1_0 
      %end; 
      df: ;
   run;
   data &freq_gradient;
      set _vg_pgrad;
      drop ds:;
   run;

   /* Store gradient data along with model data for use later (KRI Slider Bar calcualtion */
   /* First the severity info */
   /* Collect the KRIs names and their values, along with distribution info*/
   data _vg_sev_added_info;
      set &sev_dsname;
      where CCUID=&CCUID;
   run;

   data _vg_sev_added_info;
      length scale_transform $ 12;
      set _vg_sev_added_info;
      scale_transform = "&SCALE_TRANSFORM";
      keep
      %do i=1 %to &sparamcount;
         param_&i
      %end;
      %if &sev_KRI_count gt 0 %then %do;
         param_1_:
      %end;
      paramcount distribution ccuid scale_transform; 
   run;

   data &sev_gradient;
      set
      %do i=1 %to 3;
         _vg_sev_added_info
      %end; ;
      set &sev_gradient;
   run;

   /* Now store in VARPRJ */
   %let sev_grad_ds = %str();
   %ORVU_find_dataset( namein=sev_gradient, libs = VARPRJ VARPRJ_M, nameout=sev_grad_ds);
   %if &sev_grad_ds NE %str() %then %do;
      data varprj.sev_gradient;
         set &sev_grad_ds;
      run;
      /* TODO: eventually, java will delete sev_gradient and freq_gradient prior to simulating, so this will
               be unnecessary */
      proc sql noprint;
         delete from VARPRJ.sev_gradient where ccuid = .;
         delete from VARPRJ.sev_gradient where ccuid = &ccUid;
      quit;
      data varprj.sev_gradient;
         set VARPRJ.sev_gradient &sev_gradient;
      run;
   %end;
   %else %do;
      data varprj.sev_gradient;
         set &sev_gradient;
      run;
   %end;
   proc sort data=VARPRJ.sev_gradient; by ccuid quantile; run;

   /* Now Frequency info */
   /* Collect the KRIs names and their values, along with distribution info*/
   data _NULL_;
      set &freq_dsname;
      where CCUID=&CCUID;
      if &freq_KRI_count gt 0 then do;
         call symput("SelectedKRIs",SelectedKRIs);
      end;
   run;

/* First rename KRI coefficients to distinguish them from KRI values (insert Param_1_
   in the beginning) Frequency only- severity KRIs already labeled this way */
   data _vg_freq_added_info;
      set &freq_dsname;
      where ( (CCUID = &CCUID.) and (seviter = &siteration) ); 
      %if &freq_KRI_count gt 0 %then %do;
         length new_kri_list $2000;
         array kri[*] &SelectedKRIs;
         do i=1 to dim(kri);
            new_kri_list = trim(new_kri_list) || " " || "Param_1_" || trim(vname(kri[i]));
         end;
         call symput("new_list",new_kri_list);
      %end;
   run;

   data _vg_freq_added_info;
      set _vg_freq_added_info;
      %if &freq_KRI_count gt 0 %then %do;
         array old_kri[*] &SelectedKRIs;
         array new_kri[*] &new_list.;
         do i=1 to dim(old_kri);
            new_kri[i] = old_kri[i];
         end;
      %end;
      keep  Fparam_:   
            paramcount 
            distribution 
            ccuid 
            %if &freq_KRI_count gt 0 %then %do;
            Param_1_:
            %end;
            intercept 
            %if &freq_dummy_count NE 0 %then %do;
               dummy: 
            %end;
            selectedKRIs 
            ccuid 
            seviter;
   run;

   data &freq_gradient;
      set
      %do i=1 %to 3;
         _vg_freq_added_info
      %end; ;
      set &freq_gradient;
   run;

   /* Now store in VARPRJ */
   %let freq_grad_ds = %str();
   %ORVU_find_dataset( namein=freq_gradient, libs = VARPRJ VARPRJ_M, nameout=freq_grad_ds);
   %if &freq_grad_ds NE %str() %then %do;
      data VARPRJ.freq_gradient;
         set &freq_grad_ds;
      run;
      /* TODO: eventually, java will delete sev_gradient and freq_gradient prior to simulating, so this will
               be unnecessary */
      proc sql noprint;
         delete from VARPRJ.freq_gradient where ccuid = .;
         delete from VARPRJ.freq_gradient where ccuid = &ccUid;
      quit;
      data varprj.freq_gradient;
         set VARPRJ.freq_gradient &freq_gradient;
      run;
   %end;
   %else %do;
      data varprj.freq_gradient;
         set &freq_gradient;
      run;
   %end;
   proc sort data=VARPRJ.freq_gradient; by ccuid quantile; run;

   %let &abort_flag = 0;

%DONE:
   /* ALL OF THE ERROR HANDLING IS DONE BY ORVA_CONF_INTERVAL.  MESSAGES ARE DISPLAYED THERE */

   %if %upcase(&debug) eq FALSE %then %do;
      proc datasets library=work memtype=data nolist nodetails;
         delete _vg_: ;
      quit;
   %end;

%mend;


%macro freq_kri_gradient( firm_kri_data=,    /* data where firm level KRIs are stored */
                         N_F_KRIs=,          /* Number of firm level KRIs */
                         firm_kri_list=,     /* space delimited string of firm level KRIs (names) */
                         N_RC_KRIs=,         /* Number of risk category level KRIs */
                         RC_KRI_LIST=,       /* space delimited string of risk category level KRIs (names) */
                         N_DUMMIES=,
                         combined_kri_data=, /* data set with firm and risk category level KRIs and stratum means */
                         prefix=, 
                         means=, 
                         ingrad=, 
                         outgrad=
                       );

   /* Overall Mean (sum over strata) */
   proc fcmp;
      array rc_means[*] &MEANS;
      mean = 0;
      do i=1 to dim(rc_means);
         mean = mean+rc_means[i];
      end;
      call symputx("lambda",put(mean,BEST20.));
   run;

   /*Need The change in Var wrt Mean */
   data _null_;
      %do i=1 %to &fparamcount;
         FPARAM_&i. = &&Fparam&i.;
      %end;
      %MomDer&FDIST.(deriv=dfparam1_dmean);
      set &ingrad;
      DFMean = dfparam1_dmean*DFParam_1;
      if _N_ = 1 then do;
         call symputx("DFParam_1_1",put(DFMean,BEST20.));
      end;
      if _N_ = 2 then do;
         call symputx("DFParam_1_2",put(DFMean,BEST20.));
      end;
      if _N_ = 3 then do;
         call symputx("DFParam_1_3",put(DFMean,BEST20.));
      end;
   run;

   %if &N_F_KRIs gt 0 %then %do;
      /* Get the derivatives for the firm level kris */
      proc fcmp data=&firm_kri_data out = _vg_firm_der;
         array kri[*] &FIRM_KRI_LIST;
         array dkri[*] %do i=1 %to &N_F_KRIs; 
                          %let n=%scan(&FIRM_KRI_LIST,&i);
                          &Prefix&n 
                       %end; ;

         %do var_points = 1 %to 3;
            do i=1 to &N_F_KRIs;
               dkri[i] = &&DFParam_1_&var_points.*&lambda*kri[i] ;
            end;
            output;
         %end;

         keep %do i=1 %to &N_F_KRIs;
                 %let n=%scan(&FIRM_KRI_LIST,&i);
                 &Prefix&n 
              %end; ;
      run;
   %end;
   %if &N_RC_KRIs gt 0 %then %do;
   /* Get the derivatives for the loss group level kris and intercpet*/
      data _vg_rcder_prep;
         set &combined_kri_data.;
         array kri[*] &RC_KRI_LIST;
         array mean_by_kri[*] %do i=1 %to &N_RC_KRIs; 
                          %let n=%scan(&RC_KRI_LIST,&i);
                          M_&n 
                       %end; ;
         retain %do i=1 %to &N_RC_KRIs; 
                          %let n=%scan(&RC_KRI_LIST,&i);
                          M_&n 0
                       %end; ;
         do i=1 to &N_RC_KRIs;
            mean_by_kri[i] = mean_by_kri[i] + stratum_mean*kri[i];
         end;
         %if &FIRM_KRI_LIST ne %str() %then %do;
            drop &FIRM_KRI_LIST;
         %end;

      run;

      data _vg_rcder_prep;
         set _vg_rcder_prep end=last;
         if ~last then delete;
      run;

      proc fcmp data=_vg_rcder_prep out = _vg_rc_der;
         array m_kri[*] M_:;
         array dkri[*] %do i=1 %to &N_RC_KRIs; 
                          %let n=%scan(&RC_KRI_LIST,&i);
                          &Prefix&n 
                       %end; ;
         %do var_points = 1 %to 3;
            do i=1 to &N_RC_KRIs;
               dkri[i] =&&DFParam_1_&var_points.*m_kri[i] ;
            end;
            output;
         %end;

         keep %do i=1 %to &N_RC_KRIs;
                 %let n=%scan(&RC_KRI_LIST,&i);
                 &Prefix&n 
              %end; ;
      run;
   %end;
   %if &N_DUMMIES gt 0 %then %do;
      /* Finally, get derivative wrt DUMMIES */
      data _vg_dummy_combined;
         set &combined_kri_data.;
         where BLUID_KRI in (&FREQ_DUMMY_RCS);
      run;
      proc fcmp data=_vg_dummy_combined out=_vg_dummy_prep;
         array IDS[*] &FREQ_DUMMY_RCS;
         array VALS[*] &FREQ_DUMMY_VALS;
         i=1;
         do while (IDS[i] ne BLUID_KRI);
            i=i+1;
         end;
         D_DUMMY = VALS[i]*stratum_mean;
         output; 
      run;

      proc sql noprint; 
         SELECT 
                  BLUID_KRI, D_DUMMY 
         INTO 
                  :D_IDs SEPARATED BY " ",
                  :D_DUMMIES SEPARATED BY " "
         FROM 
                  _vg_dummy_prep;
      quit;
   
      proc fcmp out = _vg_dum_der;
      
         array ddum[*] %do i=1 %to &N_DUMMIES; 
                          %let n=%scan(&D_IDs,&i);
                          &Prefix.DUMMY_&n 
                       %end; ;
         array dm_dd[*] &D_DUMMIES;
         %do var_points = 1 %to 3;
            do i=1 to &N_DUMMIES;
               ddum[i] =&&DFParam_1_&var_points.*dm_dd[i] ;
            end;
            output;
         %end;

         keep %do i=1 %to &N_DUMMIES; 
                          %let n=%scan(&D_IDs,&i);
                          &Prefix.DUMMY_&n 
                       %end; ;
      run;
   %end;

   data _vg_intercept_der;
      %do var_points = 1 %to 3;
         &prefix.Intercept = &&DFParam_1_&var_points.*&lambda;
         output;
      %end;
   run;

   data &outgrad; 
      set &ingrad; 
      set _vg_intercept_der;
      %if &N_F_KRIs gt 0 %then %do;
         set _vg_firm_der;
      %end;
      %if &N_RC_KRIs gt 0 %then %do;
         set _vg_rc_der;
      %end;
      %if &N_DUMMIES gt 0 %then %do;
         set _vg_dum_der;
      %end;
      run;


      

%mend freq_kri_gradient;

/* for variable descriptions, see freq_kri_gradient macro in this file */
%macro sev_kri_gradient( firm_kri_data=, 
                         N_F_KRIs=,
                         firm_kri_list=,
                         rc_kri_data=, 
                         N_RC_KRIs=,
                         RC_KRI_LIST=,
                         prefix=, 
                         scales=, 
                         ingrad=, 
                         outgrad=
                       );


   %if &N_F_KRIs gt 0 %then %do;
      /* Need the sum of DVaRDScale*Scale for each var value (for firm level KRIs derivs) */
         data _vg_temp;
            set &ingrad;
            array dscales[*] DSParam_1_:;
            sum=0;
            %do i=1 %to &nsev_mix;
               %let scl = %scan(&SCALES,&i,' ');
               sum = sum + dscales[&i]*&SCL;
            %end;
         run;
         proc sql noprint; 
            select sum
            into : RC_FACTOR separated by " "
            from _vg_temp
         run;
         quit;

      /* Get the derivatives for the firm level kris */
      proc fcmp data=&firm_kri_data out = _vg_firm_der;
         array kri[*] &FIRM_KRI_LIST;
         array rc_factors[*] &RC_FACTOR;
         array dkri[*] %do i=1 %to &N_F_KRIs; 
                          %let n=%scan(&FIRM_KRI_LIST,&i);
                          &Prefix&n 
                       %end; ;
         do var_points = 1 to 3;
            do i=1 to &N_F_KRIs;
               dkri[i] = rc_factors[var_points]*kri[i] ;
            end;
            output;
         end;

         keep %do i=1 %to &N_F_KRIs;
                 %let n=%scan(&FIRM_KRI_LIST,&i);
                 &Prefix&n 
              %end; ;
      run;
   %end;

   %if &N_RC_KRIs gt 0 %then %do;
      /* add the scales to the loss group kri dataset */
      data _vg_temp;
         %do i=1 %to &nsev_mix;
            %let rcid = %scan(&SEV_RC_IDS,&i,' ');
            %let scale = %scan(&scales,&i,' ');
            BLUID_KRI = &rcid;
            scale = &scale;
            output;
         %end;
      run;

      /* Get the derivatives for the loss group level kris */
      proc sort data = _vg_temp;
         by BLUID_KRI;
      run;

      proc sort data = &rc_kri_data;
         by BLUID_KRI;
      run;

      data _vg_temp;
        MERGE _vg_temp &rc_kri_data; 
        BY BLUID_KRI; 
      RUN: 


      data _vg_dscale;
         set &ingrad;
         keep DSparam_1_:;
      run;
      proc transpose data=_vg_dscale out=_vg_dscale; run;

      data _vg_temp;
         set _vg_temp;
         set _vg_dscale;
         retain 
         %do i=1 %to &N_RC_KRIs.;
            %let suffix = %scan(&RC_KRI_LIST.,&i,' ');
            &prefix&suffix._1 0
            &prefix&suffix._2 0
            &prefix&suffix._3 0
         %end; ;
         %do i=1 %to &N_RC_KRIs.;
            %let suffix = %scan(&RC_KRI_LIST.,&i,' ');
            &prefix&suffix._1 = &prefix&suffix._1 + COL1*scale*&suffix.;
            &prefix&suffix._2 = &prefix&suffix._2 + COL2*scale*&suffix.;
            &prefix&suffix._3 = &prefix&suffix._3 + COL3*scale*&suffix.;
         %end; ;
      run;

      data _vg_temp;
         set _vg_temp end=last;
         if ~last then delete;
         keep &prefix.:;
      run;

      proc fcmp data = _vg_temp out=_vg_rc_der;
         %do i=1 %to 3;
            %do j=1 %to &N_RC_KRIs.;
               %let suffix = %scan(&RC_KRI_LIST.,&j,' ');
               &prefix.&suffix. = &prefix.&suffix._&i.;
            %end;
            output;
         %end;
         keep %do j=1 %to &N_RC_KRIs.;
                 %let suffix = %scan(&RC_KRI_LIST.,&j,' ');
                 &prefix.&suffix.
              %end; ;
      run;

   %end;
   data &outgrad; 
      set &ingrad; 
      %if &N_F_KRIs gt 0 %then %do;
         set _vg_firm_der;
      %end;
      %if &N_RC_KRIs gt 0 %then %do;
         set _vg_rc_der;
      %end;
   run;
%mend sev_kri_gradient;


/* Use FFT to get derivative of VaR wrt parameters */
%macro fft_param_grad( 
                      varpct=, /* Dataset with VaR values */
                      n_disc_points=, /* Number of discretization points */
                      disc_int_width=,
                      outgrad=,  /* output dataset */
                      abort_fpg=,
                      abort_reason=
                 );

   %let &abort_fpg = 1;
   %let npoints = &n_disc_points;
   %let h = &disc_int_width;

   /* TODO: this code is copied from orva_simulate.  SHould get the result from there and pass it in
            rather than recomputing */
   /*--- Extract MAXLOSS from sPriority Dataset ---*/
   data _NULL_; 
      set &sPriority end=LAST; 
      WHERE CCUID = &ccUID AND SOURCE_TYPE NE "R";
      retain MaximumLoss 0;
      MaximumLoss = max(MaximumLoss,max_loss);
      if LAST then do;
         if MaximumLoss eq 0 then do;
            MaximumLoss=10000000000;
         end;
         call symputX("MAXLOSS",put(MaximumLoss, BEST20.));
      end;
   run;
   /* Get the index of the discretized severity corresponding to maxloss */
   %let max_loss_ind = %sysfunc(round(&maxloss./&h.));
   %let max_loss_ind = %sysfunc(max(&max_loss_ind.,1));

   /* Generate points on which to compute the severity derivatives */
   DATA derivs_indata;
      DO i=1 to &npoints.;
         quant = (2*i-1)*&h./2;
         output;
      end;
   run;

   /* Calculate the derivatives of severity cdf wrt severity parameters */
   /* If there are severity KRIs, may need to do this for each risk category */
   /* The severity macros look for Param_1, not SParam_1*/
   %do i=2 %to &SPARAMCOUNT;
      %let PARAM_&i. = &&Sparam_&i.;
   %end;
   %do i=1 %to &nsev_mix;
      %let Param_1=%scan(&SEV_SCALES, &i, ' ');
      %if %sysevalf(&SPARAMCOUNT > 1, BOOLEAN) %then %do;
         %CdfDer&SDIST.(scalefactor=1, 
                        indata=derivs_indata,
                        outdata=_vg_grad_dist&i);
         /* Check that proc model returned something.  
            TODO: with 9.3 proc model will put return code into in SYSRC, so check that instead */
         %orvl_ds_exists(_vg_grad_dist&i, "orva_var_gradient", false);
         %if &ds_nobs eq 0 %then %do;
            %let &abort_reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_model_output_error, noquote, &CCUID));
            %goto DONE_FFT;
         %end;
      %end;
   %end;

   proc fcmp library=&ORVAR_CMPLIB; 

      /***** DECLARE ARRAYS *****/

      /* var quantiles, var values, output dataset */
      array qv[3] /nosymbols;
      array var_vals[3] /nosymbols;
      array results[3,&tot_cols.]/nosymbols;
      /* Array containing the weights used in the severity mixture */
      array weights[&NSEV_MIX] &SPARAM1_WEIGHTS;

      /* discretized severity- real and imaginary.  Second copy is for freq. derivs */
      array f[1] /NOSYMBOLS;
      array f2[1] /NOSYMBOLS;
      array im_f[1] /NOSYMBOLS;
      array im_f2[1] /NOSYMBOLS;

      /* derivative of aggregate distribution wrt freq. params- real and imaginary */
      %do y = 1 %to &FPARAMCOUNT;
         array dfdf&y[1] /NOSYMBOLS;
         array im_dfdf&y[1] /NOSYMBOLS;
      %end;

      /* discretized aggregate distribution- real and imaginary */
      array g[1] /NOSYMBOLS;
      array im_g[1] /NOSYMBOLS;

      /* derivative of aggregate cdf  at VaR wrt freq. parameters */
      %do y = 1 %to &FPARAMCOUNT;
         array dQdf&y[1] /NOSYMBOLS;
      %end;

      /* derivative of aggreagate cdf at VaR */
      array dQdV[3] /NOSYMBOLS;

      /* indices corresponding to the var values and the quantiles */
      array var_indices[3] /NOSYMBOLS;
      array quant_indices[3] /NOSYMBOLS;

      /* derivative of severity distribution wrt parameters- real and imaginary */
      /* Need 1 set for each scale parameter (loss grouping) */
      %DO i=1 %TO &NSEV_MIX;   
         %DO z=2 %TO &SPARAMCOUNT.;      
            array dfdS&z._&i.[1] /NOSYMBOLS;   
         %END;
      %END;
      /* These will hold the combined (mixed) values */
      %DO z=2 %TO &SPARAMCOUNT;
         array dfdS&z.[1] /NOSYMBOLS;
         array im_dfdS&z.[1] /NOSYMBOLS;

         /* derivative of aggregate severity distribution wrt parameters- real and imaginary */
         array dgdS&z.[1] /NOSYMBOLS;
         array im_dgdS&z.[1] /NOSYMBOLS;

         /* derivative of aggregate cdf wrt severity parameters */
         array dQdS&z.[1] /NOSYMBOLS;
      %end;

      /* dynamically allocate arrays to avoid stack overflow */
      call dynamic_array( f, &Npoints);
      call dynamic_array( f2, &Npoints);
      call dynamic_array( im_f, &Npoints);
      call dynamic_array( im_f2, &Npoints);

      %do y = 1 %to &FPARAMCOUNT;
         call dynamic_array( dfdf&y, &Npoints);
         call dynamic_array( im_dfdf&y, &Npoints);
         call dynamic_array( dQdf&y, &Npoints);
      %end;

      call dynamic_array( g, &npoints);
      call dynamic_array( im_g, &npoints);

      %DO i=1 %TO &NSEV_MIX;   
         %DO z=2 %TO &SPARAMCOUNT.;      
            call dynamic_array(dfdS&z._&i., &Npoints);   
         %END;
      %END;   
      %DO z=2 %TO &SPARAMCOUNT;
         call dynamic_array( dfds&z., &Npoints);
         call dynamic_array( im_dfds&z., &Npoints);
         call dynamic_array( dgds&z., &Npoints);
         call dynamic_array( im_dgds&z., &Npoints);
         call dynamic_array( dQdS&z., &Npoints);
      %end;

      /***** GET QUANTILES *****/
      /* Load the quantiles */
      rc = read_array("&varpct", qv, "quantile");

      var_vals[1] = &P_1;
      var_vals[2] = &P_2;
      var_vals[3] = &P_3;

      /* Figure out the order of the quantiles (they ar now provided in order) */
      quant_indices[1] = 3;
      quant_indices[2] = 2;
      quant_indices[3] = 1;

      /***** GET indices *****/
      /* Find the indices of of g corresponding to the VaR values */
      do i=1 to 3;
         var_indices[i] = max(floor(var_vals[quant_indices[i]]/&h+1),1);
      end;
      /***** LOAD DERIVATIVES *****/
      /*Load in the derivatives of the f's wrt sev. params */
      %DO i=1 %TO &NSEV_MIX;
         %DO z=2 %TO &SPARAMCOUNT.;
            rc = read_array("_vg_grad_dist&i", dfdS&z._&i., "DParam_&z.");
         %end;
      %END;

      /* We actually loaded derivative of Fx wrt parameters.
         f[i] = Fx[i+1] - Fx[i], where Fx is cdf of sev. dist. */
      DO i=0 to &Npoints.-2;
         %DO i=1 %TO &NSEV_MIX;
            %do z=2 %to &SPARAMCOUNT.;
               dfdS&z._&i.[&Npoints.-i] = dfdS&z._&i.[&Npoints.-i] - dfdS&z._&i.[&Npoints.-i-1];
            %end;
         %END;
      end;

      /* Now get the values for the full mixture */
      DO i=1 to &Npoints.;
         %DO z = 2 %TO &SPARAMCOUNT;
            dfdS&z[i] = 0;
            %DO i=1 %to &NSEV_MIX;
               dfdS&z[i] = weights[&i]*dfdS&z._&i.[i];
            %END;
         %END;
      end;

      /***** DISCRETIZE SEVERITY *****/
      %if %trim(%upcase(&SDIST)) eq MIXDIST %then %do;
          param_xlimit = &param_xlimit;
          param_pnlimit = &param_pnlimit;
      %end;
      DO i=1 to &Npoints.;
         f[i] = 0; 
      end;
      %DO i=2 %TO &SPARAMCOUNT;
         Param_&i = &&SPARAM_&i.;
      %END;
      %DO w=1 %TO &NSEV_MIX;
         Param_1 = %scan(&SEV_SCALES, &w, ' ');
         %VCdf&SDIST.(&h/2, prior);
         f[1]=f[1]+weights[&w]*prior;
         DO i=2 to &Npoints.;
            %VCdf&SDIST.(&h*(2*i-1)/2, next);
            f[i] = f[i] + weights[&w]*(next - prior); 
            prior = next;
         end;
      %END;

      /* Initialize imaginary component */
      do i=1 to &Npoints.;
         im_f[i] = 0;
      end;

      /* Correct for truncation due to maxloss.  All mass to the right of maxloss gets placed at maxloss.  
         Maxloss is treated as fixed and not a function of the parameters for convenience  */
      do i=(&max_loss_ind.+1) to &Npoints.;
         f[&max_loss_ind.] = f[&max_loss_ind.] + f[i];
         f[i] = 0;
         %do z=2 %to &SPARAMCOUNT;
            dfdS&z[&max_loss_ind.] = dfdS&z[&max_loss_ind.] + dfdS&z[i];
            dfdS&z[i] = 0;
         %end;
      end;

      /***** AGGREGATE DISTRIBUTION *****/
      /* Transform */
      rc = zfft(f,im_f,&Npoints.,&Npoints.,&Npoints.,-1);
      do i=1 to &Npoints.;
         g[i] = f[i];   im_g[i] = im_f[i];
         f2[i] = f[i];  im_f2[i] = im_f[i];
      end;

      /* Apply pgf */
      call pgf&FDIST(%do y=1 %to &FPARAMCOUNT; &&FPARAM&y, %end; g,im_g,&Npoints.);
      
      /* Inverse Transform */
      rc = zfft(g,im_g,&Npoints.,&Npoints.,&Npoints.,1);

      /* Rescale */
      do i=1 to &Npoints.;
         g[i] = g[i]/&Npoints.;
      end;
      /***** DQDVaR *****/
      do i=1 to 3;
         if g[var_indices[i]] eq 0 then do;
            var_indices[i] = 1;
         end;
         if var_indices[i] ne 1 then do;

            dQdV[quant_indices[i]] = g[var_indices[i]]/&h;
         end;
      end;

      /***** GET DER. OF QUANTILE WRT TO PARAMETERS *****/

      /***** FREQ DERIVS *****/
      /* Frequency parameter- we already computed transform of f */
      /* Apply der. of pgf wrt frequency params*/
      %do y=1 %to &FPARAMCOUNT;
         call pgfder&FDIST(%do y=1 %to &FPARAMCOUNT; dfdf&y,im_dfdf&y,&&FPARAM&y, %end; f,im_f,&Npoints.);
      %end;
      /* Inverse Transform (dfdl becomes dgdl, f is severity pmf and g is aggregate pmf) */
      %do y=1 %to &FPARAMCOUNT;
         rc = zfft(dfdf&y,im_dfdf&y,&Npoints.,&Npoints.,&Npoints.,1);
      %end;
      /* Rescale */
      do i=1 to &Npoints.;
         %do y=1 %to &FPARAMCOUNT;
            dfdf&y[i] = dfdf&y[i]/&Npoints.;
         %end;
      end;      

      /***** SEV DERIVS *****/
      /* Severity parameter- start with the derivative of f wrt parameters.  Do not 
         transform, as there is no good way to predict the tail behavior of the 
         derivatives (see Embrechts and Frei 2009 for description of how one can
         use a transformation to stabilize FFT) */
      do i=1 to &Npoints.;
         %do z=2 %to &SPARAMCOUNT;
            im_dfds&z.[i] = 0;   
         %end;
      end;

      %do z=2 %to &SPARAMCOUNT;
         rc = zfft(dfdS&z.,im_dfdS&z.,&Npoints.,&Npoints.,&Npoints.,-1);
      %end;

      call pgfCompDer&FDIST(%do y=1 %to &FPARAMCOUNT; &&FPARAM&y, %end;f2,im_f2,&Npoints.);

      /* compute complex product */
      %do z=2 %to &SPARAMCOUNT;
         do i=1 to &Npoints.;
            temp1 = f2[i]*dfds&z.[i] - im_f2[i]*im_dfds&z.[i];
            temp2 = f2[i]*im_dfds&z.[i] + dfds&z.[i]*im_f2[i];
            dfds&z.[i] = temp1;
            im_dfds&z.[i] = temp2;
         end;
      %end;

      /* Inverse Transform */
      %do z=2 %to &SPARAMCOUNT;
         rc = zfft(dfdS&z.,im_dfdS&z.,&Npoints.,&Npoints.,&Npoints.,1);
      %end;

      /* Rescale */
      %do z=2 %to &SPARAMCOUNT;
         do i=1 to &Npoints.;
            dfdS&z.[i] = dfdS&z.[i]/&Npoints.;
         end;
      %end;

      /***** QUANTILE DERIVS *****/
      /* Now get partials of the quantile functions */
      %DO y=1 %TO &FPARAMCOUNT;
         dQdf&y[1] = dfdf&y[1];
      %end;

      %DO z=2 %TO &SPARAMCOUNT;
         dQdS&z.[1] = dfdS&z.[1];
      %end;
      do i=2 to &npoints.;
         %DO y=1 %TO &FPARAMCOUNT;
            dQdf&y[i] = dQdf&y[i-1] + dfdf&y[i];
         %end;
         %DO z=2 %TO &SPARAMCOUNT;
            dQdS&z.[i] = dQdS&z.[i-1] + dfdS&z.[i];
         %end;
      end;

      /***** RESULTS *****/
      do i=1 to 3;
         if var_indices[i] ne 1 then do;
            %do m = 1 %to &NSEV_MIX;
               %let weight = %scan(&SPARAM1_WEIGHTS,&m,' ');
               /* DVaR /DScales */
               results[quant_indices[i],&m.] = &weight.*
               %if &SCALE_TRANSFORM eq ID %then %do;
                  var_vals[quant_indices[i]]/ %scan(&SEV_SCALES, &m, ' ');
               %end;
               %else %do;
                 /* SEV_SCALES holds the log of the sev_scales */
                  var_vals[quant_indices[i]]/exp(%scan(&SEV_SCALES, &m, ' '));
               %end;
            %end;
            %DO z=2 %to &SPARAMCOUNT;
               %let ind = %eval(&NSEV_MIX + &z - 1);
               results[quant_indices[i],&ind] = -1*dQdS&z[var_indices[i]]/dQdV[quant_indices[i]];
            %end;   
            %do y = 1 %to &FPARAMCOUNT;
               %let ind = %eval(&SPARM_TOT + &y);
               results[quant_indices[i],&ind] = -1*dQdf&y[var_indices[i]]/dQdV[quant_indices[i]];
            %end;
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT.+1] = qv[quant_indices[i]];
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT+2] = var_vals[quant_indices[i]];
         end;
         else do;
            %DO z=1 %to %eval(&SPARAMCOUNT + &NSEV_MIX -1);
               results[quant_indices[i],&z.] =0;
            %end;   
            %do y = 1 %to &FPARAMCOUNT;
               results[quant_indices[i],&SPARM_TOT+&y] = 0;
            %end;
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT.+1] = qv[quant_indices[i]];
            results[quant_indices[i],&SPARM_TOT+&FPARAMCOUNT.+2] = var_vals[quant_indices[i]];
         end;
      end;

      /* How the results are labeled depends on whether or not there are severity risk 
         category level KRIs */
      rc = write_array("&outgrad.",results, 
         %if %sysevalf(&SEV_KRI_COUNT > 0, boolean) %then %do;
            %do z=1 %to &NSEV_MIX;
               /* Get the name of the loss grouping to append to the appropriate scale */
               %let suffix = %scan(&SEV_RC_IDS,&z,' ');
               "DSParam_1_&suffix.",
            %end;
            %do z=2 %to &SPARAMCOUNT;
               /*Now do the rest of the parameters */
               "DSparam_&z.",
            %end;
         %end;
         %else %do;
            %do z=1 %to &SPARAMCOUNT; 
               "DSParam_&z.", 
            %end;
         %end;
         %do y=1 %to &FPARAMCOUNT; 
            "DFParam_&y.", 
         %end;
         "Quantile", "VaR");
   run;

   /* If there are severity KRIs, add in the derivative with respect to severity param_1 */
   %if %sysevalf(&SEV_KRI_COUNT > 0, boolean) %then %do;
      data &outgrad.;
         set &outgrad;
         DSparam_1 = 0;
         %do z=1 %to &nsev_mix;
            %let suffix = %scan(&SEV_RC_IDS,&z,' ');
            %let scale = %scan(&SEV_SCALES,&z,' '); 
            %if &SCALE_TRANSFORM = ID %then %do;
               DSparam_1 = DSparam_1 + DSParam_1_&suffix.*&SCALE./&SParam_1;
            %end; 
            %else %do;
               DSparam_1 = DSparam_1 +  DSParam_1_&suffix.*exp(&SCALE.);
            %end;
         %end;
   %end;
  
   %let &abort_fpg = 0; 
   %DONE_FFT: 
   
%mend fft_param_grad;

%macro normal_param_grad( varpct = ,
                          outgrad =,
                          abort_npg =, 
                          abort_reason=
                        );

   %let &abort_npg = 1;

   /* This dataset will be fed into proc model to get derivatives with respect parameters */
   data _vg_norm;
      set &varpct;
      %do i=1 %to &FPARAMCOUNT;
         FPARAM_&i = &&FPARAM&i.;
      %end;
      %if %upcase(&SDIST) eq MIXDIST %then %do;
         Param_xLimit  = &Param_xLimit;
         Param_pnLimit = &Param_pnLimit;
      %end;
      %do i=2 %to &SPARAMCOUNT;
        PARAM_&i. = &&Sparam_&i.;
      %end;
      smean = 0;
      svar = 0;
      %do k=1 %to &nsev_mix;
         Param_1_&k. = %scan(&SEV_SCALES, &k, ' ');;
         Param_1 = Param_1_&k;
         %moments&SDIST;
         smean      = smean +  %scan(&SPARAM1_WEIGHTS, &k, ' ')*mean;
         svar       = svar + %scan(&SPARAM1_WEIGHTS, &k, ' ')*std**2;
      %end;
      
      %moments&FDIST;
      fmean      = mean;
      fvar       = variance;
      agg_mean   = fmean*smean;
      agg_var    = fmean*svar + fvar*smean**2;
      var1       = max(probit(quantile/100)*sqrt(agg_var) + agg_mean,0);
      %do i=2 %to %eval(&NSEV_MIX + &SPARAMCOUNT + &FPARAMCOUNT - 1);
         var&i = var1;
      %end;
      index = _N_;
      output;
      keep fparam: param_: var: quantile smean svar fmean fvar index;
   run;
   %do j=1 %to 3;
      data _vg_norm&j;
         set _vg_norm;
         where index = &j;
         call symputx("current_var_guess",var1);
         call symputx("current_var_actual",var_value);
      run;

      /* Handle the VaR <= 0 case: */
      %if %sysevalf( (&current_var_guess eq 0) or (&current_var_actual eq 0),BOOLEAN) %then %do;
         data _vg_norm_out&j;
            set  _vg_norm&j;
            %do i=1 %to &nsev_mix;
               DSParam_1_&i = 0;
            %end;
            %do i=2 %to &SPARAMCOUNT;
               DSParam_&i = 0;
            %end;
            %do i=1 %to &FPARAMCOUNT;
               DFParam_&i = 0;
            %end;
            keep
               quantile
               var_value
               %do i=1 %to &nsev_mix;
                  DSParam_1_&i
               %end;
               %do i=2 %to &SPARAMCOUNT;
                  DSParam_&i
               %end;
               %do i=1 %to &FPARAMCOUNT;
                  DFParam_&i 
              %end;
            ;
         run;
      %end;
      %else %do;
         proc model data=_vg_norm&j noprint;
            dependent 
               %do i=1 %to %eval(&NSEV_MIX + &SPARAMCOUNT + &FPARAMCOUNT - 1); 
                  var&i 
               %end;
            ;
            retain
               %do i=1 %to &nsev_mix;
                  Param_1_&i  %scan(&SEV_SCALES, &i, ' ')
               %end;
               %do i=2 %to &SPARAMCOUNT;
                  Param_&i &&SParam_&i
               %end;
               %do i=1 %to &FPARAMCOUNT;
                  FParam_&i &&FParam&i
               %end;
            ;
            smean = 0;
            svar = 0;
            %do k=1 %to &nsev_mix;
               Param_1 = Param_1_&k;
               %moments&SDIST;
               smean      = smean + %scan(&SPARAM1_WEIGHTS, &k, ' ')*mean;
               svar       = svar + %scan(&SPARAM1_WEIGHTS, &k, ' ')*std**2;
            %end;
            %moments&FDIST;
            fmean      = mean;
            fvar       = variance;
            agg_mean   = fmean*smean;
            agg_var    = fmean*svar + fvar*smean**2;      
            var1 =  probit(quantile/100)*sqrt(agg_var) + agg_mean;
            %do i=2 %to %eval(&NSEV_MIX + &SPARAMCOUNT + &FPARAMCOUNT - 1);
               var&i = pred.var1;
            %end;
            %do i=1 %to &nsev_mix;
               DSParam_1_&i = getder(pred.var&i,Param_1_&i);
            %end;
            %do i=2 %to &SPARAMCOUNT;
               %let ind = %eval(&nsev_mix+&i-1);
               DSParam_&i = getder(pred.var&ind,Param_&i);
            %end;
            %do i=1 %to &FPARAMCOUNT;
               %let ind = %eval(&SPARAMCOUNT + &nsev_mix -1 + &i);
               DFParam_&i = getder(pred.var&ind,FParam_&i);
            %end;
            outvars 
               %do i=1 %to &nsev_mix;
                  DSParam_1_&i
               %end;
               %do i=2 %to &SPARAMCOUNT;
                  DSParam_&i
               %end;
               %do i=1 %to &FPARAMCOUNT;
                  DFParam_&i 
               %end;
               quantile
               var_value;
            solve 
               %do i=1 %to &nsev_mix;
                  Param_1_&i
               %end;
               %do i=2 %to &SPARAMCOUNT;
                  Param_&i
               %end;
               %do i=1 %to &FPARAMCOUNT;
                  FParam_&i
               %end;
               /out=_vg_norm_out&j(keep=
                                   quantile
                                   var_value 
                                   %do i=1 %to &nsev_mix;
                                      DSParam_1_&i
                                   %end;
                                   %do i=2 %to &SPARAMCOUNT;
                                      DSParam_&i
                                   %end;
                                   %do i=1 %to &FPARAMCOUNT;
                                      DFParam_&i 
                                   %end;
                                  );
         run;
         quit;
         /* Check that proc model returned something.  
            TODO: with 9.3 proc model will put return code info in SYSRC, so check that instead */
         %orvl_ds_exists(_vg_norm_out&j, "orva_var_gradient", false);
         %if &ds_nobs eq 0 %then %do;
            %let &abort_reason = &orvar_error %sysfunc(sasmsg(&orvads, vargrad_no_model_output_error, noquote, &CCUID));
            %goto DONE_NORM;
         %end; 
      %end;
   %end;

   data &outgrad;
      set _vg_norm_out1 _vg_norm_out2 _vg_norm_out3;
      VaR = var_value;
      drop var_value;
   run;

   /* Adjust the severity scale parameter results */
   %if %sysevalf(&SEV_KRI_COUNT > 0, boolean) %then %do;
      data &outgrad;
         set &outgrad;
         DSParam_1 = 0;
         %do i=1 %to &NSEV_MIX;
            %let suffix = %scan(&SEV_RC_IDS,&i,' ');
            %let scale = %scan(&SEV_SCALES,1,' ');
            DSParam_1_&suffix.  = DSParam_1_&i;
            %if &SCALE_TRANSFORM = ID %then %do;
               DSparam_1 = DSparam_1 + DSParam_1_&suffix.*&SCALE./&SParam_1;
            %end; 
            %else %do;
               DSparam_1 = DSparam_1 + DSParam_1_&suffix.*&SCALE.*exp(&SParam_1);
            %end;
         %end;
         drop 
            %do i=1 %to &NSEV_MIX;
               DSParam_1_&i
            %end;
            ;
      run;
   %end;
   %else %do;
      data &outgrad;
         set &outgrad;
         DSParam_1 = DSParam_1_1;
         drop DSParam_1_1;
      run;
   %end;

   %let &abort_npg = 0;

   %DONE_NORM:

%mend normal_param_grad;
