/*****************************************************************************************
 *  Copyright (C), 2008
 *  SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 * 
 *  NAME:    orvl_int_range_missing_losses                                             
 *  PURPOSE: Checks missing range dataset against the loss dataset.
 *           Check each loss's FIRST_DATE_OF_EVENT against missing ranges. 
 *           If a loss falls within a missing range, print WARNING to the log and
 *           remove entry form the missing_range.
 *  SUPPORT: mimuha
 *
 *  PARAMETERS:
 *    OUTLI_LOSS      = Libref for the loss output datasets
 *    OUTLIB_ORG      = Libref for the org output datasets
 *    DATA_VERSION_ID = Version uid
 *****************************************************************************************/
 %macro orvl_int_range_missing_losses(OUTLIB_LOSS, OUTLIB_ORG, DATA_VERSION_ID);

  /* 
   * *************************************************
   * Make sure the required dataset has records. 
   * If it doesn't, orvl_ds_exists sets LOAD_STATUS to 2.
   * *************************************************
   */
 
   %orvl_ds_exists(&OUTLIB_ORG..MISSING_RANGE, "", false);
   %if &ds_nobs eq 0 %then %do;
     %orvl_cmn_status_success;
     %return;
   %end;
   
   %orvl_ds_exists(&OUTLIB_LOSS..INTERNAL_LOSS_EVENT, "", false);
   %if &ds_nobs eq 0 %then %do;
     %orvl_cmn_status_success;
     %return;
   %end;
 
   /* 
    * *************************************************
    * get parent_load_date (for incremental load) 
    * *************************************************
    */
   %let PARENT_UID= .;
   %let PARENT_LOAD_DATE=0;
   proc sql noprint;
       select PARENT_UID into: PARENT_UID from VARDATA.DATA_VERSION_HIER where VERSION_UID =&DATA_VERSION_ID;
   quit;
   %if &PARENT_UID ne . %then %do;
     data _NULL_; set VARDATA.LOAD_INFORMATION;
       where VERSION_UID = &DATA_VERSION_ID and DATA_TYPE = 'I';
       call symputx('PARENT_LOAD_DATE', LOAD_DATE);
     run;
   %end;
   
   /*
    * **********************************************************
    * Call %orva_assignown on missing_range and rename columns
    * *********************************************************
    */
   data _tmp_range; set &OUTLIB_ORG..MISSING_RANGE
     (rename=(INTERNAL_BL_UID =MR_INTERNAL_BL_UID
              INTERNAL_RC_UID =MR_INTERNAL_RC_UID
              PROCESS_UID     =MR_PROCESS_UID 
              LEGAL_ENTITY_UID=MR_LEGAL_ENTITY_UID
              CLASS_1         =MR_CLASS_1
              CLASS_2         =MR_CLASS_2
              CLASS_3         =MR_CLASS_3
              CLASS_4         =MR_CLASS_4
              CLASS_5         =MR_CLASS_5
              CLASS_6         =MR_CLASS_6
              CLASS_7         =MR_CLASS_7
              CLASS_8         =MR_CLASS_8
              CLASS_9         =MR_CLASS_9
              CLASS_10        =MR_CLASS_10
              FEED_REF_ID     =MR_FEED_REF_ID
     ));
     %if &PARENT_LOAD_DATE ne 0 %then %do;
       where DATETIME_LOADED > &PARENT_LOAD_DATE;
     %end;
     %orva_assignown(refid=MR_FEED_REF_ID);
     drop DATETIME_LOADED VERSION_UID;
   run;

   /* 
    * ************************************************************
    * Call %orva_assignown on loss table and keep minimal columns
    * ***********************************************************
    */
   data _tmp_losses; set &OUTLIB_LOSS..INTERNAL_LOSS_EVENT;
     %if &PARENT_LOAD_DATE ne 0 %then %do;
        where DATETIME_LOADED > &PARENT_LOAD_DATE;
     %end;
     %orva_assignown();
     keep OWN INTERNAL_BL_UID INTERNAL_RC_UID PROCESS_UID LEGAL_ENTITY_UID
          CLASS_1-CLASS_10 FIRST_DATE_OF_EVENT;
   run;
   
   /* 
    * ************************************************************
    * join losses and ranges by owner and range dates
    * ***********************************************************
    */
   proc sql noprint; 
      create table _tmp_losses_m as select * from 
         _tmp_losses L left join _tmp_range R on
         L.OWN = R.OWN and 
         L.FIRST_DATE_OF_EVENT GE R.START_MISSING_DATE and 
         L.FIRST_DATE_OF_EVENT LE R.END_MISSING_DATE where  MISSING_RANGE_UID ne .
         and (  MR_INTERNAL_BL_UID = . or MR_INTERNAL_BL_UID = INTERNAL_BL_UID) 
         and (MR_INTERNAL_RC_UID = . or MR_INTERNAL_RC_UID = INTERNAL_RC_UID) 
         and ( MR_PROCESS_UID = . or MR_PROCESS_UID = PROCESS_UID )
         and ( MR_LEGAL_ENTITY_UID = . or MR_LEGAL_ENTITY_UID = LEGAL_ENTITY_UID)
         and ( MR_CLASS_1 = . or MR_CLASS_1 = CLASS_1)
         and ( MR_CLASS_2 = . or MR_CLASS_2 = CLASS_2) 
         and ( MR_CLASS_3 = . or MR_CLASS_3 = CLASS_3)
         and ( MR_CLASS_4 = . or MR_CLASS_4 = CLASS_4)
         and ( MR_CLASS_5 = . or MR_CLASS_5 = CLASS_5)
         and ( MR_CLASS_6 = . or MR_CLASS_6 = CLASS_6)
         and ( MR_CLASS_7 = . or MR_CLASS_7 = CLASS_7)
         and ( MR_CLASS_8 = . or MR_CLASS_8 = CLASS_8)
         and ( MR_CLASS_9 = . or MR_CLASS_9 = CLASS_9)
         and ( MR_CLASS_10 = . or MR_CLASS_10 = CLASS_10);
   quit;

  
   %orvl_ds_exists(_tmp_losses_m, "", false);
   %if &ds_nobs eq 0 %then %do;
      %orvl_cmn_status_success;
      %goto MR_CLEANUP;
   %end;
   
   /* 
    * ************************************************************
    * check all cross classification variables
    * ***********************************************************
    */
   data _tmp_losses_m; set _tmp_losses_m;
    
      keep MISSING_RANGE_UID START_MISSING_DATE END_MISSING_DATE MR_:;
   run;
   
   proc sort data=_tmp_losses_m NODUPS; 
      by MISSING_RANGE_UID; 
   run;
   
   /* 
    * ************************************************************
    * drop missing range observations that we found losses for
    * ***********************************************************
    */
   %orvl_ds_exists(_tmp_losses_m, "", false);
   %if &ds_nobs eq 0 %then %do;
      %orvl_cmn_status_success;
      %goto MR_CLEANUP;
   %end;

   data _tmp_losses_m
        %orvl_log_ds_init();
      set _tmp_losses_m;
      length temp_str $200.;
      %orvl_log_ds_modify();
      temp_str = "STATUS WARNING: " || sasmsg("&orvlds", "irml_losses_found_so_dropping_range", "noquote",
                                              strip(put(int(MR_FEED_REF_ID), best15.)));
      %orvl_log_output(strip(temp_str));
   run;

   proc sql noprint; 
      Delete From &OUTLIB_ORG..MISSING_RANGE
         Where MISSING_RANGE_UID in (Select MISSING_RANGE_UID From _tmp_losses_m);
   quit;

   %orvl_cmn_status_success;
   
%MR_CLEANUP:
   /*- cleanup temporary datasets -*/                                                                                               
   proc datasets library=work memtype=data nolist nodetails;
      delete _tmp_losses _tmp_losses_m _tmp_range;       
   quit;        

   

%mend orvl_int_range_missing_losses;

/*****************************************************
%orvsetup(version=3.3);
options mprint nosymbolgen;
%orvl_int_range_missing_losses(LOSS, ORG, 23, 23);
 ******************************************************/
