/*************************************************************************************
 * $Id: orvp_signon.sas,v 1.1.4.16.2.6.2.2.2.1.2.1 2012/05/03 14:57:54 sasdje Exp $
 * Copyright (C), 2004
 * SAS Institute Inc., Cary, N.C. 27513, U.S.A.  All rights reserved.
 *
 * NAME   : orvp_SignOn.sas
 * PURPOSE: Macro to sign on to hosts for parallel processing.
 *          Runs the %orvsetup macro on each host.
 * SUPPORT: jariou
 *
 * PARAMETERS :
 *
 *   hostData         = Dataset containg host ids and port numbers for a SAS Connect
 *                      style network.  SAS Connect must be running on each host
 *                      specified in the dataset.
 *   gridData         = Dataset containing grid computing network info.
 *   user             = User ID (for SAS Connect only).
 *   password         = password (for SAS Connect only, should be encrypted)
 *   status           = status dataset.
 *
 * NOTES : this macro sets the following global MACRO vars:
 *   NumHosts          = number of hosts specified in hostData or grid network.
 *   NumHostsConnected = number of hosts successfully connected to.
 *   Status&iHost      = connection status of ith host.
 *   Host&iHost        = ith host name.
 *   RWork&iHost       = remote WORK libname for ith host.
 *
 **************************************************************************************/
%macro orvp_SignOn(  hostData  = , 
                     gridData  = , 
                     user      = , 
                     password  = , 
                     status    = , 
                     var_version = 4.2,
                     ctool     =
                     );
   %local maxHosts;

   data &status;
      length reason $100;
      status = 1;
      reason = "%sysfunc(sasmsg(&orvlds, unknown_error, noquote))";
   run;

   /*--- turn off SYNTAXCHECK, causes problems ---*/
   %local syntaxopt; 
   %let syntaxopt = %sysfunc(getoption(SYNTAXCHECK));

   %local notesopt;  
   %let notesopt  = %sysfunc(getoption(NOTES));

   %local mprintopt; 
   %let mprintopt = %sysfunc(getoption(MPRINT));

   options NOSYNTAXCHECK NONOTES;

   /*--- MACROVAR: Number of hosts in &hostData Dataset ---*/
   %global NumHosts;
   %let NumHosts = 0;

   /*--- MACROVAR: Number of hosts successfully connected to ---*/
   %global NumHostsConnected;
   %let NumHostsConnected = 0;

   /**********************************************************************/
   /*--- SAS Connect Network Specified ---*/
   /**********************************************************************/
   %if "&hostData" ne "" %then %do;
      /*--- Make sure that there is something in hostData ---*/
      %let NumHosts = 0;
      %let dsid     = %sysfunc(open(&hostData,I));

      %if &dsid ^= 0 %then %do;
         %let NumHosts = %sysfunc(attrn(&dsid,nlobs));
         %let rc = %sysfunc(close(&dsid));
      %end;

      %if &NumHosts eq 0 %then %do;
         %put &orvar_warning %sysfunc(sasmsg(&orvlds, son_cant_sign_on_error, noquote));

         data &status;
            length reason $100;
            status = 1;
            reason = "%sysfunc(sasmsg(&orvlds, son_empty_host_ds_error, noquote))";
         run;

         %return;
      %end;

      /*--- Assign each a Host an ID in dataset ---*/
      data &hostData; 
         set &hostData end = LAST;
         iHost = _N_;
      run;
   %end;
   /**********************************************************************/
   /*--- SAS Grid Network Specified ---*/
   /**********************************************************************/
   %else %if "&gridData" ne "" %then %do;
      /*--- Make sure that there is something in hostData ---*/
      %let nobs = 0;
      %let dsid = %sysfunc(open(&gridData,I));

      %if &dsid ^= 0 %then %do;
         %let nobs = %sysfunc(attrn(&dsid,nlobs));
         %let rc   = %sysfunc(close(&dsid));
      %end;

      %if &nobs eq 0 %then %do;
         %put &orvar_warning %sysfunc(sasmsg(&orvlds, son_cant_signon_grid_error, noquote));

         data &status;
            length reason $100;
            status = 1;
            reason = "%sysfunc(sasmsg(&orvlds, son_empty_grid_ds_error, noquote))";
         run;

         %return;
      %end;

      /*--- set up necessary options from &gridData ---*/
      data _NULL_; set &gridData;
         call symput('metaserver', trim(left(metaserver)));
         call symput('metaport', metaport);
         call symput('metarepository', trim(left(metarepository)));
         call symput('metauser', trim(left(metauser)));
         call symput('metapass', trim(left(metapass)));
         call symput('maxhosts', maxhosts);
      run;

      options metaserver     = "&metaserver";
      options metaport       = &metaport;
      options metarepository = "&metarepository";
      options metauser       = "&metauser";
      options metapass       = "&metapass";
      
      /*--- enable Grid Network ---*/
      %let rc = %sysfunc(grdsvc_enable(_all_, resource = SASMain));

      %if &rc ne 0 %then %do;
         %put &orvar_warning %sysfunc(sasmsg(&orvlds, son_sas_grid_error, noquote));

         data &status;
            length reason $100;
            status = 1;
            reason = "%sysfunc(sasmsg(&orvlds, son_sas_grid_error, noquote))";
         run;

         %return;
      %end;

      /*--- How many nodes availble? ---*/
      %let NumHosts = %sysfunc(grdsvc_nnodes(""));
      %put ORVP_SIGNON: %sysfunc(sasmsg(&orvlds, son_available_in_sas_grid_status, noquote, &NumHosts));
   %end;
   /**********************************************************************/
   /*--- No Parallel Network Specified ---*/
   /**********************************************************************/
   %else %do;
      %put &orvar_warning %sysfunc(sasmsg(&orvlds, son_specify_opts_error, noquote));

      data &status;
         length reason $100;
         status = 1;
         reason = "%sysfunc(sasmsg(&orvlds, son_no_specified_ds_error, noquote))";
      run;

      %return;
   %end;
   
   /*--- declare Host ID and status macros vars ---*/
   %do iHost = 1 %to &NumHosts;
      %global Host&iHost Status&iHost;
      %let Host&iHost = &iHost;
      %let Status&iHost = 0;
   %end;

   /*--- adjust num hosts if necessary ---*/
   %if "&maxHosts" ne "" %then %do;
      %if &maxHosts < &NumHosts %then %do;
          %let NumHosts = &maxHosts;
      %end;
   %end;

   /*-- loop over hosts and try to connect --*/
   %do iHost = 1 %to &NumHosts;
      %if "&hostData" ne "" %then %do;
         data _null_; 
            set &hostData(where =(iHost = &iHost));
            call symput('Port', trim(left(Port)));
            call symput('Host', trim(left(Host)));
         run; 

         %put ORVP_SIGNON: %sysfunc(sasmsg(&orvlds, son_starting_status, noquote, &iHost, &Host, &Port));
         %let Host&iHost = &Host &Port;

         %if &user NE %str() AND &password NE %str() %then %do;
            signon Host&iHost user = &user password = &password macvar = Status&iHost;
         %end;
         %else %do;
            signon Host&iHost macvar = Status&iHost;
         %end;
      %end;
      %else %do;
         signon Host&iHost macvar = Status&iHost;
      %end;

      %if &&Status&iHost eq 0 or &&Status&iHost eq 2 %then %do;
         %let NumHostsConnected = %eval(&NumHostsConnected + 1);
         libname RWork&iHost slibref = WORK server = Host&iHost;
      %end;
      %else %do;
         %let Status&iHost = -1;  
      %end;
   %end;

   data _VarVersion_;
      length Var_Version $16.;
      Var_Version = &var_version;
   run;

   /*-- run the %orvsetup macro on each host --*/
   %do iHost = 1 %to &NumHosts;
      %if &&Status&iHost ne -1 %then %do;
         options remote = Host&iHost;
         %syslput rem_ctool       = &ctool;  
         %syslput rem_mprintopt   = &mprintopt;  
         %syslput rem_notesopt    = &notesopt;  

         rsubmit remote = Host&iHost cwait = yes;
            options &rem_notesopt NOSYNTAXCHECK &rem_mprintopt;
            /*-- upload all the initialization files --*/
            proc upload  data = xml_lookup;            run;
            proc upload  data = sparmnames;            run;
            proc upload  data = sparmnames_2;          run;
            proc upload  data = fparmnames;            run;
            proc upload  data = fparmnames_2;          run;
            proc upload  data = scaletrans;            run;
            proc upload  data = orvfuncs index = YES;  run;
            proc upload  data = _VarVersion_;          run;

            data _null_;
               set _VarVersion_;
               call symput("Var_Version", Var_Version);
            run;

            /*--- initialize VaR ---*/
            %orvsetup(version = &VAR_VERSION);                 
            %orva_init(resetOption = NONE);

            %nrstr(%%)sysrput rem_ORVAR_SETUP = &ORVAR_SETUP;
         endrsubmit;

         %if "&rem_ORVAR_SETUP" ^= "1" %then %do;
            %put &orvar_error %sysfunc(sasmsg(&orvlds, son_not_set_for_orvar_error, noquote, &iHost, &&host, &iHost));
            %let Status&iHost = -1;
            %let NumHostsConnected = %eval(&NumHostsConnected - 1);
         %end;
      %end;
   %end;

   /*--- reset session's options ---*/
   options &syntaxopt &notesopt;

   data &status;
      length reason $100;

      if &NumHostsConnected < 1 then do;
         put "&orvar_warning %sysfunc(sasmsg(&orvlds, son_cant_signon_hosts_error, noquote))";
         status = 1;
         reason = "%sysfunc(sasmsg(&orvlds, son_no_hosts_error, noquote))";
      end;
      else do;
         put "ORVP_SIGNON: %sysfunc(sasmsg(&orvlds, son_success_connect_status, noquote, &NumHostsConnected))";
         status = 0;
         reason = "%sysfunc(sasmsg(&orvlds, son_connected_status, noquote, &NumHostsConnected))";
      end;
   run;
%mend orvp_SignOn;
